! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_ape_mod
    use param
    use erolib_general_mod
    use erolib_ffits_col_mod
    use erolib_sort_mod
    use erolib_tbintable_mod
    implicit none

    type, public :: tApeCorr

        integer :: nelements
        type(r4_tColumn) :: cts
        type(r4_tColumn) :: bkg
        type(r4_tColumn) :: exp
        type(r4_tColumn) :: eef
        type(r4_tColumn) :: radius
        type(r4_tColumn) :: pois

    contains

        procedure :: init => init_tApeCorr
        procedure :: free => free_tApeCorr
        procedure :: writeToFits => write_tApeCorr
        procedure :: writeToFits_frow_lrow => write_frow_lrow_tApeCorr
        generic ::  write => writeToFits, writeToFits_frow_lrow
        procedure :: readFromFits => read_tApeCorr
        procedure :: readFromFits_frow_lrow => read_frow_lrow_tApeCorr
        generic :: read => readFromFits, readFromFits_frow_lrow
        procedure :: apply_index => apply_index_tApeCorr
        procedure :: insert_into_tBinTable => insert_into_tBinTable_tApeCorr

    end type tApeCorr

contains

      !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tApeCorr
    !> allocate the memory for this type
    !
    !
    !> @param[inout] this
    !> @param[in]  nelements
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tApeCorr(this, nelements, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) ::  nelements

        class(tApeCorr) :: this
        character(*), parameter :: function = "init_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = nelements
        call this%cts%init("ape_cts", nelements, status, tunit="[counts]")
        call this%bkg%init("ape_bkg", nelements, status, tunit="[counts]")
        call this%exp%init("ape_exp", nelements, status, tunit="[s]")
        call this%eef%init("ape_eef", nelements, status)
        call this%radius%init("ape_radius", nelements, status, tunit="[sec]")
        call this%pois%init("ape_pois", nelements, status)

    end subroutine init_tApeCorr

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tApeCorr
    !> free the tApeCorr
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tApeCorr(this, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, jj

        class(tApeCorr) :: this
        character(*), parameter :: function = "free_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = 0
        call this%cts%free(status)
        call this%bkg%free(status)
        call this%exp%free(status)
        call this%eef%free(status)
        call this%radius%free(status)
        call this%pois%free(status)

    end subroutine free_tApeCorr

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tApeCorr
    !> write data to FITS fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[in, optional] extname
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tApeCorr(this, unit, status, extname)
        implicit none

        integer, intent(inout) :: status
        class(tApeCorr) :: this
        integer :: ii, jj
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), intent(in), optional :: extname

        character(*), parameter :: function = "write_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        if(present(extname)) call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%cts%write(unit, status, frow=frow, lrow=lrow)
            call this%bkg%write(unit, status, frow=frow, lrow=lrow)
            call this%exp%write(unit, status, frow=frow, lrow=lrow)
            call this%eef%write(unit, status, frow=frow, lrow=lrow)
            call this%radius%write(unit, status, frow=frow, lrow=lrow)
            call this%pois%write(unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine write_tApeCorr

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tApeCorr
    !> read the apecorr data from a fits file
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_tApeCorr(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        class(tApeCorr) :: this
        integer :: ii
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), parameter :: function = "read_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%cts%read(unit, status, frow=frow, lrow=lrow)
            call this%bkg%read(unit, status, frow=frow, lrow=lrow)
            call this%exp%read(unit, status, frow=frow, lrow=lrow)
            call this%eef%read(unit, status, frow=frow, lrow=lrow)
            call this%radius%read(unit, status, frow=frow, lrow=lrow)
            call this%pois%read(unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine read_tApeCorr

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tApeCorr
    !> write a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_frow_lrow_tApeCorr(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tApeCorr) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "write_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        call this%cts%write(unit, status, frow=frow, lrow=lrow)
        call this%bkg%write(unit, status, frow=frow, lrow=lrow)
        call this%exp%write(unit, status, frow=frow, lrow=lrow)
        call this%eef%write(unit, status, frow=frow, lrow=lrow)
        call this%radius%write(unit, status, frow=frow, lrow=lrow)
        call this%pois%write(unit, status, frow=frow, lrow=lrow)

    end subroutine write_frow_lrow_tApeCorr

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tApeCorr
    !> read a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_frow_lrow_tApeCorr(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tApeCorr) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "read_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        call this%cts%read(unit, status, frow=frow, lrow=lrow)
        call this%bkg%read(unit, status, frow=frow, lrow=lrow)
        call this%exp%read(unit, status, frow=frow, lrow=lrow)
        call this%eef%read(unit, status, frow=frow, lrow=lrow)
        call this%radius%read(unit, status, frow=frow, lrow=lrow)
        call this%pois%read(unit, status, frow=frow, lrow=lrow)

    end subroutine read_frow_lrow_tApeCorr

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_tApeCorr(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tApeCorr) :: this

        integer, dimension(:), intent(in) :: order

        character(*), parameter :: function = "apply_index_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = size(order)
        call this%cts%apply_index(order, status)
        call this%bkg%apply_index(order, status)
        call this%exp%apply_index(order, status)
        call this%eef%apply_index(order, status)
        call this%radius%apply_index(order, status)
        call this%pois%apply_index(order, status)

    end subroutine apply_index_tApeCorr

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tApeCorr
    !> insert the contents into a tbintable description
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout] table
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine insert_into_tBinTable_tApeCorr(this, table, status)
        implicit none

        integer, intent(inout) :: status

        type(tBinTable), intent(inout) :: table

        integer :: ii, jj

        class(tApeCorr) :: this
        character(*), parameter :: function = "insert_into_tBinTable_tApeCorr"

        if(status > 100 .or. status == s_errRep) return

        call table%insert_ffits_col(this%cts, status)
        call table%insert_ffits_col(this%bkg, status)
        call table%insert_ffits_col(this%exp, status)
        call table%insert_ffits_col(this%eef, status)
        call table%insert_ffits_col(this%radius, status)
        call table%insert_ffits_col(this%pois, status)

    end subroutine insert_into_tBinTable_tApeCorr

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge all columns of two tApe types
    !
    !
    !> @param[in] Ape_A, Ape_B
    !> @param[in] Ape_C
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_Ape(Ape_A, Ape_B, Ape_C, status)
        implicit none

        integer, intent(inout) :: status

        class(tApeCorr), intent(in) :: Ape_A
        class(tApeCorr), intent(in) :: Ape_B
        class(tApeCorr), intent(out) :: Ape_C

        character(*), parameter :: function = "merge_Ape"

        if(status > 100 .or. status == s_errRep) return

        if(Ape_A%nelements == 0 .and. Ape_B%nelements == 0) then
            call wrlog(function, "cats to merge have zero elements - nothing to be done here")
            return
        end if

        Ape_C%nelements = Ape_A%nelements + Ape_B%nelements

        call merge_tColumn(Ape_A%cts, Ape_B%cts, Ape_C%cts, status)
        call merge_tColumn(Ape_A%bkg, Ape_B%bkg, Ape_C%bkg, status)
        call merge_tColumn(Ape_A%exp, Ape_B%exp, Ape_C%exp, status)
        call merge_tColumn(Ape_A%eef, Ape_B%eef, Ape_C%eef, status)

        call merge_tColumn(Ape_A%radius, Ape_B%radius, Ape_C%radius, status)
        call merge_tColumn(Ape_A%pois, Ape_B%pois, Ape_C%pois, status)

    end subroutine merge_Ape

end module catprep_ape_mod
