! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_catextend_mod
    use param
    use erolib_general_mod
    use erolib_ffits_col_mod
    use erolib_sort_mod
    use erolib_tbintable_mod
    use catprep_varichck_mod
    use catprep_specfit_mod

    implicit none

    type, public :: tCatExtend

        integer(i4) :: nelements
        integer(i4) :: nsigma
        type(str_tColumn) :: souuid ! unique source id
        type(i4_tColumn) :: new ! flag for sources which have not been registered in the catalogs before
        ! source confusion number is a number indicating the complexity to cleary associate the source with sources in other catalogs
        ! for each associtated source identification the number of SC increases by 1
        ! therefore 0 = potential new source
        ! moreover each associated source is also checked against close sources which also increase the number by 1 (but ignoring the original association)
        ! therefore 1 = source with a unique associated source -> inherit SOUUID / IAUNAM
        ! all numbers >= 2 complex region which has to be manually looked at if one were to associate sources
        type(i4_tColumn) :: sc
        ! source candidates for the identification taking only first order candiates into account
        type(str_tColumn) :: sc1 ! SOUUID of the 1. best associated source
        type(str_tColumn) :: sc2 ! SOUUID of the 2. best associated source
        type(str_tColumn) :: sc3 ! SOUUID of the 3. best associated source
        type(str_tColumn) :: sc4 ! SOUUID of the 4. best associated source
        type(str_tColumn) :: sc5 ! SOUUID of the 5. best associated source
        type(str_tColumn) :: sc6 ! SOUUID of the 6. best associated source
        type(str_tColumn) :: sc7 ! SOUUID of the 7. best associated source
        type(str_tColumn) :: sc8 ! SOUUID of the 8. best associated source
        type(str_tColumn) :: sc9 ! SOUUID of the 9. best associated source

        type(r8_tColumn) :: qsc1 ! quality of the assocition to the 1. best associated source
        type(r8_tColumn) :: qsc2 ! quality of the assocition to the 2. best associated source
        type(r8_tColumn) :: qsc3 ! quality of the assocition to the 3. best associated source
        type(r8_tColumn) :: qsc4 ! quality of the assocition to the 4. best associated source
        type(r8_tColumn) :: qsc5 ! quality of the assocition to the 5. best associated source
        type(r8_tColumn) :: qsc6 ! quality of the assocition to the 6. best associated source
        type(r8_tColumn) :: qsc7 ! quality of the assocition to the 7. best associated source
        type(r8_tColumn) :: qsc8 ! quality of the assocition to the 8. best associated source
        type(r8_tColumn) :: qsc9 ! quality of the assocition to the 9. best associated source

        !        type(i4_tColumn) :: region ! region classification 1 = inside else outside
        type(i4_tColumn) :: unique ! source will/would be handled with this catalog
        type(str_tColumn) :: iaunam ! iauname
        !        type(i4_tColumn) :: rX ! number of sources within the vicinity of X (TBD) arc seconds of a source

        !        type(r8_tColumn) :: tstart ! tstart of stdhdr for this run
        !        type(r8_tColumn) :: tstop  ! tstop of stdhdr for this run

        ! this allows at later stages when merging multiple single scans to possibly identify transient behaviour
        type(str_tColumn) :: comment ! variable coments

    contains

        procedure :: init => init_tCatExtend
        procedure :: free => free_tCatExtend
        procedure :: writeToFits => write_tCatExtend
        procedure :: writeToFits_frow_lrow => write_frow_lrow_tCatExtend
        generic ::  write => writeToFits, writeToFits_frow_lrow
        procedure :: readFromFits => read_tCatExtend
        procedure :: readFromFits_frow_lrow => read_frow_lrow_tCatExtend
        generic :: read => readFromFits, readFromFits_frow_lrow
        procedure :: apply_index => apply_index_tCatExtend
        procedure :: insert_into_tBinTable => insert_into_tBinTable_tCatExtend

    end type tCatExtend

contains

      !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatExtend
    !> allocate the memory for this type
    !
    !
    !> @param[inout] this
    !> @param[in]  nelements
    !> @param[in, optional] vs_extend
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tCatExtend(this, n, status, vs_extend)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) ::  n

        logical, optional, intent(in) :: vs_extend

        class(tCatExtend) :: this
        character(*), parameter :: function = "init_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = n
        call this%souuid%init("souUID", n, status, nvec=14)
        call this%iaunam%init("IAUNAM", n, status, nvec=32)
        call this%new%init("NEW", n, status)
        !        call this%region%init("region", n, status)
        call this%unique%init("UNIQUE", n, status)

        call this%sc%init("SOUCONF", n, status)
        call this%sc1%init("souUID_1", n, status, nvec=14)
        call this%sc2%init("souUID_2", n, status, nvec=14)
        call this%sc3%init("souUID_3", n, status, nvec=14)
        call this%sc4%init("souUID_4", n, status, nvec=14)
        call this%sc5%init("souUID_5", n, status, nvec=14)
        call this%sc6%init("souUID_6", n, status, nvec=14)
        call this%sc7%init("souUID_7", n, status, nvec=14)
        call this%sc8%init("souUID_8", n, status, nvec=14)
        call this%sc9%init("souUID_9", n, status, nvec=14)

        call this%qsc1%init("souUID_QUAL_1", n, status)
        call this%qsc2%init("souUID_QUAL_2", n, status)
        call this%qsc3%init("souUID_QUAL_3", n, status)
        call this%qsc4%init("souUID_QUAL_4", n, status)
        call this%qsc5%init("souUID_QUAL_5", n, status)
        call this%qsc6%init("souUID_QUAL_6", n, status)
        call this%qsc7%init("souUID_QUAL_7", n, status)
        call this%qsc8%init("souUID_QUAL_8", n, status)
        call this%qsc9%init("souUID_QUAL_9", n, status)

        call this%comment%init("COMMENT", n, status, nvec=32)

    end subroutine init_tCatExtend

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatExtend
    !> free the tCatExtend
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tCatExtend(this, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, jj

        class(tCatExtend) :: this
        character(*), parameter :: function = "free_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = 0
        call this%souuid%free(status)
        call this%iaunam%free(status)
        call this%new%free(status)
        !        call this%region%free(status)
        call this%unique%free(status)

        call this%sc%free(status)

        call this%sc1%free(status)
        call this%sc2%free(status)
        call this%sc3%free(status)
        call this%sc4%free(status)
        call this%sc5%free(status)
        call this%sc6%free(status)
        call this%sc7%free(status)
        call this%sc8%free(status)
        call this%sc9%free(status)

        call this%qsc1%free(status)
        call this%qsc2%free(status)
        call this%qsc3%free(status)
        call this%qsc4%free(status)
        call this%qsc5%free(status)
        call this%qsc6%free(status)
        call this%qsc7%free(status)
        call this%qsc8%free(status)
        call this%qsc9%free(status)

        call this%comment%free(status)

    end subroutine free_tCatExtend

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatExtend
    !> write a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[in, optional] extname
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tCatExtend(this, unit, status, extname)
        implicit none

        integer, intent(inout) :: status
        class(tCatExtend) :: this
        integer :: ii, jj
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), intent(in), optional :: extname

        character(*), parameter :: function = "write_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        if(present(extname)) call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%souuid%write(unit, status, frow=frow, lrow=lrow)
            call this%iaunam%write(unit, status, frow=frow, lrow=lrow)
            call this%new%write(unit, status, frow=frow, lrow=lrow)
            !            call this%region%write(unit, status, frow=frow, lrow=lrow)
            call this%unique%write(unit, status, frow=frow, lrow=lrow)

            call this%sc%write(unit, status, frow=frow, lrow=lrow)

            call this%sc1%write(unit, status, frow=frow, lrow=lrow)
            call this%sc2%write(unit, status, frow=frow, lrow=lrow)
            call this%sc3%write(unit, status, frow=frow, lrow=lrow)
            call this%sc4%write(unit, status, frow=frow, lrow=lrow)
            call this%sc5%write(unit, status, frow=frow, lrow=lrow)
            call this%sc6%write(unit, status, frow=frow, lrow=lrow)
            call this%sc7%write(unit, status, frow=frow, lrow=lrow)
            call this%sc8%write(unit, status, frow=frow, lrow=lrow)
            call this%sc9%write(unit, status, frow=frow, lrow=lrow)

            call this%qsc1%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc2%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc3%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc4%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc5%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc6%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc7%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc8%write(unit, status, frow=frow, lrow=lrow)
            call this%qsc9%write(unit, status, frow=frow, lrow=lrow)

            call this%comment%write(unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine write_tCatExtend

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatExtend
    !> read a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_tCatExtend(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        class(tCatExtend) :: this
        integer :: ii
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), parameter :: function = "read_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%souuid%read(unit, status, frow=frow, lrow=lrow)
            call this%iaunam%read(unit, status, frow=frow, lrow=lrow)
            call this%new%read(unit, status, frow=frow, lrow=lrow)
            !            call this%region%read(unit, status, frow=frow, lrow=lrow)
            call this%unique%read(unit, status, frow=frow, lrow=lrow)

            call this%sc%read(unit, status, frow=frow, lrow=lrow)

            call this%sc1%read(unit, status, frow=frow, lrow=lrow)
            call this%sc2%read(unit, status, frow=frow, lrow=lrow)
            call this%sc3%read(unit, status, frow=frow, lrow=lrow)
            call this%sc4%read(unit, status, frow=frow, lrow=lrow)
            call this%sc5%read(unit, status, frow=frow, lrow=lrow)
            call this%sc6%read(unit, status, frow=frow, lrow=lrow)
            call this%sc7%read(unit, status, frow=frow, lrow=lrow)
            call this%sc8%read(unit, status, frow=frow, lrow=lrow)
            call this%sc9%read(unit, status, frow=frow, lrow=lrow)

            call this%qsc1%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc2%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc3%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc4%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc5%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc6%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc7%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc8%read(unit, status, frow=frow, lrow=lrow)
            call this%qsc9%read(unit, status, frow=frow, lrow=lrow)
                      !  call this%rx%read(unit, status, frow=frow, lrow=lrow)
            call this%comment%read(unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine read_tCatExtend

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatExtend
    !> write a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_frow_lrow_tCatExtend(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tCatExtend) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "write_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        call this%souuid%write(unit, status, frow=frow, lrow=lrow)
        call this%iaunam%write(unit, status, frow=frow, lrow=lrow)
        call this%new%write(unit, status, frow=frow, lrow=lrow)
        !        call this%region%write(unit, status, frow=frow, lrow=lrow)
        call this%unique%write(unit, status, frow=frow, lrow=lrow)

        call this%sc%write(unit, status, frow=frow, lrow=lrow)

        call this%sc1%write(unit, status, frow=frow, lrow=lrow)
        call this%sc2%write(unit, status, frow=frow, lrow=lrow)
        call this%sc3%write(unit, status, frow=frow, lrow=lrow)
        call this%sc4%write(unit, status, frow=frow, lrow=lrow)
        call this%sc5%write(unit, status, frow=frow, lrow=lrow)
        call this%sc6%write(unit, status, frow=frow, lrow=lrow)
        call this%sc7%write(unit, status, frow=frow, lrow=lrow)
        call this%sc8%write(unit, status, frow=frow, lrow=lrow)
        call this%sc9%write(unit, status, frow=frow, lrow=lrow)

        call this%qsc1%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc2%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc3%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc4%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc5%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc6%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc7%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc8%write(unit, status, frow=frow, lrow=lrow)
        call this%qsc9%write(unit, status, frow=frow, lrow=lrow)

        call this%comment%write(unit, status, frow=frow, lrow=lrow)

    end subroutine write_frow_lrow_tCatExtend

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatExtend
    !> read a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_frow_lrow_tCatExtend(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tCatExtend) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "read_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        call this%souuid%read(unit, status, frow=frow, lrow=lrow)
        call this%iaunam%read(unit, status, frow=frow, lrow=lrow)
        call this%new%read(unit, status, frow=frow, lrow=lrow)
        !        call this%region%read(unit, status, frow=frow, lrow=lrow)
        call this%unique%read(unit, status, frow=frow, lrow=lrow)

        call this%sc%read(unit, status, frow=frow, lrow=lrow)

        call this%sc1%read(unit, status, frow=frow, lrow=lrow)
        call this%sc2%read(unit, status, frow=frow, lrow=lrow)
        call this%sc3%read(unit, status, frow=frow, lrow=lrow)
        call this%sc4%read(unit, status, frow=frow, lrow=lrow)
        call this%sc5%read(unit, status, frow=frow, lrow=lrow)
        call this%sc6%read(unit, status, frow=frow, lrow=lrow)
        call this%sc7%read(unit, status, frow=frow, lrow=lrow)
        call this%sc8%read(unit, status, frow=frow, lrow=lrow)
        call this%sc9%read(unit, status, frow=frow, lrow=lrow)

        call this%qsc1%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc2%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc3%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc4%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc5%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc6%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc7%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc8%read(unit, status, frow=frow, lrow=lrow)
        call this%qsc9%read(unit, status, frow=frow, lrow=lrow)

        call this%comment%read(unit, status, frow=frow, lrow=lrow)

    end subroutine read_frow_lrow_tCatExtend

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_tCatExtend(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tCatExtend) :: this

        integer, dimension(:), intent(in) :: order

        character(*), parameter :: function = "apply_index_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = size(order)
        call this%souuid%apply_index(order, status)
        call this%iaunam%apply_index(order, status)
        call this%new%apply_index(order, status)
        !        call this%region%apply_index(order, status)
        call this%unique%apply_index(order, status)

        call this%sc%apply_index(order, status)

        call this%sc1%apply_index(order, status)
        call this%sc2%apply_index(order, status)
        call this%sc3%apply_index(order, status)
        call this%sc4%apply_index(order, status)
        call this%sc5%apply_index(order, status)
        call this%sc6%apply_index(order, status)
        call this%sc7%apply_index(order, status)
        call this%sc8%apply_index(order, status)
        call this%sc9%apply_index(order, status)

        call this%qsc1%apply_index(order, status)
        call this%qsc2%apply_index(order, status)
        call this%qsc3%apply_index(order, status)
        call this%qsc4%apply_index(order, status)
        call this%qsc5%apply_index(order, status)
        call this%qsc6%apply_index(order, status)
        call this%qsc7%apply_index(order, status)
        call this%qsc8%apply_index(order, status)
        call this%qsc9%apply_index(order, status)

        call this%comment%apply_index(order, status)

    end subroutine apply_index_tCatExtend

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatExtend
    !> insert the contents into a tbintable description
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout] table
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine insert_into_tBinTable_tCatExtend(this, table, status)
        implicit none

        integer, intent(inout) :: status

        type(tBinTable), intent(inout) :: table

        integer :: ii, jj

        class(tCatExtend) :: this
        character(*), parameter :: function = "insert_into_tBinTable_tCatExtend"

        if(status > 100 .or. status == s_errRep) return

        call table%insert_ffits_col(this%souuid, status)
        call table%insert_ffits_col(this%iaunam, status)
        call table%insert_ffits_col(this%new, status)
        !        call table%insert_ffits_col(this%region, status)
        call table%insert_ffits_col(this%unique, status)

        call table%insert_ffits_col(this%sc, status)

        call table%insert_ffits_col(this%comment, status)

        call table%insert_ffits_col(this%sc1, status)
        call table%insert_ffits_col(this%sc2, status)
        call table%insert_ffits_col(this%sc3, status)
        call table%insert_ffits_col(this%sc4, status)
        call table%insert_ffits_col(this%sc5, status)
        call table%insert_ffits_col(this%sc6, status)
        call table%insert_ffits_col(this%sc7, status)
        call table%insert_ffits_col(this%sc8, status)
        call table%insert_ffits_col(this%sc9, status)

        call table%insert_ffits_col(this%qsc1, status)
        call table%insert_ffits_col(this%qsc2, status)
        call table%insert_ffits_col(this%qsc3, status)
        call table%insert_ffits_col(this%qsc4, status)
        call table%insert_ffits_col(this%qsc5, status)
        call table%insert_ffits_col(this%qsc6, status)
        call table%insert_ffits_col(this%qsc7, status)
        call table%insert_ffits_col(this%qsc8, status)
        call table%insert_ffits_col(this%qsc9, status)

    end subroutine insert_into_tBinTable_tCatExtend

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge all columns of two tCatExtend types
    !
    !
    !> @param[in] CatExtend_A, CatExtend_B
    !> @param[in] CatExtend_C
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_CatExtend(Cat_A, Cat_B, Cat_C, status)
        implicit none

        integer, intent(inout) :: status

        class(tCatExtend), intent(in) :: Cat_A
        class(tCatExtend), intent(in) :: Cat_B
        class(tCatExtend), intent(out) :: Cat_C

        character(*), parameter :: function = "merge_CatExtend"

        if(status > 100 .or. status == s_errRep) return

        if(Cat_A%nelements == 0 .and. Cat_B%nelements == 0) then
            call wrlog(function, "cats to merge have zero elements - nothing to be done here")
            return
        end if

        Cat_C%nelements = Cat_A%nelements + Cat_B%nelements

        call merge_tColumn(Cat_A%souuid, Cat_B%souuid, Cat_C%souuid, status)
        call merge_tColumn(Cat_A%iaunam, Cat_B%iaunam, Cat_C%iaunam, status)
        call merge_tColumn(Cat_A%new, Cat_B%new, Cat_C%new, status)
        call merge_tColumn(Cat_A%unique, Cat_B%unique, Cat_C%unique, status)

        call merge_tColumn(Cat_A%sc, Cat_B%sc, Cat_C%sc, status)

        call merge_tColumn(Cat_A%sc1, Cat_B%sc1, Cat_C%sc1, status)
        call merge_tColumn(Cat_A%sc2, Cat_B%sc2, Cat_C%sc2, status)
        call merge_tColumn(Cat_A%sc3, Cat_B%sc3, Cat_C%sc3, status)
        call merge_tColumn(Cat_A%sc4, Cat_B%sc4, Cat_C%sc4, status)
        call merge_tColumn(Cat_A%sc5, Cat_B%sc5, Cat_C%sc5, status)
        call merge_tColumn(Cat_A%sc6, Cat_B%sc6, Cat_C%sc6, status)
        call merge_tColumn(Cat_A%sc7, Cat_B%sc7, Cat_C%sc7, status)
        call merge_tColumn(Cat_A%sc8, Cat_B%sc8, Cat_C%sc8, status)
        call merge_tColumn(Cat_A%sc9, Cat_B%sc9, Cat_C%sc9, status)

        call merge_tColumn(Cat_A%qsc1, Cat_B%qsc1, Cat_C%qsc1, status)
        call merge_tColumn(Cat_A%qsc2, Cat_B%qsc2, Cat_C%qsc2, status)
        call merge_tColumn(Cat_A%qsc3, Cat_B%qsc3, Cat_C%qsc3, status)
        call merge_tColumn(Cat_A%qsc4, Cat_B%qsc4, Cat_C%qsc4, status)
        call merge_tColumn(Cat_A%qsc5, Cat_B%qsc5, Cat_C%qsc5, status)
        call merge_tColumn(Cat_A%qsc6, Cat_B%qsc6, Cat_C%qsc6, status)
        call merge_tColumn(Cat_A%qsc7, Cat_B%qsc7, Cat_C%qsc7, status)
        call merge_tColumn(Cat_A%qsc8, Cat_B%qsc8, Cat_C%qsc8, status)
        call merge_tColumn(Cat_A%qsc9, Cat_B%qsc9, Cat_C%qsc9, status)

        call merge_tColumn(Cat_A%comment, Cat_B%comment, Cat_C%comment, status)

    end subroutine merge_CatExtend


end module catprep_catextend_mod
