! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_compcat_mod
    use param
    use erolib_tbintable_mod
    use erolib_region_mod
    use erolib_ffits_col_mod
    use erolib_str_mod
    use catprep_srclist_mod

    use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr

    implicit none

logical, private, parameter :: doTest=.false.

    ! extnames used in the files
    character(*), parameter, private :: CompCat_ext = "CATSRCL"

    ! basic type of the catalog used for the cumulative list
    type, public :: tCompCat

        integer :: unit
        character(buffsize) :: filename
        character(buffsize) :: extname
        type(tBinTable) :: table ! description of the binary table
        type(stdhdr_type) :: stdhdr

        integer :: nsouuid ! header keyword
        integer :: nelements_used ! number of elements filled
        integer :: nelements ! number of elements (rows)

        type(str_tColumn) :: souuid     ! [souuid]
        type(str_tColumn) :: iaunam     ! [iaunam]
        type(str_tColumn) :: detuid     ! [detuid]
        type(r8_tColumn) :: ra          ! [deg]
        type(r8_tColumn) :: dec         ! [deg]
        type(r4_tColumn) :: radec_err   ! [sec] 1 sigma error
        type(r4_tColumn) :: flux        !   [cgs]              Source flux 2)
        type(r4_tColumn) :: flux_err    !   [cgs]              1 sigma error 2)
        type(str_tColumn) :: comment    ! [] comment
        type(i1_tColumn) :: new         ! indicating if a source was added in the previous run
!        type(i4_tColumn) :: active         ! currently not used !DISCUSS indicating if a source was updated in the previous run
        type(i4_tColumn) :: id_cluster     ! indicating sources which have been detected in the same ermldet run and had the same id_cluster
        type(r4_tColumn) :: ext         ! if /=0.0_r4 source is extended
        type(r4_tColumn) :: ext_err     ! error of the size

        ! auxilliary column which is not written / saved
        type(i4_tColumn) :: region
        type(i4_tColumn) :: id ! representing the row number / original order


    contains
        procedure :: init => init_tCompCat
        procedure :: free => free_tCompCat
        procedure :: open => open_tCompCat
        procedure :: close => close_tCompCat
        procedure :: create_extension => create_extension_tCompCat
        procedure :: read => read_tCompCat
        procedure :: write => write_tCompCat
        procedure :: apply_index => apply_index_tCompCat
        procedure :: sort => sort_tCompCat
        procedure :: flag_region_ra_dec => flag_region_ra_dec_tCompCat
        procedure :: addnew => addnew_tCompCat
        procedure :: update => update_tCompCat

    end type tCompCat

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloacte n CompCat structures
    !
    !
    !
    !> @param[inout] CompCat
    !> @param[in] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine alloc_CompCat(CompCat, n, status)
        implicit none

        integer :: stat
        integer, intent(in) :: n
        integer, intent(inout) :: status

        class(tCompCat), allocatable, dimension(:) :: CompCat

        character(buffsize) :: errmsg
        character(*), parameter :: function = "alloc_CompCat"

        call dealloc_CompCat(CompCat, status)
        allocate(CompCat(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

    end subroutine alloc_CompCat

      !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> dealloacte CompCat structures
    !
    !
    !
    !> @param[inout] CompCat
    !> @param[inout] status
    !---------------------------------------------------------------------------
    subroutine dealloc_CompCat(CompCat, status)

        integer :: stat
        integer, intent(inout) :: status

        class(tCompCat), allocatable, dimension(:) :: CompCat

        character(*), parameter :: function = "dealloc_CompCat"

        if(allocated(CompCat)) then
            deallocate(CompCat, stat=stat)
            call FError(function, stat, "dealloc CompCat type", status)
        end if

    end subroutine dealloc_CompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> init the tCompCat type
    !> nband = number of different energy bands
    !
    !
    !> @param[in] this
    !> @param[out] n
    !> @param[inout] status
    !> @param[in, optional] extend
    !---------------------------------------------------------------------------

    subroutine init_tCompCat(this, n, status, extend, extra)
        implicit none

        integer, intent(inout) :: status
        class(tCompCat) :: this
        integer, intent(in) :: n

        logical, optional, intent(in) :: extend
        integer, optional, intent(in) :: extra

        character(*), parameter :: function = "init_tCompCat"

        if(status > 100 .or. status == s_errRep) return
        this%nelements = n
        if(present(extra)) then
            if(extra < 0) then
                call wrstop(function, "tried to initialize comparison catalog with a negative number of extra elements")
            end if
            this%nelements = this%nelements + extra
            this%nelements_used = n
        else
            this%nelements_used = 0
        end if

        this%nsouuid = 0

        this%extname = CompCat_ext

        call this%souuid%init(    "souUID",        this%nelements, status, nvec=14)
        call this%detuid%init(    "detUID",        this%nelements, status, nvec=32)
        call this%iaunam%init(    "IAUNAM",        this%nelements, status, nvec=32)
        call this%ra%init(        "RA",            this%nelements, status, tunit="[deg]")
        call this%dec%init(       "DEC",           this%nelements, status, tunit="[deg]")
        call this%radec_err%init( "RADEC_ERR",     this%nelements, status, tunit="[sec]")
        call this%flux%init(      "ML_FLUX_0",     this%nelements, status)
        call this%flux_err%init(  "ML_FLUX_ERR_0", this%nelements, status)
        call this%comment%init(   "COMMENT",       this%nelements, status, nvec=32)
        call this%new%init(       "NEW",           this%nelements, status)
!        call this%active%init(    "ACTIVE",        this%nelements, status)
        call this%id_cluster%init("ID_CLUSTER",    this%nelements, status)
        call this%ext%init(       "EXT",           this%nelements, status, tunit="[sec]")
        call this%ext_err%init(   "EXT_ERR",       this%nelements, status, tunit="[sec]")

        call this%region%init(    "REGION",        this%nelements, status)
        call this%id%init(        "ID",            this%nelements, status)

    end subroutine init_tCompCat

    subroutine init_tCompCat_20200728(this, n, status, extend, extra)
        implicit none

        integer, intent(inout) :: status
        class(tCompCat) :: this
        integer, intent(in) :: n

        logical, optional, intent(in) :: extend
        integer, optional, intent(in) :: extra

        character(*), parameter :: function = "init_tCompCat_20200728"

call wrstop(function, "obsolete")

        if(status > 100 .or. status == s_errRep) return
        this%nelements = n
        if(present(extra)) then
            if(extra < 0) then
                call wrstop(function, "tried to initialize comparison catalog with a negative number of extra elements")
            end if
            this%nelements = this%nelements + extra
            this%nelements_used = n
        else
            this%nelements_used = 0
        end if

        this%nsouuid = 0

        this%extname = CompCat_ext

        call this%souuid%init("souUID", n, status, nvec=14)
        call this%detuid%init("detUID", n, status, nvec=32)
        call this%iaunam%init("IAUNAM", n, status, nvec=32)
        call this%ra%init("RA", n, status, tunit="[deg]")
        call this%dec%init("DEC", n, status, tunit="[deg]")
        call this%radec_err%init("RADEC_ERR", n, status, tunit="[sec]")
        call this%flux%init("ML_FLUX_0", n, status)
        call this%flux_err%init("ML_FLUX_ERR_0", n, status)
        call this%comment%init("COMMENT", n, status, nvec=32)
        call this%new%init("NEW", n, status)
!        call this%active%init("ACTIVE", n, status)
        call this%id_cluster%init("ID_CLUSTER", n, status)
        call this%ext%init("EXT", n, status, tunit="[sec]")
        call this%ext_err%init("EXT_ERR", n, status, tunit="[sec]")

        call this%region%init("REGION", n, status)
        call this%id%init("ID", n, status)

    end subroutine init_tCompCat_20200728

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the tCompCat type
    !
    !
    !> @param[in] this
    !> @param[out] n
    !> @param[inout] status
    !> @param[in, optional] nband
    !---------------------------------------------------------------------------

    subroutine free_tCompCat(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tCompCat) :: this

        character(*), parameter :: function = "free_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = 0

        call this%souuid%free(status)
        call this%detuid%free(status)
        call this%iaunam%free(status)
        call this%ra%free(status)
        call this%dec%free(status)
        call this%radec_err%free(status)

        call this%flux%free(status)
        call this%flux_err%free(status)
        call this%comment%free(status)
        call this%new%free(status)
!        call this%active%free(status)
        call this%id_cluster%free(status)

        call this%ext%free(status)
        call this%ext_err%free(status)

        call this%region%free(status)
        call this%id%free(status)

    end subroutine free_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> routine to open the CompCat
    !
    !
    !
    !> @param[inout] this
    !> @param[in] filename, rwmode
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine open_tCompCat(this, filename, rwmode, status, delete)
        implicit none

        integer, intent(inout) :: status
        class(tCompCat) :: this

        character(*), intent(in) :: filename
        integer, intent(in) :: rwmode

        integer :: exists

        integer, intent(in), optional :: delete

        character(*), parameter :: function = "open_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        this%filename = filename

        call ffits_get_iounit(this%unit, status)
        if(rwmode /= create) then
            call ffits_open_file(this%unit, filename, rwmode, status)
            if(status == ok) call wrlog(function, "CompCat " // trim(this%filename) // " successfully opened")
        else
            call ffits_file_exists(filename, exists, status)
            if(present(delete) .and. exists == 1) then
                if(delete == true) call delete_file_fortran(filename, status)
            end if
            call ffits_create_file(this%unit, filename, fblocksize, status)
            if(status == ok) call wrlog(function, "CompCat " // trim(this%filename) // " successfully created")
        end if

    end subroutine open_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> close the tCompCat file
    !
    !
    !
    !> @param[inout] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine close_tCompCat(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tCompCat) :: this

        character(*), parameter :: function = "close_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        call ffits_close_file(this%unit, status)
        call ffits_free_iounit(this%unit, status)

    end subroutine close_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCompCat
    !> create the extension of this type in a FITS file
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tCompCat(this, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ncol ! number of columns in the file

        class(tCompCat) :: this
        character(*), parameter :: function = "create_extension_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        ncol = 20

        call this%table%init(ncol, status)

        call this%table%insert_ffits_col(this%souuid, status)
        call this%table%insert_ffits_col(this%detuid, status)
        call this%table%insert_ffits_col(this%iaunam, status)
        call this%table%insert_ffits_col(this%ra, status)
        call this%table%insert_ffits_col(this%dec, status)
        call this%table%insert_ffits_col(this%radec_err, status)

        call this%table%insert_ffits_col(this%flux, status)
        call this%table%insert_ffits_col(this%flux_err, status)
        call this%table%insert_ffits_col(this%comment, status)
        call this%table%insert_ffits_col(this%new, status)
!        call this%table%insert_ffits_col(this%active, status)
        call this%table%insert_ffits_col(this%id_cluster, status)

        call this%table%insert_ffits_col(this%ext, status)
        call this%table%insert_ffits_col(this%ext_err, status)

        call this%table%insertFits(this%unit, this%extname, status)

    end subroutine create_extension_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the CompCat type from a FITS file
    !
    !
    !
    !> @param[in] this
    !> @param[out] filename
    !> @param[inout] status
    !> @param[in, optional] extend, used
    !---------------------------------------------------------------------------

    subroutine read_tCompCat(this, status, extend, extra)
        implicit none

        integer, intent(inout) :: status
        integer :: nelements, rowsize, frow, lrow, nrows, ii

        integer, intent(in), optional :: extra

        class(tCompCat) :: this
        logical, optional, intent(in) :: extend

        character(fcolnamelength) :: col
        character(*), parameter :: function = "read_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        call ffits_movnam_hdu(this%unit, BINARY_TBL, CompCat_ext, status)
        call ffits_get_num_rows(this%unit, nelements, status)
        call ffits_get_rowsize(this%unit, rowsize, status)

        status = ok
        if(present(extra)) then
            if(extra < 0) then
                call wrstop(function, "the number of extra elements is less than zero - contact the developers")
            end if
            if(present(extend)) then
                call this%init(nelements, status, extend, extra=extra)
            else
                call this%init(nelements, status, extra=extra)
            end if

        else
            if(present(extend)) then
                call this%init(nelements, status, extend)
            else
                call this%init(nelements, status)
            end if
        end if

        call fread_stdhdr(this%unit, this%stdhdr, status)

        status = ignore
        call ffits_read_key(this%unit, "nsouuid", this%nsouuid, status)
        if(status /= ignore) then
            call wrlog(function, "comparison catalog has no keyword nsouuid. nsouuid will be derived from the nelements in the catalog")
            this%nsouuid = nelements
        end if
        status = ok

        if(nelements == 0 .or. status == s_errRep) return

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%souuid%read(this%unit, status, frow=frow, lrow=lrow)
            call this%detuid%read(this%unit, status, frow=frow, lrow=lrow)
            call this%iaunam%read(this%unit, status, frow=frow, lrow=lrow)
            call this%ra%read(this%unit, status, frow=frow, lrow=lrow)
            call this%dec%read(this%unit, status, frow=frow, lrow=lrow)
            call this%radec_err%read(this%unit, status, frow=frow, lrow=lrow)

            call this%flux%read(this%unit, status, frow=frow, lrow=lrow)
            call this%flux_err%read(this%unit, status, frow=frow, lrow=lrow)
            call this%comment%read(this%unit, status, frow=frow, lrow=lrow)
            call this%new%read(this%unit, status, frow=frow, lrow=lrow)
            call this%id_cluster%read(this%unit, status, frow=frow, lrow=lrow)
            call this%ext%read(this%unit, status, frow=frow, lrow=lrow)
            call this%ext_err%read(this%unit, status, frow=frow, lrow=lrow)

!            call this%active%read(this%unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine read_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write the CompCat type to a FITS file
    !
    !
    !
    !> @param[in] this
    !> @param[out] filename
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tCompCat(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tCompCat) :: this
        integer :: rowsize, ii, frow, nelements, lrow, nrows

        character(*), parameter :: function = "write_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
        if(status /= ignore) then
            !~       print *, "create extension"
            status = ok
            call this%create_extension(status)
        else
            status = ok
        !~       print *, "extension already exists"
        end if

        call ffits_get_rowsize(this%unit, rowsize, status)
        nelements = this%nelements

        call fwrite_stdhdr(this%unit, this%stdhdr, status)
        call ffits_update_key(this%unit, "nsouuid", this%nsouuid, "number of registered souuid for this skyfield", status)

        ! write 1D arrays
        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%detuid%write(this%unit, status, frow=frow, lrow=lrow)
            call this%souuid%write(this%unit, status, frow=frow, lrow=lrow)
            call this%iaunam%write(this%unit, status, frow=frow, lrow=lrow)
            call this%ra%write(this%unit, status, frow=frow, lrow=lrow)
            call this%dec%write(this%unit, status, frow=frow, lrow=lrow)
            call this%radec_err%write(this%unit, status, frow=frow, lrow=lrow)

            call this%flux%write(this%unit, status, frow=frow, lrow=lrow)
            call this%flux_err%write(this%unit, status, frow=frow, lrow=lrow)
            call this%comment%write(this%unit, status, frow=frow, lrow=lrow)
            call this%new%write(this%unit, status, frow=frow, lrow=lrow)
!            call this%active%write(this%unit, status, frow=frow, lrow=lrow)
            call this%id_cluster%write(this%unit, status, frow=frow, lrow=lrow)

            call this%ext%write(this%unit, status, frow=frow, lrow=lrow)
            call this%ext_err%write(this%unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine write_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_tCompCat(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tCompCat) :: this

        integer, dimension(:), intent(in) :: order

        character(*), parameter :: function = "apply_index_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = size(order)
        call this%souuid%apply_index(order, status)
        call this%detuid%apply_index(order, status)
        call this%iaunam%apply_index(order, status)
        call this%ra%apply_index(order, status)
        call this%dec%apply_index(order, status)
        call this%radec_err%apply_index(order, status)

        call this%flux%apply_index(order, status)
        call this%flux_err%apply_index(order, status)
        call this%comment%apply_index(order, status)
        call this%new%apply_index(order, status)
!        call this%active%apply_index(order, status)
        call this%id_cluster%apply_index(order, status)

        call this%ext%apply_index(order, status)
        call this%ext_err%apply_index(order, status)

        call this%region%apply_index(order, status)
        call this%id%apply_index(order, status)

    end subroutine apply_index_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge all entries of the columns with a second tCompCat
    !
    !
    !> @param[in] Cat_A, Cat_B
    !> @param[in] Cat_C
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_CompCat(Cat_A, Cat_B, Cat_C, status)
        implicit none

        integer, intent(inout) :: status

        class(tCompCat), intent(in) :: Cat_A
        class(tCompCat), intent(in) :: Cat_B
        class(tCompCat), intent(out) :: Cat_C

        character(*), parameter :: function = "merge_CompCat"

        if(status > 100 .or. status == s_errRep) return

        ! copy all important informations
        Cat_C%table        = Cat_A%table
        Cat_C%stdhdr        = Cat_A%stdhdr

        if(Cat_A%nelements == 0 .and. Cat_B%nelements == 0) then
            call wrlog(function, "cats to merge have zero elements - nothing to be done here")
            return
        end if

        Cat_C%nelements = Cat_A%nelements + Cat_B%nelements

        call merge_tColumn(Cat_A%souuid, Cat_B%souuid, Cat_C%souuid, status)
        call merge_tColumn(Cat_A%detuid, Cat_B%detuid, Cat_C%detuid, status)
        call merge_tColumn(Cat_A%iaunam, Cat_B%iaunam, Cat_C%iaunam, status)
        call merge_tColumn(Cat_A%ra, Cat_B%ra, Cat_C%ra, status)
        call merge_tColumn(Cat_A%dec, Cat_B%dec, Cat_C%dec, status)
        call merge_tColumn(Cat_A%radec_err, Cat_B%radec_err, Cat_C%radec_err, status)

        call merge_tColumn(Cat_A%flux, Cat_B%flux, Cat_C%flux, status)
        call merge_tColumn(Cat_A%flux_err, Cat_B%flux_err, Cat_C%flux_err, status)

        call merge_tColumn(Cat_A%comment, Cat_B%comment, Cat_C%comment, status)
        call merge_tColumn(Cat_A%new, Cat_B%new, Cat_C%new, status)
!        call merge_tColumn(Cat_A%active, Cat_B%active, Cat_C%active, status)
        call merge_tColumn(Cat_A%id_cluster, Cat_B%id_cluster, Cat_C%id_cluster, status)

        call merge_tColumn(Cat_A%ext, Cat_B%ext, Cat_C%ext, status)
        call merge_tColumn(Cat_A%ext_err, Cat_B%ext_err, Cat_C%ext_err, status)

        call merge_tColumn(Cat_A%region, Cat_B%region, Cat_C%region, status)
        call merge_tColumn(Cat_A%id, Cat_B%id, Cat_C%id, status)


    end subroutine merge_CompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine to mark sources in regions by rectangle ra_min / ...
    !> the overlap parameter defining the boarder region (deg)
    !
    !> @param[in] this
    !> @param[in, optional] overlap
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine flag_region_ra_dec_tCompCat(this, ra_min_in, ra_max_in, dec_min_in, dec_max_in, status, overlap, value)
        implicit none

        integer, intent(inout) :: status
        class(tCompCat) :: this

        real(r8) :: dec_min, dec_max
        real(r8), intent(in) :: ra_min_in, ra_max_in, dec_min_in, dec_max_in
        real(r8), dimension(this%nelements) :: ra, dec, ra_min, ra_max, overlap_ra
        integer :: ii, old_value
        integer, intent(in), optional :: value
        real(r8), intent(in), optional :: overlap

        character(*), parameter :: function = "flag_region_ra_dec_tCompCat"

        ra = mod(this%ra%value, 360.0_r8)
        dec = this%dec%value
        ra_min = mod(ra_min_in, 360.0_r8)
        ra_max = mod(ra_max_in, 360.0_r8)

        ! potentially correct for negative ra
        where (ra < tiny(0.0_r8))
            ra = ra + 360.0_r8
        end where

        where (ra_min < tiny(0.0_r8))
            ra_min = ra_min + 360.0_r8
        end where

        where (ra_max < tiny(0.0_r8))
            ra_max = ra_max + 360.0_r8
        end where

        overlap_ra = overlap / cos ( dec * PI / 180.0_r8)

        if(present(overlap)) then
            ! correction center for the ra overlap
            ! adjust the overlap for declination 1 / cos (angle)
            ra_min = ra_min - overlap_ra
            ra_max = ra_max + overlap_ra
            dec_min = dec_min_in - overlap
            dec_max = dec_max_in + overlap
        else
            ! ra_min / ra_max remain unchanged besides the corrections for negative values and mod 360
            dec_min = dec_min_in
            dec_max = dec_max_in
        end if

        do ii = 1, this%nelements

            ! differ two cases to replace the old region value with a custom value or 1 if it is within a region
            if(present(value)) then
                ! since we are on a sphere and 0/360 is the same we have
                ! two cases ra_min > ra_ max and ra_min < ra_max
                if(ra_min(ii) < ra_max(ii)) then
                    old_value = this%region%value(ii)
                    this%region%value(ii) = value
                    if(ra(ii) < ra_min(ii)) this%region%value(ii) = old_value
                    if(ra(ii) > ra_max(ii)) this%region%value(ii) = old_value
                    if(dec(ii) < dec_min) this%region%value(ii) = old_value
                    if(dec(ii) > dec_max) this%region%value(ii) = old_value
                else
                    if(ra(ii) >= ra_min(ii)) this%region%value(ii) = value
                    if(ra(ii) <= ra_max(ii)) this%region%value(ii) = value
                    if(dec(ii) >= dec_min) this%region%value(ii) = value
                    if(dec(ii) <= dec_max) this%region%value(ii) = value
                end if

            else
                if(ra_min(ii) < ra_max(ii)) then
                    old_value = this%region%value(ii)
                    this%region%value(ii) = 1
                    if(ra(ii) < ra_min(ii)) this%region%value(ii) = old_value
                    if(ra(ii) > ra_max(ii)) this%region%value(ii) = old_value
                    if(dec(ii) < dec_min) this%region%value(ii) = old_value
                    if(dec(ii) > dec_max) this%region%value(ii) = old_value
                else
                    if(ra(ii) >= ra_min(ii)) this%region%value(ii) = 1
                    if(ra(ii) <= ra_max(ii)) this%region%value(ii) = 1
                    if(dec(ii) >= dec_min) this%region%value(ii) = 1
                    if(dec(ii) <= dec_max) this%region%value(ii) = 1
                end if

            end if

        end do

    end subroutine flag_region_ra_dec_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> sort the type entries according to one array
    !
    !
    !
    !> @param[in] this
    !> @param[in] array
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_tCompCat(this, array, status, ascend)
        implicit none

        integer, intent(inout) :: status
        integer, optional :: ascend

        class(tCompCat) :: this
        class(*), dimension(:) :: array

        integer, dimension(:), allocatable :: order

        real(r8),allocatable, dimension(:) :: tmparray_r8
        real(r4),allocatable, dimension(:) :: tmparray_r4
        integer(i4),allocatable, dimension(:) :: tmparray_i4
        integer(i2),allocatable, dimension(:) :: tmparray_i2
        integer(i1),allocatable, dimension(:) :: tmparray_i1

        character(*), parameter :: function = "sort_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        if(this%nelements /= size(array)) then
            call wrerr(function // " tCompCat(_ext) type", s_err, "length of sorting array does not conform to length of type")
            status = s_errRep
        end if

        select type(array)
            type is ( real(r4) )
            tmparray_r4 = array
            call sort_real(tmparray_r4 ,status, order)
            type is ( real(r8) )
            tmparray_r8 = array
            call sort_double(tmparray_r8 ,status, order)
            type is ( integer(i4) )
            tmparray_i4 = array
            call sort_i4(tmparray_i4 ,status, order)
            type is ( integer(i2) )
            tmparray_i2 = array
            call sort_i2(tmparray_i2 ,status, order)
            type is ( integer(i1) )
            tmparray_i1 = array
            call sort_i1(tmparray_i1 ,status, order)
            class default
            call wrstop(function, "sorting tCompCat with the current array type is not supported for tCompCat")
        end select

        if(present(ascend)) then
            if(ascend == false) call reverse_array(order)
        end if


        call this%apply_index(order, status)

        call dealloc(order, status)
        call dealloc(tmparray_i1, status)
        call dealloc(tmparray_i2, status)
        call dealloc(tmparray_i4, status)
        call dealloc(tmparray_r4, status)
        call dealloc(tmparray_r8, status)

    end subroutine sort_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> add the new sources in a catalog to the existing comparison catalog
    !
    !
    !> @param[in] this
    !> @param[in] CatSrcList
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine addnew_tCompCat(this, CatSrcList, status)
        implicit none

        integer, intent(inout) :: status

        class(tCompCat) :: this
        type(tCatSrcList), intent(inout) :: CatSrcList

        integer :: ii, counter
        integer :: skyfield

        character(*), parameter :: function = "addnew_tCompCat"

        if(status > 100 .or. status == s_errRep) return

        if(CatSrcList%nelements == 0) return

        counter = this%nelements_used + 1
        skyfield = str2int(CatSrcList%stdhdr%skyfield, status)
        do ii = 1, CatSrcList%nelements
            if(CatSrcList%extend%new%value(ii) == true) then
                this%nsouuid = this%nsouuid + 1
                select case (CatSrcList%stdhdr%obs_mode)
                    case ("SURVEY")
                        write(this%souuid%value(counter), '(A1, I6.6, A1, I5.5)') "s",  skyfield, "_", this%nsouuid
                    case ("POINTING")
                        write(this%souuid%value(counter), '(A1, I6.6, A1, I5.5)') "p",  skyfield, "_", this%nsouuid
                    case default
                        call FError(function, s_err,  "obs_mode " // trim(CatSrcList%stdhdr%obs_mode) // " unknown", status)
                        return
                end select

                CatSrcList%extend%souuid%value(ii) = this%souuid%value(counter)
                this%detuid%value(counter) = CatSrcList%detuid%value(ii)
                this%iaunam%value(counter) = ""

                this%ra%value(counter) = CatSrcList%ra%value(ii)
                this%dec%value(counter) = CatSrcList%dec%value(ii)
                this%radec_err%value(counter) = CatSrcList%radec_err%value(ii)

                this%flux%value(counter) = CatSrcList%eb0%flux%value(ii)
                this%flux_err%value(counter) = CatSrcList%eb0%flux_err%value(ii)
                this%comment%value(counter) = ""

                this%new%value(counter) = true
!                this%active%value(counter) = false
                this%id_cluster%value(counter) = CatSrcList%id_cluster%value(ii)

                this%ext%value(counter) = CatSrcList%ext%value(ii)
                this%ext_err%value(counter) = CatSrcList%ext_err%value(ii)

                counter = counter + 1
            end if
        end do

        this%nelements_used = counter

    end subroutine addnew_tCompCat

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> update the existing comparison catalog with uniquely matched source
    !
    !
    !> @param[in] this
    !> @param[in] CatSrcList
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine update_tCompCat(this, CatSrcList, status)
        implicit none

        integer, intent(inout) :: status

        class(tCompCat) :: this
        type(tCatSrcList), intent(in) :: CatSrcList

        integer :: ii, counter

        character(*), parameter :: function = "update_tCompCat"

        if(status > 100 .or. status == s_errRep) return
        if(CatSrcList%nelements == 0) return

        do ii = 1, CatSrcList%nelements

            if(CatSrcList%id_match%value(ii) > 0) then

                counter = CatSrcList%id_match%value(ii)

                this%detuid%value(counter) = CatSrcList%detuid%value(ii)
                this%ra%value(counter) = CatSrcList%ra%value(ii)
                this%dec%value(counter) = CatSrcList%dec%value(ii)
                this%radec_err%value(counter) = CatSrcList%radec_err%value(ii)

                this%flux%value(counter) = CatSrcList%eb0%flux%value(ii)
                this%flux_err%value(counter) = CatSrcList%eb0%flux_err%value(ii)
                this%comment%value(counter) = ""

                this%new%value(counter) = false
!                this%active%value(counter) = true
                this%id_cluster%value(counter) = CatSrcList%id_cluster%value(ii)

                this%ext%value(counter) = CatSrcList%ext%value(ii)
                this%ext_err%value(counter) = CatSrcList%ext_err%value(ii)

            end if
        end do

    end subroutine update_tCompCat

end module catprep_compcat_mod
