! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_eb_mod
    use param
    use erolib_general_mod
    use erolib_ffits_col_mod
    use erolib_sort_mod
    use erolib_tbintable_mod
    implicit none

    type, public :: tEB

        integer :: nelements

        character(1) :: eband_str

        type(r4_tColumn) :: scts      !   [counts]           Source counts 2)
        type(r4_tColumn) :: scts_err  !   [counts]           1 sigma error 2)
        type(r4_tColumn) :: rate      !   [counts/sec]       Vignetting corrected source count rate 2)
        type(r4_tColumn) :: rate_err  !   [counts/sec]       1 sigma error 2)

        type(r4_tColumn) :: flux      !   [cgs]              Source flux 2)
        type(r4_tColumn) :: flux_err  !   [cgs]              1 sigma error 2)
        type(r4_tColumn) :: det_like  !                      Detection likelihood 1) 2)
        type(r4_tColumn) :: bg_map    !   [counts/pixel]     Background at source location 2)

        integer :: eband_zero

        type(r4_tColumn) :: exp_map   !   [sec]              Vignetting corrected exposure at source location 2)
        type(r4_tColumn) :: vig       !                      Average vignetting factor at source location 2)
        type(r4_tColumn) :: eef

    contains

        procedure :: init => init_tEB
        procedure :: free => free_tEB
        procedure :: writeToFits => write_tEB
        procedure :: writeToFits_frow_lrow => write_frow_lrow_tEB
        generic ::  write => writeToFits, writeToFits_frow_lrow
        procedure :: readFromFits => read_tEB
        procedure :: readFromFits_frow_lrow => read_frow_lrow_tEB
        generic :: read => readFromFits, readFromFits_frow_lrow
        procedure :: apply_index => apply_index_tEB
        procedure :: insert_into_tBinTable => insert_into_tBinTable_tEB
        procedure :: set => set_tEB

    end type tEB

contains

      !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> allocate the memory for this type
    !
    !
    !> @param[inout] this
    !> @param[in]  nelements
    !> @param[in, optional] eband
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tEB(this, nelements, status, eband)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) ::  nelements

        integer, intent(in), optional :: eband

        class(tEB) :: this
        character(*), parameter :: function = "init_tEB"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = nelements

        this%eband_str = ""

        if(present(eband)) then

            ! write(this%eband_str, "(A1, I1)") "_", eband
            write(this%eband_str, "(I1)") eband

            call this%scts%init("ML_CTS_" // this%eband_str, nelements, status)
            call this%scts_err%init("ML_CTS_ERR_" // this%eband_str, nelements, status)
            call this%rate%init("ML_RATE_" // this%eband_str, nelements, status)
            call this%rate_err%init("ML_RATE_ERR_" // this%eband_str, nelements, status)

            call this%flux%init("ML_FLUX_" // this%eband_str, nelements, status)
            call this%flux_err%init("ML_FLUX_ERR_" // this%eband_str, nelements, status)
            call this%det_like%init("DET_LIKE_" // this%eband_str, nelements, status)
            call this%bg_map%init("ML_BKG_" // this%eband_str, nelements, status)

            if(eband == 0) then
                this%eband_zero = true
            else
                this%eband_zero = false
                call this%exp_map%init("ML_EXP_" // this%eband_str, nelements, status)
                call this%eef%init("ML_EEF_" // this%eband_str, nelements, status)
                call this%vig%init("VIGNET_" // this%eband_str, nelements, status)
            end if

        else

            call this%scts%init("ML_CTS", nelements, status)
            call this%scts_err%init("ML_CTS_ERR", nelements, status)
            call this%rate%init("ML_RATE", nelements, status)
            call this%rate_err%init("ML_RATE_ERR", nelements, status)

            call this%flux%init("ML_FLUX", nelements, status)
            call this%flux_err%init("ML_FLUX_ERR", nelements, status)
            call this%det_like%init("DET_LIKE", nelements, status)
            call this%bg_map%init("ML_BKG", nelements, status)

            this%eband_zero = false

            call this%exp_map%init("ML_EXP", nelements, status)
            call this%eef%init("ML_EEF", nelements, status)
            call this%vig%init("VIGNET", nelements, status)

        end if

    end subroutine init_tEB

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> free the tEB
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tEB(this, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, jj

        class(tEB) :: this
        character(*), parameter :: function = "free_tEB"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = 0
        call this%scts%free(status)
        call this%scts_err%free(status)
        call this%rate%free(status)
        call this%rate_err%free(status)

        call this%flux%free(status)
        call this%flux_err%free(status)
        call this%det_like%free(status)
        call this%bg_map%free(status)
        call this%exp_map%free(status)

        call this%eef%free(status)
        call this%vig%free(status)

    end subroutine free_tEB

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> write a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[in, optional] extname
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tEB(this, unit, status, extname)
        implicit none

        integer, intent(inout) :: status
        class(tEB) :: this
        integer :: ii, jj
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), intent(in), optional :: extname

        character(*), parameter :: function = "write_tEB"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        if(present(extname)) call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%scts%write(unit, status, frow=frow, lrow=lrow)
            call this%scts_err%write(unit, status, frow=frow, lrow=lrow)
            call this%rate%write(unit, status, frow=frow, lrow=lrow)
            call this%rate_err%write(unit, status, frow=frow, lrow=lrow)

            call this%flux%write(unit, status, frow=frow, lrow=lrow)
            call this%flux_err%write(unit, status, frow=frow, lrow=lrow)
            call this%det_like%write(unit, status, frow=frow, lrow=lrow)
            call this%bg_map%write(unit, status, frow=frow, lrow=lrow)

            if(this%eband_zero == false) then
                call this%exp_map%write(unit, status, frow=frow, lrow=lrow)
                call this%eef%write(unit, status, frow=frow, lrow=lrow)
                call this%vig%write(unit, status, frow=frow, lrow=lrow)
            end if

        end do

    end subroutine write_tEB

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> read a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_tEB(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        class(tEB) :: this
        integer :: ii
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), parameter :: function = "read_tEB"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%scts%read(unit, status, frow=frow, lrow=lrow)
            call this%scts_err%read(unit, status, frow=frow, lrow=lrow)
            call this%rate%read(unit, status, frow=frow, lrow=lrow)
            call this%rate_err%read(unit, status, frow=frow, lrow=lrow)

            call this%flux%read(unit, status, frow=frow, lrow=lrow)
            call this%flux_err%read(unit, status, frow=frow, lrow=lrow)
            call this%det_like%read(unit, status, frow=frow, lrow=lrow)
            call this%bg_map%read(unit, status, frow=frow, lrow=lrow)

            if(this%eband_zero == false) then
                call this%exp_map%read(unit, status, frow=frow, lrow=lrow)
                call this%eef%read(unit, status, frow=frow, lrow=lrow)
                call this%vig%read(unit, status, frow=frow, lrow=lrow)
            end if

        end do

    end subroutine read_tEB

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> write a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_frow_lrow_tEB(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tEB) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "write_tEB"

        if(status > 100 .or. status == s_errRep) return

        call this%scts%write(unit, status, frow=frow, lrow=lrow)
        call this%scts_err%write(unit, status, frow=frow, lrow=lrow)
        call this%rate%write(unit, status, frow=frow, lrow=lrow)
        call this%rate_err%write(unit, status, frow=frow, lrow=lrow)

        call this%flux%write(unit, status, frow=frow, lrow=lrow)
        call this%flux_err%write(unit, status, frow=frow, lrow=lrow)
        call this%det_like%write(unit, status, frow=frow, lrow=lrow)
        call this%bg_map%write(unit, status, frow=frow, lrow=lrow)

        if(this%eband_zero == false) then
            call this%exp_map%write(unit, status, frow=frow, lrow=lrow)
            call this%eef%write(unit, status, frow=frow, lrow=lrow)
            call this%vig%write(unit, status, frow=frow, lrow=lrow)
        end if

    end subroutine write_frow_lrow_tEB

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> read a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_frow_lrow_tEB(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tEB) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "read_tEB"

        if(status > 100 .or. status == s_errRep) return

        call this%scts%read(unit, status, frow=frow, lrow=lrow)
        call this%scts_err%read(unit, status, frow=frow, lrow=lrow)
        call this%rate%read(unit, status, frow=frow, lrow=lrow)
        call this%rate_err%read(unit, status, frow=frow, lrow=lrow)

        call this%flux%read(unit, status, frow=frow, lrow=lrow)
        call this%flux_err%read(unit, status, frow=frow, lrow=lrow)
        call this%det_like%read(unit, status, frow=frow, lrow=lrow)
        call this%bg_map%read(unit, status, frow=frow, lrow=lrow)

        if(this%eband_zero == false) then
            call this%exp_map%read(unit, status, frow=frow, lrow=lrow)
            call this%eef%read(unit, status, frow=frow, lrow=lrow)
            call this%vig%read(unit, status, frow=frow, lrow=lrow)
        end if

    end subroutine read_frow_lrow_tEB

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_tEB(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tEB) :: this

        integer, dimension(:), intent(in) :: order

        character(*), parameter :: function = "apply_index_tEB"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = size(order)
        call this%scts%apply_index(order, status)
        call this%scts_err%apply_index(order, status)
        call this%rate%apply_index(order, status)
        call this%rate_err%apply_index(order, status)

        call this%flux%apply_index(order, status)
        call this%flux_err%apply_index(order, status)
        call this%det_like%apply_index(order, status)
        call this%bg_map%apply_index(order, status)

        if(this%eband_zero == false) then
            call this%exp_map%apply_index(order, status)
            call this%eef%apply_index(order, status)
            call this%vig%apply_index(order, status)
        end if

    end subroutine apply_index_tEB

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> insert the contents into a tbintable description
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout] table
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine insert_into_tBinTable_tEB(this, table, status)
        implicit none

        integer, intent(inout) :: status

        type(tBinTable), intent(inout) :: table

        integer :: ii, jj

        class(tEB) :: this
        character(*), parameter :: function = "insert_into_tBinTable_tEB"

        if(status > 100 .or. status == s_errRep) return

        call table%insert_ffits_col(this%scts, status)
        call table%insert_ffits_col(this%scts_err, status)
        call table%insert_ffits_col(this%rate, status)
        call table%insert_ffits_col(this%rate_err, status)

        call table%insert_ffits_col(this%flux, status)
        call table%insert_ffits_col(this%flux_err, status)
        call table%insert_ffits_col(this%det_like, status)
        call table%insert_ffits_col(this%bg_map, status)

        if(this%eband_zero == false) then
            call table%insert_ffits_col(this%exp_map, status)
            call table%insert_ffits_col(this%eef, status)
            call table%insert_ffits_col(this%vig, status)
        end if

    end subroutine insert_into_tBinTable_tEB

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tEB
    !> set the values of an order index to the provided values
    !
    !
    !> @param[inout] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine set_tEB(this, order, scts, scts_err, rate, rate_err, flux, flux_err, det_like, bg_map, status, exp_map, vig, eef)
        implicit none

        integer, intent(inout) :: status
        class(tEB) :: this
        integer, intent(in) :: order

        real(r4), intent(in) :: scts
        real(r4), intent(in) :: scts_err
        real(r4), intent(in) :: rate
        real(r4), intent(in) :: rate_err
        real(r4), intent(in) :: flux
        real(r4), intent(in) :: flux_err
        real(r4), intent(in) :: det_like
        real(r4), intent(in) :: bg_map
        real(r4), intent(in), optional :: exp_map
        real(r4), intent(in), optional :: vig
        real(r4), intent(in), optional :: eef

        this%scts%value(order) = scts
        this%scts_err%value(order) = scts_err
        this%rate%value(order) = rate
        this%rate_err%value(order) = rate_err
        this%flux%value(order) = flux
        this%flux_err%value(order) = flux_err
        this%det_like%value(order) = det_like
        this%bg_map%value(order) = bg_map

        if(present(exp_map)) this%exp_map%value(order) = exp_map
        if(present(vig)) this%vig%value(order) = vig
        if(present(eef)) this%eef%value(order) = eef

    end subroutine set_tEB

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge all columns of two tEB types
    !
    !
    !> @param[in] EB_A, EB_B
    !> @param[in] EB_C
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_EB(EB_A, EB_B, EB_C, status)
        implicit none

        integer, intent(inout) :: status

        class(tEB), intent(in) :: EB_A
        class(tEB), intent(in) :: EB_B
        class(tEB), intent(out) :: EB_C

        character(*), parameter :: function = "merge_EB"

        if(status > 100 .or. status == s_errRep) return

        if(EB_A%nelements == 0 .and. EB_B%nelements == 0) then
            call wrlog(function, "cats to merge have zero elements - nothing to be done here")
            return
        end if

        EB_C%nelements = EB_A%nelements + EB_B%nelements

        if(EB_A%eband_str /= EB_B%eband_str) then
            call wrstop(function, "cannot merge energy band types with a different band")
        else
            EB_C%eband_str = EB_A%eband_str
        end if

        call merge_tColumn(EB_A%scts, EB_B%scts, EB_C%scts, status)
        call merge_tColumn(EB_A%scts_err, EB_B%scts_err, EB_C%scts_err, status)
        call merge_tColumn(EB_A%rate, EB_B%rate, EB_C%rate, status)
        call merge_tColumn(EB_A%rate_err, EB_B%rate_err, EB_C%rate_err, status)

        call merge_tColumn(EB_A%flux, EB_B%flux, EB_C%flux, status)
        call merge_tColumn(EB_A%flux_err, EB_B%flux_err, EB_C%flux_err, status)
        call merge_tColumn(EB_A%det_like, EB_B%det_like, EB_C%det_like, status)
        call merge_tColumn(EB_A%bg_map, EB_B%bg_map, EB_C%bg_map, status)

        if((EB_A%eband_zero == false) .and. (EB_B%eband_zero == false)) then
            call merge_tColumn(EB_A%exp_map, EB_B%exp_map, EB_C%exp_map, status)
            call merge_tColumn(EB_A%vig, EB_B%vig, EB_C%vig, status)
            call merge_tColumn(EB_A%eef, EB_B%eef, EB_C%eef, status)
            EB_C%eband_zero = false
        else
            EB_C%eband_zero = true
        end if

    end subroutine merge_EB

end module catprep_eb_mod
