! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_match_cats_mod
    use param
    use catprep_srclist_mod
    use catprep_compcat_mod
    use catprep_skymap_mod
    use erolib_ffits_mod
    use erolib_general_mod
    use erolib_str_mod
    use erolib_sort_mod
    use erolib_math_mod
    implicit none

    type tQMatch_r4
        ! n_ / m_ are the prefixes for information to the n X m Matrix
        integer :: n_nelements, m_nelements
        integer, dimension(:), allocatable :: n_order, m_order
        real(r4), dimension(:, :), allocatable :: value
        type(tCatSrcList) :: Cat_N
        type(tCompCat) :: Cat_M

    contains
        procedure :: init => init_tQMatch_r4
        procedure :: free => free_tQMatch_r4
        procedure :: match_distance => match_distance_tQMatch_r4 ! calculate values of how likely a match between two sources is based on the distance and apply a cut to sources which are too far apart
        procedure :: determine_sc => determine_sc_tQMatch_r4 ! determine the source confusion for all n source with the m comparison sources based on the predetermined values
        procedure :: sort_sc => sort_sc_tQMatch_r4
        procedure :: sort_best => sort_best_tQMatch_r4
         ! return the id(s) of comparison sources which have been uniquely matched and are none zero (own skyfield but comparison catalog)
         ! this routine is used find the source which need to be updated in the comparison catalog file
        procedure :: get_unique_match_id => get_unique_match_id_tQMatch_r4
    end type tQMatch_r4


contains


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> routine to init a tQMatch_r4 type
    !
    !
    !
    !> @param[inout] this
    !> @param[in] nelements, melements
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tQMatch_r4(this, Cat_N, Cat_M, status)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this
        type(tCatSrcList) :: Cat_N
        type(tCompCat) :: Cat_M

        integer :: ii

        character(*), parameter :: function = "init_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        this%Cat_N = Cat_N
        this%Cat_M = Cat_M

        this%n_nelements = Cat_N%nelements
        this%m_nelements = Cat_M%nelements

        call alloc(this%n_order, this%n_nelements, status)
        call alloc(this%m_order, this%m_nelements, status)
        call alloc(this%value, this%Cat_N%nelements, this%Cat_M%nelements, status)

        do ii = 1, this%n_nelements
            this%n_order(ii) = ii
        end do

        do ii = 1, this%m_nelements
            this%m_order(ii) = ii
        end do

    end subroutine init_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> routine to free a tQMatch_r4 type
    !
    !
    !
    !> @param[inout] this
    !> @param[in] nelements, melements
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tQMatch_r4(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        character(*), parameter :: function = "free_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        this%n_nelements = 0
        this%m_nelements = 0
        call dealloc(this%n_order, status)
        call dealloc(this%m_order, status)
        call dealloc(this%value, status)

        call this%Cat_N%free(status)
        call this%Cat_M%free(status)

    end subroutine free_tQMatch_r4

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> calculate the match matrix based on distances tQMatch_r4 type
    !
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine match_distance_tQMatch_r4(this, status, nsigma)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        integer :: ii

        ! potential cut on the distance in units of sigma
        real(r8), optional, intent(in) :: nsigma

        character(*), parameter :: function = "match_distance_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        if(this%Cat_N%nelements == 0 ) then
            call wrerr(function, warn, "the input catalog has zero elements no match possible")
            return
        end if

        if(this%Cat_M%nelements == 0 ) then
            call wrerr(function, warn, "the comparison catalog has zero elements no matches can be determined")
            return
        end if

        ! 1. calculate distances between sources
        do ii = 1, this%n_nelements
            this%value(ii, :) = distance_gc_multi(this%Cat_N%ra%value(ii), this%Cat_N%dec%value(ii), this%Cat_M%ra%value, this%Cat_M%dec%value, true)
            ! convert distance into arcseconds
            this%value(ii, :) = this%value(ii, :)*3600.0_r8
            this%value(ii, :) = this%value(ii, :) / sqrt(this%Cat_N%radec_err%value(ii)**2 + this%Cat_M%radec_err%value**2)
        end do

        ! apply sigma cut
        if(present(nsigma)) then
            where (this%value > nsigma)
                this%value = 0
            elsewhere
                this%value = exp(-this%value/2.0)
            end where
        end if


    end subroutine match_distance_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> calculate the source confusion number for each input source N in cat_N of the tQMatch_r4 type
    !> flag new sources and set ids for uniquely matched sources
    !
    !
    !> @param[inout] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine determine_sc_tQMatch_r4(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        integer :: ii, jj
        integer :: sc_m, m_nelements
        integer, dimension(this%Cat_N%nelements, this%Cat_M%nelements) :: sc_nm

        real(r8), dimension(this%Cat_M%nelements) :: value
        integer, allocatable, dimension(:) :: order

        character(*), parameter :: function = "determine_sc_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        if(this%Cat_N%nelements == 0 ) then
            call wrerr(function, warn, "the input catalog has zero elements no match possible")
            return
        end if

        if(this%Cat_M%nelements == 0 ) then
            call wrerr(function, warn, "the comparison catalog has zero elements treat all found sources as new")
            do ii = 1, this%Cat_N%nelements
                if(this%Cat_N%region%value(ii) == true) then
                    this%Cat_N%extend%new%value(ii) = true
                    this%Cat_N%extend%unique%value(ii) = true
                end if
            end do
            return
        end if

        ! create a table with found matches set to 1

        sc_nm(:, :) = 0
        where (this%value > 0)
            sc_nm = 1
        end where

        ! determine for each source in the comparison catalog the number of potential source identifications
        ! set the elements in the SC matrix to this number
        ! all matches of any input source with the source M will be set to the determined SC
        !
        !   0101    0201
        !   0100  > 0200
        !   1000    1000
        do jj = 1, this%m_nelements
            sc_m = sum(sc_nm(:, jj))
            sc_nm(:,jj) = sc_nm(:,jj) * sc_m
        end do

        ! determine the total SC number for any given input source
        ! by summing over the SCm
        ! 0201      3
        ! 0200 > SC 2
        ! 1000      1
        do ii = 1, this%n_nelements
            this%Cat_N%extend%sc%value(ii) = sum(sc_nm(ii, :))

            ! if a source is new the sum equals zero and the source must be inside the region
            if(this%Cat_N%extend%sc%value(ii) == 0 .and. this%Cat_N%region%value(ii) == true) then
                this%Cat_N%extend%new%value(ii) = true
                this%Cat_N%extend%unique%value(ii) = true
                cycle
            end if

            ! if a source is uniquely identified sc must be 1
            ! make a reference of the id value which is only non zero if the source was identified within the same skyfield
            if(this%Cat_N%extend%sc%value(ii) == 1) then
                order = where_i4(sc_nm(ii, :), [1], status)
                this%Cat_N%id_match%value(ii) = this%Cat_M%id%value(order(1))
                this%Cat_N%extend%souuid%value(ii) = this%Cat_M%souuid%value(order(1))
                this%Cat_N%extend%iaunam%value(ii) = this%Cat_M%iaunam%value(order(1))
                ! if the identified source is from the "same" skyfield the id match is > 0 and thus source is identified as unique
                if(this%Cat_N%id_match%value(ii) > 0) then
                    this%Cat_N%extend%unique%value(ii) = true
                end if
            end if

            ! for each source sort the likelyhood for a source match in descending order
            value = this%value(ii, :)
            call sort(value, status, order)
            ! sort is ascending therefore reverse it to descending
            call reverse_array(order)

            ! note the best matches in SOUUID and likelyhood
            jj = order(1)
            this%Cat_N%extend%sc1%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc1%value(ii) = this%value(ii, jj)
            ! go to next source if the "input" source has only one partner
            if(count(sc_nm(ii,:) >= 1) == 1) cycle

            jj = order(2)
            this%Cat_N%extend%sc2%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc2%value(ii) = this%value(ii, jj)
            if(count(sc_nm(ii,:) >= 1) == 2) cycle

            jj = order(3)
            this%Cat_N%extend%sc3%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc3%value(ii) = this%value(ii, jj)
            if(count(sc_nm(ii,:) >= 1) == 3) cycle

            jj = order(4)
            this%Cat_N%extend%sc4%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc4%value(ii) = this%value(ii, jj)
            if(count(sc_nm(ii,:) >= 1) == 4) cycle

            jj = order(5)
            this%Cat_N%extend%sc5%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc5%value(ii) = this%value(ii, jj)
            if(count(sc_nm(ii,:) >= 1) == 5) cycle

            jj = order(6)
            this%Cat_N%extend%sc6%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc6%value(ii) = this%value(ii, jj)
            if(count(sc_nm(ii,:) >= 1) == 6) cycle

            jj = order(7)
            this%Cat_N%extend%sc7%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc7%value(ii) = this%value(ii, jj)
            if(count(sc_nm(ii,:) >= 1) == 7) cycle

            jj = order(8)
            this%Cat_N%extend%sc8%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc8%value(ii) = this%value(ii, jj)
            if(count(sc_nm(ii,:) >= 1) == 8) cycle

            jj = order(9)
            this%Cat_N%extend%sc9%value(ii) = this%Cat_M%souuid%value(jj)
            this%Cat_N%extend%qsc9%value(ii) = this%value(ii, jj)

        end do

    end subroutine determine_sc_tQMatch_r4

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> sort matrix to have the SC number in ascending order
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_sc_tQMatch_r4(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        integer(i4), dimension(this%n_nelements) :: sc
        integer(i4), allocatable, dimension(:) :: order

        character(*), parameter :: function = "sort_sc_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return
        if(this%Cat_N%nelements == 0) return

        ! first sort in ascending order of the source confusion number

        sc = this%Cat_N%extend%sc%value
        call sort(sc, status, order)
        this%n_order = this%n_order(order)
        this%value = this%value(this%n_order, :)
        call this%Cat_N%apply_index(order, status)

        call dealloc(order, status)

    end subroutine sort_sc_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> important: the structure must be first sorted in SC to locate the "starting" sort position
    !> important: this only applies for the case that the start row (and column) are unkown
    !> sort matrix that each source in Cat_N and Cat_M are only "used" once and the best value (max) is written on the diagonal elements
    !> the sorting only uses input sources which have more than one source "partner"
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_best_tQMatch_r4(this, status, startrow, startcol)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        integer :: ii, jj, row, col
        integer, intent(in), optional :: startrow, startcol
        integer, dimension(1) :: start
        integer, dimension(2) :: pos_best
        integer(i4), dimension(this%n_nelements) :: sc

        character(*), parameter :: function = "sort_best_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        ! return the first row for which more than one source was found in the comparison catalog
        sc = this%Cat_N%extend%sc%value
        start = minloc(sc, mask = sc .GT. 1)

        if(start(1) < 1) then
            call wrlog(function, "sort best not possible: since no SC > 1 was found")
            return
        end if

        jj = 1

        if(present(startrow)) then
            start(1) = startrow
        end if

        if(present(startcol)) then
            jj = startcol
        end if

        ! -----------------
        ! sort the (sub) matrix
        ! -----------------
        do ii = start(1), this%n_nelements
            pos_best = maxloc(this%value(ii: , jj:))
            row = ii + pos_best(1) - 1
            col = jj + pos_best(2) - 1
            ! swap row
            call swap(this%value(ii , :), this%value(row ,:))
            call swap(this%n_order(ii), this%n_order(row))
            ! swap col
            call swap(this%value(:, jj), this%value(:, col))
            call swap(this%m_order(jj), this%m_order(col))
            jj = jj + 1
            if(jj > this%m_nelements) then
                call wrlog(function, "the number of input sources exceeds the number of sources in the comparison catalog")
                exit
            end if
        end do
        ! -------------------------

        ! apply the changes in order to the catalog lists
        call this%Cat_N%apply_index(this%n_order, status)
        call this%Cat_M%apply_index(this%m_order, status)

    end subroutine sort_best_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return the ids of unique matches (comparison catsrclist of own skyfield)
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine get_unique_match_id_tQMatch_r4(this, id, status)
        implicit none

        integer, intent(inout) :: status
        integer :: n
        integer, allocatable, dimension(:), intent(out) :: id
        class(tQMatch_r4) :: this

        character(*), parameter :: function = "get_unique_match_id_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        n = count(this%Cat_N%extend%sc%value == 1)

    end subroutine get_unique_match_id_tQMatch_r4

end module catprep_match_cats_mod
