! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_mllist_mod

  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr
  use detect_io, only : fget_mllist, mllist_type
  use catprep_srclist_mod
  use erolib_sort_mod
  implicit none

contains

  ! subroutines and functions dealing with the ermldet output of G.L. and using them with the srclist type

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the mllist file and insert the values into the srclist type
  !
  !
  !
  !> @param[in] filename, detuidname
  !> @param[out] catsrclist
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine insert_mllist_catsrclist(filename, det_alg_str, detuid_prefix, catsrclist, status)
    implicit none

    character(*),      intent(in)    :: filename, det_alg_str, detuid_prefix
    type(tCatSrcList), intent(inout) :: catsrclist
    integer,           intent(inout) :: status

    type(mllist_type) :: mllist
    character(buffsize) :: rfilename,sass_parint
    character(7) :: msg
    integer :: max_prefix_size

    integer(i4), dimension(:), allocatable :: src_id_order ! positions in the mllist array for tel id, band id, src id (UNIQUE)
    integer(i4), dimension(:), allocatable :: src_id ! value of the order position in the mllist array for tel id, band id, src id (UNIQUE VALUES)
    integer(i4), dimension(:), allocatable :: id_array ! positions in the mllist array for tel id, band id and combined (ALL INDICES for ONE UNIQUE VALUE)

    integer(i4), dimension(:), allocatable :: order, order_col

    integer :: ii, jj ! for loop variables
    integer :: ldigit, rdigit, result, tel_id, band_id

    integer :: iounit, hdutype

    real(r8) :: glat, glon

    character(*), parameter :: function = "insert_mllist_catsrclist"

    if(status > 100 .or. status == s_errRep) return

    call fget_mllist(filename, mllist, status)

    ! manually read skyfield from mllist file
    call ffits_get_iounit(iounit, status)
    call ffits_open_file(iounit, filename, readonly, status)
    !        call ffits_movabs_hdu(iounit, 2, hdutype, status)
    call ffits_movnam_hdu(iounit, BINARY_TBL, "SRCLIST", status)

    call fread_stdhdr(iounit, catsrclist%stdhdr, status)

    call ffits_close_file(iounit, status)
    call ffits_free_iounit(iounit, status)


    if(size(mllist%data%id_src) == 0) then
      call wrerr(function, warn, "no entries in " // trim(filename) // " found - exit without changes")
      call catsrclist%init(0, status)
      return
    end if

    ! get the index of the unique values
    src_id_order = unique(mllist%data%id_src, status)

    ! all unique src_id values
    src_id = mllist%data(src_id_order)%id_src

    ! sort the unique id_src and get an index array of the positions
    call sort(src_id, status, order)

    ! change the order array of the original index array with the sorted index array
    src_id_order = src_id_order(order) ! now the src_id_order is sorted with ascending src id values (which are also unique)
    ! the src_id will also be changed to ascending (just in case it is not)
    src_id = src_id(order)

    if(size(src_id_order) < 1) then
      call wrstop(function, "the number of input sources is zero or less")
    end if

    ! set and allocate a new extended catsrclist
    call catsrclist%init(size(src_id_order), status)

    catsrclist%id_src%value = mllist%data(src_id_order)%id_src
    catsrclist%id_cluster%value = mllist%data(src_id_order)%id_cluster
    catsrclist%ra%value = mllist%data(src_id_order)%ra
    catsrclist%dec%value = mllist%data(src_id_order)%dec
    catsrclist%radec_err%value = mllist%data(src_id_order)%radec_err

    !   ! TODO determine sourceowner
    catsrclist%owner%value = mllist%data(src_id_order)%owner

    ! souuid and detuid are derived values based on the filename of the ermldet output file being processed
    !         call wrlog(function, "determine detuid differently - not based on filename")

    ! split the input filename of any "/" indicating relative paths
        
    call get_environment_variable("SASS_PARINT", sass_parint)

    if(sass_parint == "PIPELINE") then
      result = index(filename, "/", .true.)
      rfilename = filename(result+1:)
      do ii = 1, size(src_id_order)
        rdigit = len_trim(rfilename) - 5 ! due to .fits
        ldigit = rdigit - 8 ! last nine digits
        write(catsrclist%detuid%value(ii),'(A16,A2,I5.5,A9)') rfilename(1:16)&
          , det_alg_str, mllist%data(src_id_order(ii))%id_src, rfilename(ldigit:rdigit)
      end do
    else
      max_prefix_size = catsrclist%detuid_size - 7
      if(len_trim(detuid_prefix)>max_prefix_size) then
        do ii = 1, size(src_id_order)
          write(msg,'(A2,I5.5)') det_alg_str, mllist%data(src_id_order(ii))%id_src
          catsrclist%detuid%value(ii) = detuid_prefix(1:max_prefix_size)//msg
        end do
      else
        do ii = 1, size(src_id_order)
          write(msg,'(A2,I5.5)') det_alg_str, mllist%data(src_id_order(ii))%id_src
          catsrclist%detuid%value(ii) = trim(detuid_prefix)//msg
        end do
      end if
    end if

    order_col = where_str(mllist%cname, ["ID_SRC"], status)
    catsrclist%id_src%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ID_CLUSTER"], status)
    catsrclist%id_cluster%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["RA"], status)
    catsrclist%ra%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["DEC"], status)
    catsrclist%dec%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["RADEC_ERR"], status)
    catsrclist%radec_err%tunit = mllist%cunit(order_col(1))

    order_col = where_str(mllist%cname, ["LII"], status)
    catsrclist%lii%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["BII"], status)
    catsrclist%bii%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["DIST_NN"], status)
    catsrclist%dist_nn%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["EXT"], status)
    catsrclist%ext%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["EXT_ERR"], status)
    catsrclist%ext_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["EXT_LIKE"], status)
    catsrclist%ext_like%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RADIUS"], status)
    catsrclist%cutrad%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["MASKFRAC"], status)
    catsrclist%maskfrac%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["RAWX"], status)
    catsrclist%rawx%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["RAWY"], status)
    catsrclist%rawy%tunit = mllist%cunit(order_col(1))

    order_col = where_str(mllist%cname, ["HR1"], status)
    catsrclist%hr_1%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["HR2"], status)
    catsrclist%hr_2%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["HR3"], status)
    catsrclist%hr_3%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["HR1_ERR"], status)
    catsrclist%hr_1_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["HR2_ERR"], status)
    catsrclist%hr_2_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["HR3_ERR"], status)
    catsrclist%hr_3_err%tunit = mllist%cunit(order_col(1))

    order_col = where_str(mllist%cname, ["APE_CTS"], status)
    catsrclist%ape%cts%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["APE_BKG"], status)
    catsrclist%ape%bkg%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["APE_EXP"], status)
    catsrclist%ape%exp%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["APE_EEF"], status)
    catsrclist%ape%eef%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["APE_RADIUS"], status)
    catsrclist%ape%radius%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["APE_POIS"], status)
    catsrclist%ape%pois%tunit = mllist%cunit(order_col(1))

    ! Band 0:
    order_col = where_str(mllist%cname, ["ML_CTS"], status)
    catsrclist%eb0%scts%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_CTS_ERR"], status)
    catsrclist%eb0%scts_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE"], status)
    catsrclist%eb0%rate%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE_ERR"], status)
    catsrclist%eb0%rate_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX"], status)
    catsrclist%eb0%flux%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX_ERR"], status)
    catsrclist%eb0%flux_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["DET_LIKE"], status)
    catsrclist%eb0%det_like%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_BKG"], status)
    catsrclist%eb0%bg_map%tunit   = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EXP"], status)
    catsrclist%eb0%exp_map%tunit  = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["VIGNET"], status)
    catsrclist%eb0%vig%tunit      = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EEF"], status)
    catsrclist%eb0%eef%tunit      = mllist%cunit(order_col(1))

    ! Band 1:
    order_col = where_str(mllist%cname, ["ML_CTS"], status)
    catsrclist%eb1%scts%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_CTS_ERR"], status)
    catsrclist%eb1%scts_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE"], status)
    catsrclist%eb1%rate%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE_ERR"], status)
    catsrclist%eb1%rate_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX"], status)
    catsrclist%eb1%flux%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX_ERR"], status)
    catsrclist%eb1%flux_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["DET_LIKE"], status)
    catsrclist%eb1%det_like%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_BKG"], status)
    catsrclist%eb1%bg_map%tunit   = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EXP"], status)
    catsrclist%eb1%exp_map%tunit  = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["VIGNET"], status)
    catsrclist%eb1%vig%tunit      = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EEF"], status)
    catsrclist%eb1%eef%tunit      = mllist%cunit(order_col(1))

    ! Band 2:
    order_col = where_str(mllist%cname, ["ML_CTS"], status)
    catsrclist%eb2%scts%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_CTS_ERR"], status)
    catsrclist%eb2%scts_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE"], status)
    catsrclist%eb2%rate%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE_ERR"], status)
    catsrclist%eb2%rate_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX"], status)
    catsrclist%eb2%flux%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX_ERR"], status)
    catsrclist%eb2%flux_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["DET_LIKE"], status)
    catsrclist%eb2%det_like%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_BKG"], status)
    catsrclist%eb2%bg_map%tunit   = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EXP"], status)
    catsrclist%eb2%exp_map%tunit  = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["VIGNET"], status)
    catsrclist%eb2%vig%tunit      = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EEF"], status)
    catsrclist%eb2%eef%tunit      = mllist%cunit(order_col(1))

    ! Band 3:
    order_col = where_str(mllist%cname, ["ML_CTS"], status)
    catsrclist%eb3%scts%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_CTS_ERR"], status)
    catsrclist%eb3%scts_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE"], status)
    catsrclist%eb3%rate%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE_ERR"], status)
    catsrclist%eb3%rate_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX"], status)
    catsrclist%eb3%flux%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX_ERR"], status)
    catsrclist%eb3%flux_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["DET_LIKE"], status)
    catsrclist%eb3%det_like%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_BKG"], status)
    catsrclist%eb3%bg_map%tunit   = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EXP"], status)
    catsrclist%eb3%exp_map%tunit  = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["VIGNET"], status)
    catsrclist%eb3%vig%tunit      = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EEF"], status)
    catsrclist%eb3%eef%tunit      = mllist%cunit(order_col(1))

    ! Band 4:
    order_col = where_str(mllist%cname, ["ML_CTS"], status)
    catsrclist%eb4%scts%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_CTS_ERR"], status)
    catsrclist%eb4%scts_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE"], status)
    catsrclist%eb4%rate%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_RATE_ERR"], status)
    catsrclist%eb4%rate_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX"], status)
    catsrclist%eb4%flux%tunit     = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_FLUX_ERR"], status)
    catsrclist%eb4%flux_err%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["DET_LIKE"], status)
    catsrclist%eb4%det_like%tunit = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_BKG"], status)
    catsrclist%eb4%bg_map%tunit   = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EXP"], status)
    catsrclist%eb4%exp_map%tunit  = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["VIGNET"], status)
    catsrclist%eb4%vig%tunit      = mllist%cunit(order_col(1))
    order_col = where_str(mllist%cname, ["ML_EEF"], status)
    catsrclist%eb4%eef%tunit      = mllist%cunit(order_col(1))


    ! all these values are always present for each combination of SRC ID , TEL ID and BAND ID

    catsrclist%lII%value = mllist%data(src_id_order)%lII
    catsrclist%bII%value = mllist%data(src_id_order)%bII

    catsrclist%dist_nn%value = mllist%data(src_id_order)%dist_nn
    catsrclist%ext%value = mllist%data(src_id_order)%ext
    catsrclist%ext_err%value = mllist%data(src_id_order)%ext_err
    catsrclist%ext_like%value = mllist%data(src_id_order)%ext_like
    catsrclist%cutrad%value = mllist%data(src_id_order)%cutrad
    catsrclist%maskfrac%value = mllist%data(src_id_order)%maskfrac

    catsrclist%rawx%value = mllist%data(src_id_order)%rawx
    catsrclist%rawy%value = mllist%data(src_id_order)%rawy

    ! as in detect_io_mod.f90 for the mllist type
    catsrclist%hr_1%value = mllist%data(src_id_order)%hr1
    catsrclist%hr_1_err%value = mllist%data(src_id_order)%hr1_err
    catsrclist%hr_2%value =  mllist%data(src_id_order)%hr2
    catsrclist%hr_2_err%value = mllist%data(src_id_order)%hr2_err
    catsrclist%hr_3%value = mllist%data(src_id_order)%hr3
    catsrclist%hr_3_err%value = mllist%data(src_id_order)%hr3_err

    ! ape columns
    catsrclist%ape%cts%value = mllist%data(src_id_order)%ape_cts
    catsrclist%ape%bkg%value = mllist%data(src_id_order)%ape_bkg
    catsrclist%ape%exp%value = mllist%data(src_id_order)%ape_exp
    catsrclist%ape%eef%value = mllist%data(src_id_order)%ape_eef
    catsrclist%ape%radius%value = mllist%data(src_id_order)%ape_radius
    catsrclist%ape%pois%value = mllist%data(src_id_order)%ape_pois



    ! step through all individual sources and fill the correct columns
    do ii = 1, size(src_id)
      order = where_i4(mllist%data%id_src, [src_id(ii)], status)

      if(size(order) < 1) then
        call wrstop(function, "the selected source was not found")
      end if

      do jj = 1, size(order)
        tel_id = mllist%data(order(jj))%id_inst
        band_id = mllist%data(order(jj))%id_band

        ! select tel_id
        select case (tel_id)
          case (0)
              ! 0 is ignored
          case (1)
            select case(band_id)
              case(0)
                call catsrclist%eb0%set(ii, mllist%data(order(jj))%scts, mllist%data(order(jj))%scts_err, &
                  mllist%data(order(jj))%rate, mllist%data(order(jj))%rate_err, &
                  mllist%data(order(jj))%flux, mllist%data(order(jj))%flux_err, &
                  mllist%data(order(jj))%like, mllist%data(order(jj))%bg_map, status)
              case(1)
                call catsrclist%eb1%set(ii, mllist%data(order(jj))%scts, mllist%data(order(jj))%scts_err, &
                  mllist%data(order(jj))%rate, mllist%data(order(jj))%rate_err, &
                  mllist%data(order(jj))%flux, mllist%data(order(jj))%flux_err, &
                  mllist%data(order(jj))%like, mllist%data(order(jj))%bg_map, &
                  status, mllist%data(order(jj))%exp_map, mllist%data(order(jj))%vignetting, &
                  mllist%data(order(jj))%eef)
              case(2)
                call catsrclist%eb2%set(ii, mllist%data(order(jj))%scts, mllist%data(order(jj))%scts_err, &
                  mllist%data(order(jj))%rate, mllist%data(order(jj))%rate_err, &
                  mllist%data(order(jj))%flux, mllist%data(order(jj))%flux_err, &
                  mllist%data(order(jj))%like, mllist%data(order(jj))%bg_map, &
                  status, mllist%data(order(jj))%exp_map, mllist%data(order(jj))%vignetting, &
                  mllist%data(order(jj))%eef)
              case(3)
                call catsrclist%eb3%set(ii, mllist%data(order(jj))%scts, mllist%data(order(jj))%scts_err, &
                  mllist%data(order(jj))%rate, mllist%data(order(jj))%rate_err, &
                  mllist%data(order(jj))%flux, mllist%data(order(jj))%flux_err, &
                  mllist%data(order(jj))%like, mllist%data(order(jj))%bg_map, &
                  status, mllist%data(order(jj))%exp_map, mllist%data(order(jj))%vignetting, &
                  mllist%data(order(jj))%eef)
              case(4)
                call catsrclist%eb4%set(ii, mllist%data(order(jj))%scts, mllist%data(order(jj))%scts_err, &
                  mllist%data(order(jj))%rate, mllist%data(order(jj))%rate_err, &
                  mllist%data(order(jj))%flux, mllist%data(order(jj))%flux_err, &
                  mllist%data(order(jj))%like, mllist%data(order(jj))%bg_map, &
                  status, mllist%data(order(jj))%exp_map, mllist%data(order(jj))%vignetting, &
                  mllist%data(order(jj))%eef)

            end select

          case default
            call wrstop (function, "different id_inst found (other than 0 / 1)")

        end select
      end do

    end do

    call dealloc(order, status)
    call dealloc(order_col, status)
    call dealloc(id_array, status)
    call dealloc(src_id, status)
    call dealloc(src_id_order, status)

  end subroutine insert_mllist_catsrclist

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> map id_src of mllist to the src_id  in catsrclist returning an array of indices mapping it
  !
  !
  !
  !> @param[in] id_src
  !> @param[out] src_id
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine map_id_src_2_src_id(id_src, src_id, mapping, status)
    implicit none

    integer, intent(inout) :: status

    integer(i4), dimension(:), intent(in) :: id_src, src_id
    integer(i4), dimension(size(id_src)), intent(out) :: mapping
    integer :: ii, jj

    character(*), parameter :: function = "map_id_src_2_src_id"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(id_src)
      do jj = 1, size(src_id)
        if(id_src(ii) == src_id(jj)) then
          mapping(ii) = jj
          exit
        end if
      end do
    end do

  end subroutine map_id_src_2_src_id


end module catprep_mllist_mod
