! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_skymap_mod
  use erolib_general_mod
  use erolib_ffits_mod
  use erolib_str_mod
  use erolib_sort_mod
  use erolib_region_mod
  use erolib_tbintable_mod
  implicit none

  integer, private :: max_fields = 9

  type tSkymap

    integer :: unit
    character(buffsize) :: filename
    character(buffsize) :: extname
    type(tBinTable) :: table ! description of the binary table

    integer(i4) :: nelements

    type(i4_tColumn) :: smapnr
    type(r8_tColumn) :: ra_min
    type(r8_tColumn) :: ra_max
    type(r8_tColumn) :: dec_min
    type(r8_tColumn) :: dec_max
    type(i2_tColumn) :: owner
    type(i2_tColumn) :: n_nbrs
    type(str_tColumn) :: dir ! directory of internal catalog
    type(str_tColumn) :: name ! filename of latest internal catalog
    type(str_tColumn) :: dir_lastsscan ! directory of last single scan
    type(str_tColumn) :: name_lastsscan ! filename of last single scan
    type(str_tColumn) :: dir_sscan ! directory of latest single scan
    type(str_tColumn) :: name_sscan ! filename of latest single scan
    type(r8_tColumn) :: tstart_sscan ! tstart of latest single scan

    type(i4_tColumn) :: field1
    type(i4_tColumn) :: field2
    type(i4_tColumn) :: field3
    type(i4_tColumn) :: field4
    type(i4_tColumn) :: field5
    type(i4_tColumn) :: field6
    type(i4_tColumn) :: field7
    type(i4_tColumn) :: field8
    type(i4_tColumn) :: field9

    type(i4_tColumn) :: field1_order
    type(i4_tColumn) :: field2_order
    type(i4_tColumn) :: field3_order
    type(i4_tColumn) :: field4_order
    type(i4_tColumn) :: field5_order
    type(i4_tColumn) :: field6_order
    type(i4_tColumn) :: field7_order
    type(i4_tColumn) :: field8_order
    type(i4_tColumn) :: field9_order

  contains
    procedure, non_overridable :: apply_index => apply_index_tSkymap
    procedure, non_overridable :: get_skyfield_boundary => get_skyfield_boundary_tSkymap
    procedure, non_overridable :: read => read_tSkymap
    procedure, non_overridable :: write => write_tSkymap
    procedure, non_overridable :: get_adjacent_skyfields => get_adjacent_skyfields_tSkymap
    procedure, non_overridable :: determine_skyfield => determine_skyfield_tSkymap
    procedure, non_overridable :: open => open_tSkymap
    procedure, non_overridable :: close => close_tSkymap
    procedure, non_overridable :: init => init_tSkymap
    procedure, non_overridable :: free => free_tSkymap
    procedure, non_overridable :: derive_order => derive_order_tSkymap
    procedure, non_overridable :: create_extension => create_extension_tSkymap
    procedure, non_overridable :: get_catalog_name => get_catalog_name_tSkymap
    procedure, non_overridable :: get_adjacent_catalog_name => get_adjacent_catalog_name_tSkymap
    procedure, non_overridable :: set_and_write_new_catalog_name => set_and_write_new_catalog_name_tSkymap
    procedure, non_overridable :: set_and_write_sscan => set_and_write_sscan_tSkymap
    procedure, non_overridable :: set_and_write_lastsscan => set_and_write_lastsscan_tSkymap
    procedure, non_overridable :: determine_skyfield_radec_box => determine_skyfield_radec_box_tSkymap
    procedure, non_overridable :: check_skyfield => check_skyfield_tSkymap

  end type tSkymap

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> determine the skyfield for a specific RA / DEC coordinate
  !
  !
  !
  !> @param[in] this, ra, dec
  !> @param[out] skyfield
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine determine_skyfield_tSkymap(this, ra, dec, skyfield, status, owner)
    implicit none

    integer, intent(inout) :: status
    real(r8), intent(in) :: ra, dec
    real(r8) :: ra_intern
    integer(i4), intent(out) :: skyfield
    class(tSkymap) :: this
    integer :: ii, counter

    integer, optional, intent(out) :: owner

    character(*), parameter :: function = "determine_skyfield_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    counter = 0

    ra_intern = mod(ra, 360.0_r8)

    if(ra < 0) ra_intern =  ra_intern + 360.0_r8
    if(dec < -90.0_r8 .or. dec > +90.0_r8) call wrstop(function, "central region position is outside coordinate system of FK5")

    do ii = 1, this%nelements
      if(this%ra_min%value(ii) < ra_intern) then
        if(this%ra_max%value(ii) > ra_intern) then
          if(this%dec_min%value(ii) < dec) then
            if(this%dec_max%value(ii) > dec) then
              skyfield = this%smapnr%value(ii)
              if(present(owner)) owner = int(this%owner%value(ii), 4)
              counter = counter + 1
            end if
          end if
        end if
      end if
    end do

    if(counter /= 1) then
      call wrstop(function, "could not determine skyfield - contact developers")
    end if

  end subroutine determine_skyfield_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> return the boundary for a specific skyfield
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] ra_min, ra_max, dec_min, dec_max
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_skyfield_boundary_tSkymap(this, skyfield, ra_min, ra_max, dec_min, dec_max, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this
    integer, intent(in) :: skyfield

    integer :: ii

    real(r8), intent(out) :: ra_min, ra_max, dec_min, dec_max

    character(*), parameter :: function = "get_skyfield_boundary_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! if 0 maps are found try reading in the map
    if(this%nelements == 0) call this%read(status)

    ! if there are still no skymaps report error
    if(this%nelements == 0) call wrstop(function, "no skyfields in provided skymaps file")

    do ii = 1, this%nelements
      if(this%smapnr%value(ii) == skyfield) then
        ra_min = this%ra_min%value(ii)
        ra_max = this%ra_max%value(ii)
        dec_min = this%dec_min%value(ii)
        dec_max = this%dec_max%value(ii)
        return
      end if
    end do

    call wrerr(function, s_err, "skyfield could not be found in skymaps file")
    status = s_errRep

  end subroutine get_skyfield_boundary_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the skyfields adjacent to skyfields
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_adjacent_skyfields_tSkymap(this, skyfield, skyfield_out, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    integer, dimension(:), intent(in) :: skyfield
    integer(i4), intent(out), allocatable, dimension(:) :: skyfield_out

    integer, dimension(:), allocatable :: order
    integer :: order_index, n_nbrs

    character(*), parameter :: function = "get_adjacent_skyfields_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! find the smapnr matching the skyfield nr
    order = where_i4(this%smapnr%value, skyfield, status)

    if(size(order) /= 1) then
      call wrstop(function, "could not find skyfield match in skymap file")
    end if

    order_index = order(1)
    n_nbrs = this%n_nbrs%value(order_index)
    ! allocate with the max number of skyfields
    call alloc(skyfield_out, n_nbrs, status)

    skyfield_out(1) = this%field1%value(order_index)
    skyfield_out(2) = this%field2%value(order_index)
    skyfield_out(3) = this%field3%value(order_index)
    skyfield_out(4) = this%field4%value(order_index)
    skyfield_out(5) = this%field5%value(order_index)
    skyfield_out(6) = this%field6%value(order_index)

    ! 6 skyfields are always adjacent - only special cases are 7,8,9 for adjacent skyfields

    if(this%n_nbrs%value(order(1)) > 6) then
      skyfield_out(7) = this%field7%value(order(1))
    end if

    if(this%n_nbrs%value(order(1)) > 7) then
      skyfield_out(8) = this%field8%value(order(1))
    end if

    if(this%n_nbrs%value(order(1))>8) then
      skyfield_out(9) = this%field9%value(order(1))
    end if

    call dealloc(order, status)

  end subroutine get_adjacent_skyfields_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tSkymap
  !> read the skymap data from a fits file
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tSkymap(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this
    integer :: ii, unit, derive_order
    integer :: frow, lrow, nelements, rowsize, nrows

    character(*), parameter :: function = "read_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    unit = this%unit

    call ffits_get_num_rows(this%unit, nelements, status)
    call ffits_get_rowsize(this%unit, rowsize, status)
    call this%init(nelements, status)

    derive_order = false

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      call this%smapnr%read(unit, status, frow=frow, lrow=lrow)
      call this%ra_min%read(unit, status, frow=frow, lrow=lrow)
      call this%ra_max%read(unit, status, frow=frow, lrow=lrow)
      call this%dec_min%read(unit, status, frow=frow, lrow=lrow)
      call this%dec_max%read(unit, status, frow=frow, lrow=lrow)
      call this%owner%read(unit, status, frow=frow, lrow=lrow)
      call this%n_nbrs%read(unit, status, frow=frow, lrow=lrow)

      call this%field1%read(unit, status, frow=frow, lrow=lrow)
      call this%field2%read(unit, status, frow=frow, lrow=lrow)
      call this%field3%read(unit, status, frow=frow, lrow=lrow)
      call this%field4%read(unit, status, frow=frow, lrow=lrow)
      call this%field5%read(unit, status, frow=frow, lrow=lrow)
      call this%field6%read(unit, status, frow=frow, lrow=lrow)
      call this%field7%read(unit, status, frow=frow, lrow=lrow)
      call this%field8%read(unit, status, frow=frow, lrow=lrow)
      call this%field9%read(unit, status, frow=frow, lrow=lrow)

      ! since the original skyfield file does not contain certain
      ! the code tries to read these

      status = ignore
      call ffits_get_colnum(unit, .false., this%dir%name, this%dir%num, status)
      if(status == ignore) then
        status = ok
        call this%dir%read(unit, status, frow=frow, lrow=lrow)
      end if

      status = ignore
      call ffits_get_colnum(unit, .false., this%name%name, this%name%num, status)
      if(status == ignore) then
        status = ok
        call this%name%read(unit, status, frow=frow, lrow=lrow)
      end if

      status = ignore
      call ffits_get_colnum(unit, .false., this%dir_lastsscan%name, this%dir_lastsscan%num, status)
      if(status == ignore) then
        status = ok
        call this%dir_lastsscan%read(unit, status, frow=frow, lrow=lrow)
      end if

      status = ignore
      call ffits_get_colnum(unit, .false., this%name_lastsscan%name, this%name_lastsscan%num, status)
      if(status == ignore) then
        status = ok
        call this%name_lastsscan%read(unit, status, frow=frow, lrow=lrow)
      end if

      status = ignore
      call ffits_get_colnum(unit, .false., this%dir_sscan%name, this%dir_sscan%num, status)
      if(status == ignore) then
        status = ok
        call this%dir_sscan%read(unit, status, frow=frow, lrow=lrow)
      end if

      status = ignore
      call ffits_get_colnum(unit, .false., this%name_sscan%name, this%name_sscan%num, status)
      if(status == ignore) then
        status = ok
        call this%name_sscan%read(unit, status, frow=frow, lrow=lrow)
      end if

      status = ignore
      call ffits_get_colnum(unit, .false., this%tstart_sscan%name, this%tstart_sscan%num, status)
      if(status == ignore) then
        status = ok
        call this%tstart_sscan%read(unit, status, frow=frow, lrow=lrow)
      end if

      ! derive fieldX_order in case the column does not exist
      status = ignore
      call ffits_get_colnum(unit, .false., this%field1_order%name, this%field1_order%num, status)
      if(status == ignore) then
        status = ok
        call this%field1_order%read(unit, status, frow=frow, lrow=lrow)
      end if

      if(status == ignore) then
        ! since the first field1_order column was found -
        ! assume that the file contains the other columns as well
        status = ok
        call this%field2_order%read(unit, status, frow=frow, lrow=lrow)
        call this%field3_order%read(unit, status, frow=frow, lrow=lrow)
        call this%field4_order%read(unit, status, frow=frow, lrow=lrow)
        call this%field5_order%read(unit, status, frow=frow, lrow=lrow)
        call this%field6_order%read(unit, status, frow=frow, lrow=lrow)
        call this%field7_order%read(unit, status, frow=frow, lrow=lrow)
        call this%field8_order%read(unit, status, frow=frow, lrow=lrow)
        call this%field9_order%read(unit, status, frow=frow, lrow=lrow)

      else
        ! derive the field_order
        status = ok
        derive_order = true
      end if
    end do

    if(derive_order == true) then
      call this%derive_order(status)
    end if

  end subroutine read_tSkymap

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> derive the fieldX_order
  !
  !
  !
  !> @param[inout] this
  !> @param[in]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine derive_order_tSkymap(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this
    integer(i4) :: ii
    integer(i4), allocatable, dimension(:) :: order
    integer(i4), dimension(this%nelements) :: n_nbrs, smapnr, field1, field2, field3, field4
    integer(i4), dimension(this%nelements) :: field5, field6, field7, field8, field9

    character(*), parameter :: function = "derive_order_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    n_nbrs = this%n_nbrs%value
    smapnr = this%smapnr%value
    field1 = this%field1%value
    field2 = this%field2%value
    field3 = this%field3%value
    field4 = this%field4%value
    field5 = this%field5%value
    field6 = this%field6%value
    field7 = this%field7%value
    field8 = this%field8%value
    field9 = this%field9%value

    do ii = 1, this%nelements
      order = where_i4(smapnr, [field1(ii)], status)
      if(size(order) == 1) then
        this%field1_order%value(ii) = order(1)
      else
        print *, " - 1 -"
        print *, [field1(ii)]
        call wrstop(function, "more than one matching skyfield found! - duplicate entries")
      end if

      order = where_i4(smapnr, [field2(ii)], status)
      if(size(order) == 1) then
        this%field2_order%value(ii) = order(1)
      else
        print *, " - 2 -"
        print *, [field2(ii)]
        call wrstop(function, "more than one matching skyfield found! - duplicate entries")
      end if

      order = where_i4(smapnr, [field3(ii)], status)
      if(size(order) == 1) then
        this%field3_order%value(ii) = order(1)
      else
        print *, " - 3 -"
        print *, [field3(ii)]
        call wrstop(function, "more than one matching skyfield found! - duplicate entries")
      end if

      order = where_i4(smapnr, [field4(ii)], status)
      if(size(order) == 1) then
        this%field4_order%value(ii) = order(1)
      else
        print *, " - 4 -"
        print *, [field4(ii)]
        call wrstop(function, "more than one matching skyfield found! - duplicate entries")
      end if

      order = where_i4(smapnr, [field5(ii)], status)
      if(size(order) == 1) then
        this%field5_order%value(ii) = order(1)
      else
        print *, " - 5 -"
        print *, [field5(ii)]
        call wrstop(function, "more than one matching skyfield found! - duplicate entries")
      end if

      order = where_i4(smapnr, [field6(ii)], status)
      if(size(order) == 1) then
        this%field6_order%value(ii) = order(1)
      else
        print *, " - 6 -"
        print *, [field6(ii)]
        call wrstop(function, "more than one matching skyfield found! - duplicate entries")
      end if

      if(n_nbrs(ii) > 6) then
        order = where_i4(smapnr, [field7(ii)], status)
        if(size(order) == 1) then
          this%field7_order%value(ii) = order(1)
        else
          print *, " - 7 -"
          print *, [field7(ii)]
          call wrstop(function, "more than one matching skyfield found! - duplicate entries")
        end if
      end if

      if(n_nbrs(ii) > 7) then
        order = where_i4(smapnr, [field8(ii)], status)
        if(size(order) == 1) then
          this%field8_order%value(ii) = order(1)
        else
          print *, " - 8 -"
          print *, [field8(ii)]
          print *, order
          print *, smapnr(848)
          call wrstop(function, "more than one matching skyfield found! - duplicate entries")
        end if
      end if

      if(n_nbrs(ii) > 8) then
        order = where_i4(smapnr, [field9(ii)], status)
        if(size(order) == 1) then
          this%field9_order%value(ii) = order(1)
        else
          print *, " - 9 -"
          print *, [field9(ii)]
          call wrstop(function, "more than one matching skyfield found! - duplicate entries")
        end if
      end if

    end do

    call dealloc(order, status)

  end subroutine derive_order_tSkymap

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tSkymap
  !> write data to FITS file
  !
  !
  !> @param[inout] this
  !> @param[in] unit
  !> @param[in] frow, lrow
  !> @param[in, optional] extname
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tSkymap(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this
    integer :: ii, jj
    integer :: unit
    integer :: frow, lrow, nelements, rowsize, nrows

    character(*), parameter :: function = "write_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! move to the correct extension

    nelements = this%nelements
    unit = this%unit

    status = ignore
    call ffits_movnam_hdu(unit, BINARY_TBL, this%extname, status)
    if(status /= ignore) then
      status = ok
      call this%create_extension(status)
    else
      status = ok
    end if

    call ffits_get_rowsize(unit, rowsize, status)

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      call this%smapnr%write(unit, status, frow=frow, lrow=lrow)
      call this%ra_min%write(unit, status, frow=frow, lrow=lrow)
      call this%ra_max%write(unit, status, frow=frow, lrow=lrow)
      call this%dec_min%write(unit, status, frow=frow, lrow=lrow)
      call this%dec_max%write(unit, status, frow=frow, lrow=lrow)
      call this%n_nbrs%write(unit, status, frow=frow, lrow=lrow)

      call this%dir%write(unit, status, frow=frow, lrow=lrow)
      call this%name%write(unit, status, frow=frow, lrow=lrow)
      call this%dir_lastsscan%write(unit, status, frow=frow, lrow=lrow)
      call this%name_lastsscan%write(unit, status, frow=frow, lrow=lrow)
      call this%dir_sscan%write(unit, status, frow=frow, lrow=lrow)
      call this%name_sscan%write(unit, status, frow=frow, lrow=lrow)
      call this%tstart_sscan%write(unit, status, frow=frow, lrow=lrow)

      call this%field1%write(unit, status, frow=frow, lrow=lrow)
      call this%field2%write(unit, status, frow=frow, lrow=lrow)
      call this%field3%write(unit, status, frow=frow, lrow=lrow)
      call this%field4%write(unit, status, frow=frow, lrow=lrow)
      call this%field5%write(unit, status, frow=frow, lrow=lrow)
      call this%field6%write(unit, status, frow=frow, lrow=lrow)
      call this%field7%write(unit, status, frow=frow, lrow=lrow)
      call this%field8%write(unit, status, frow=frow, lrow=lrow)
      call this%field9%write(unit, status, frow=frow, lrow=lrow)

      call this%field1_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field2_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field3_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field4_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field5_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field6_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field7_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field8_order%write(unit, status, frow=frow, lrow=lrow)
      call this%field9_order%write(unit, status, frow=frow, lrow=lrow)

    end do

  end subroutine write_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> open the skymap file
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tSkymap(this, filename, rwmode, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this
    character(*), intent(in) :: filename
    integer, intent(in) :: rwmode

    character(*), parameter :: function = "open_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    if(rwmode /= create) then
      call this%init(0, status)
      this%filename = filename
      call ffits_get_iounit(this%unit, status)
      call ffits_open_file(this%unit, this%filename, rwmode, status)
      if(status == ok) call wrlog(function, "Skymap " // trim(this%filename) // " successfully opened")
      call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
    else if (rwmode == create) then
      call ffits_get_iounit(this%unit, status)
      this%filename = filename
      call ffits_create_file(this%unit, this%filename, fblocksize, status)
    end if

  end subroutine open_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close the skymap file
  !
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tSkymap(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    character(*), parameter :: function = "close_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to init a tSkymap type
  !
  !
  !
  !> @param[inout] this
  !> @param[in] nelements
  !> @param[in, optional] extname
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine init_tSkymap(this, nelements, status, extname)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this
    character(*), optional, intent(in) :: extname

    integer, intent(in) :: nelements

    character(*), parameter :: function = "init_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    this%extname = "SMAPS"

    if(present(extname)) then
      this%extname = extname
    end if

    this%nelements = nelements

    call this%smapnr%init("SRVMAP", nelements, status)
    !        call this%smapnr%init("SMAPNR", nelements, status)
    call this%ra_min%init("RA_MIN", nelements, status)
    call this%ra_max%init("RA_MAX", nelements, status)
    call this%dec_min%init("DE_MIN", nelements, status)
    call this%dec_max%init("DE_MAX", nelements, status)
    call this%n_nbrs%init("N_NBRS", nelements, status)
    call this%owner%init("OWNER", nelements, status)

    call this%field1%init("FIELD1", nelements, status)
    call this%field2%init("FIELD2", nelements, status)
    call this%field3%init("FIELD3", nelements, status)
    call this%field4%init("FIELD4", nelements, status)
    call this%field5%init("FIELD5", nelements, status)
    call this%field6%init("FIELD6", nelements, status)
    call this%field7%init("FIELD7", nelements, status)
    call this%field8%init("FIELD8", nelements, status)
    call this%field9%init("FIELD9", nelements, status)

    call this%field1_order%init("O_FIELD1", nelements, status)
    call this%field2_order%init("O_FIELD2", nelements, status)
    call this%field3_order%init("O_FIELD3", nelements, status)
    call this%field4_order%init("O_FIELD4", nelements, status)
    call this%field5_order%init("O_FIELD5", nelements, status)
    call this%field6_order%init("O_FIELD6", nelements, status)
    call this%field7_order%init("O_FIELD7", nelements, status)
    call this%field8_order%init("O_FIELD8", nelements, status)
    call this%field9_order%init("O_FIELD9", nelements, status)

    call this%dir%init("DIR", nelements, status, nvec=512)
    call this%name%init("FILENAME", nelements, status, nvec=64)
    call this%dir_lastsscan%init("LS_DIR", nelements, status, nvec=512)
    call this%name_lastsscan%init("LS_NAME", nelements, status, nvec=64)
    call this%dir_sscan%init("S_DIR", nelements, status, nvec=512)
    call this%name_sscan%init("S_NAME", nelements, status, nvec=64)
    call this%tstart_sscan%init("S_TSTART", nelements, status)

  end subroutine init_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to free the type tSkymap
  !
  !
  !
  !> @param[in] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_tSkymap(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    character(*), parameter :: function = "free_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = 0
    call this%smapnr%free(status)
    call this%ra_min%free(status)
    call this%ra_max%free(status)
    call this%dec_min%free(status)
    call this%dec_max%free(status)
    call this%n_nbrs%free(status)
    call this%owner%free(status)

    call this%field1%free(status)
    call this%field2%free(status)
    call this%field3%free(status)
    call this%field4%free(status)
    call this%field5%free(status)
    call this%field6%free(status)
    call this%field7%free(status)
    call this%field8%free(status)
    call this%field9%free(status)

    call this%field1_order%free(status)
    call this%field2_order%free(status)
    call this%field3_order%free(status)
    call this%field4_order%free(status)
    call this%field5_order%free(status)
    call this%field6_order%free(status)
    call this%field7_order%free(status)
    call this%field8_order%free(status)
    call this%field9_order%free(status)

    call this%dir%free(status)
    call this%name%free(status)

    call this%dir_lastsscan%free(status)
    call this%name_lastsscan%free(status)

    call this%dir_sscan%free(status)
    call this%name_sscan%free(status)
    call this%tstart_sscan%free(status)


  end subroutine free_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_tSkymap(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tSkymap) :: this

    integer, dimension(:), intent(in) :: order

    character(*), parameter :: function = "apply_index_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = size(order)
    call this%smapnr%apply_index(order, status)
    call this%ra_min%apply_index(order, status)
    call this%ra_max%apply_index(order, status)
    call this%dec_min%apply_index(order, status)
    call this%dec_max%apply_index(order, status)
    call this%owner%apply_index(order, status)
    call this%n_nbrs%apply_index(order, status)
    call this%owner%apply_index(order, status)

    call this%dir%apply_index(order, status)
    call this%name%apply_index(order, status)
    call this%dir_lastsscan%apply_index(order, status)
    call this%name_lastsscan%apply_index(order, status)
    call this%dir_sscan%apply_index(order, status)
    call this%name_sscan%apply_index(order, status)
    call this%tstart_sscan%apply_index(order, status)

    call this%field1%apply_index(order, status)
    call this%field2%apply_index(order, status)
    call this%field3%apply_index(order, status)
    call this%field4%apply_index(order, status)
    call this%field5%apply_index(order, status)
    call this%field6%apply_index(order, status)
    call this%field7%apply_index(order, status)
    call this%field8%apply_index(order, status)
    call this%field9%apply_index(order, status)

    call this%field1_order%apply_index(order, status)
    call this%field2_order%apply_index(order, status)
    call this%field3_order%apply_index(order, status)
    call this%field4_order%apply_index(order, status)
    call this%field5_order%apply_index(order, status)
    call this%field6_order%apply_index(order, status)
    call this%field7_order%apply_index(order, status)
    call this%field8_order%apply_index(order, status)
    call this%field9_order%apply_index(order, status)

  end subroutine apply_index_tSkymap

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tSkymap
  !> insert the contents into a tbintable description
  !
  !
  !> @param[inout] this
  !> @param[in]
  !> @param[out]
  !> @param[inout] table
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_extension_tSkymap(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ncol

    class(tSkymap) :: this
    character(*), parameter :: function = "create_extension_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ncol = 20

    call this%table%init(ncol, status)

    call this%table%insert_ffits_col(this%smapnr, status)
    call this%table%insert_ffits_col(this%ra_min, status)
    call this%table%insert_ffits_col(this%ra_max, status)
    call this%table%insert_ffits_col(this%dec_min, status)
    call this%table%insert_ffits_col(this%dec_max, status)
    call this%table%insert_ffits_col(this%owner, status)
    call this%table%insert_ffits_col(this%n_nbrs, status)

    call this%table%insert_ffits_col(this%dir, status)
    call this%table%insert_ffits_col(this%name, status)
    call this%table%insert_ffits_col(this%dir_lastsscan, status)
    call this%table%insert_ffits_col(this%name_lastsscan, status)
    call this%table%insert_ffits_col(this%dir_sscan, status)
    call this%table%insert_ffits_col(this%name_sscan, status)
    call this%table%insert_ffits_col(this%tstart_sscan, status)

    call this%table%insert_ffits_col(this%field1, status)
    call this%table%insert_ffits_col(this%field2, status)
    call this%table%insert_ffits_col(this%field3, status)
    call this%table%insert_ffits_col(this%field4, status)
    call this%table%insert_ffits_col(this%field5, status)
    call this%table%insert_ffits_col(this%field6, status)
    call this%table%insert_ffits_col(this%field7, status)
    call this%table%insert_ffits_col(this%field8, status)
    call this%table%insert_ffits_col(this%field9, status)

    call this%table%insert_ffits_col(this%field1_order, status)
    call this%table%insert_ffits_col(this%field2_order, status)
    call this%table%insert_ffits_col(this%field3_order, status)
    call this%table%insert_ffits_col(this%field4_order, status)
    call this%table%insert_ffits_col(this%field5_order, status)
    call this%table%insert_ffits_col(this%field6_order, status)
    call this%table%insert_ffits_col(this%field7_order, status)
    call this%table%insert_ffits_col(this%field8_order, status)
    call this%table%insert_ffits_col(this%field9_order, status)

    call this%table%insertFits(this%unit, this%extname, status)

  end subroutine create_extension_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the catalog filenames of the adjacent skyfields to the target skyfield
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_adjacent_catalog_name_tSkymap(this, skyfield, filename, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    integer, intent(in) :: skyfield
    character(buffsize), intent(out), allocatable, dimension(:) :: filename

    integer, dimension(:), allocatable :: order
    integer :: order_index, order_index_adj, n_nbrs

    character(*), parameter :: function = "get_adjacent_catalog_name_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! find the smapnr matching the skyfield nr
    order = where_i4(this%smapnr%value, [skyfield], status)

    if(size(order) /= 1) then
      call wrstop(function, "could not find skyfield match in skymap file")
    end if

    order_index = order(1)
    n_nbrs = this%n_nbrs%value(order_index)
    ! allocate with the max number of skyfields
    call alloc(filename, n_nbrs, status)

    order_index_adj = this%field1_order%value(order_index)
    filename(1) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))
    order_index_adj = this%field2_order%value(order_index)
    filename(2) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))
    order_index_adj = this%field3_order%value(order_index)
    filename(3) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))
    order_index_adj = this%field4_order%value(order_index)
    filename(4) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))
    order_index_adj = this%field5_order%value(order_index)
    filename(5) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))
    order_index_adj = this%field6_order%value(order_index)
    filename(6) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))

    ! 6 skyfields are always adjacent - only special cases are 7,8,9 for adjacent skyfields

    if(this%n_nbrs%value(order(1)) > 6) then
      order_index_adj = this%field7_order%value(order_index)
      filename(7) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))

    end if

    if(this%n_nbrs%value(order(1)) > 7) then
      order_index_adj = this%field8_order%value(order_index)
      filename(8) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))
    end if

    if(this%n_nbrs%value(order(1))>8) then
      order_index_adj = this%field9_order%value(order_index)
      filename(9) = trim(this%dir%value(order_index_adj)) // trim(this%name%value(order_index_adj))
    end if

    call dealloc(order, status)

  end subroutine get_adjacent_catalog_name_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the catalog filename for the target skyfield
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_catalog_name_tSkymap(this, skyfield, filename, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    integer, intent(in) :: skyfield
    character(buffsize), intent(out) :: filename

    integer, dimension(:), allocatable :: order
    integer :: order_index

    character(*), parameter :: function = "get_catalog_name_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! find the smapnr matching the skyfield nr
    order = where_i4(this%smapnr%value, [skyfield], status)

    if(size(order) /= 1) then
      call wrstop(function, "could not find skyfield match in skymap file")
    end if

    order_index = order(1)
    filename = trim(this%dir%value(order_index)) // trim(this%name%value(order_index))

    call dealloc(order, status)

  end subroutine get_catalog_name_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write name and dir of new catalog to skymap
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_and_write_new_catalog_name_tSkymap(this, skyfield, filename, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    integer, intent(in) :: skyfield
    character(buffsize), intent(in) :: filename

    integer, dimension(:), allocatable :: order
    integer :: order_index, l_index

    character(buffsize) :: cwd

    character(*), parameter :: function = "set_and_write_new_catalog_name_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! find the smapnr matching the skyfield nr
    order = where_i4(this%smapnr%value, [skyfield], status)

    if(size(order) /= 1) then
      call wrstop(function, "could not find skyfield match in skymap file")
    end if

    order_index = order(1)
    l_index = scan(filename, "/", .true.)

    if(l_index > 0) then
      this%dir%value(order_index) = trim(filename(:l_index))
      this%name%value(order_index) = trim(filename(l_index+1:))
    else
      call getcwd(cwd, status)
      if(status /= 0) then
        call wrerr(function, s_err, "cannot get current working directory")
      end if
      this%dir%value(order_index) = trim(cwd) // "/"
      this%name%value(order_index) = trim(filename)
    end if

    call this%dir%write(this%unit, status, frow=order_index, lrow=order_index, add=false, first_row=order_index)
    call this%name%write(this%unit, status, frow=order_index, lrow=order_index, add=false, first_row=order_index)
    call dealloc(order, status)

  end subroutine set_and_write_new_catalog_name_tSkymap

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write name and dir of new catalog to skymap
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_and_write_lastsscan_tSkymap(this, skyfield, filename, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    integer, intent(in) :: skyfield
    character(buffsize), intent(in) :: filename

    integer, dimension(:), allocatable :: order
    integer :: order_index, l_index

    character(buffsize) :: cwd

    character(*), parameter :: function = "set_and_write_lastsscan_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! find the smapnr matching the skyfield nr
    order = where_i4(this%smapnr%value, [skyfield], status)

    if(size(order) /= 1) then
      call wrstop(function, "could not find skyfield match in skymap file")
    end if

    order_index = order(1)
    l_index = scan(filename, "/", .true.)

    if(l_index > 0) then
      this%dir_lastsscan%value(order_index) = trim(filename(:l_index))
      this%name_lastsscan%value(order_index) = trim(filename(l_index+1:))
    else
      call getcwd(cwd, status)
      if(status /= 0) then
        call wrerr(function, s_err, "cannot get current working directory")
      end if
      this%dir_lastsscan%value(order_index) = trim(cwd) // "/"
      this%name_lastsscan%value(order_index) = trim(filename)
    end if

    call this%dir_lastsscan%write(this%unit, status, frow=order_index, lrow=order_index, add=false, first_row=order_index)
    call this%name_lastsscan%write(this%unit, status, frow=order_index, lrow=order_index, add=false, first_row=order_index)
    call dealloc(order, status)

  end subroutine set_and_write_lastsscan_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write name and dir of new catalog to skymap
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_and_write_sscan_tSkymap(this, skyfield, filename, tstart, status)
    implicit none

    integer, intent(inout) :: status
    class(tSkymap) :: this

    integer, intent(in) :: skyfield
    real(r8), intent(in) :: tstart
    character(buffsize), intent(in) :: filename

    integer, dimension(:), allocatable :: order
    integer :: order_index, l_index

    character(buffsize) :: cwd

    character(*), parameter :: function = "set_and_write_sscan_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    ! find the smapnr matching the skyfield nr
    order = where_i4(this%smapnr%value, [skyfield], status)

    if(size(order) /= 1) then
      call wrstop(function, "could not find skyfield match in skymap file")
    end if

    order_index = order(1)
    l_index = scan(filename, "/", .true.)

    if(l_index > 0) then
      this%dir_sscan%value(order_index) = trim(filename(:l_index))
      this%name_sscan%value(order_index) = trim(filename(l_index+1:))
    else
      call getcwd(cwd, status)
      if(status /= 0) then
        call wrerr(function, s_err, "cannot get current working directory")
      end if
      this%dir_sscan%value(order_index) = trim(cwd) // "/"
      this%name_sscan%value(order_index) = trim(filename)
    end if

    call this%dir_sscan%write(this%unit, status, frow=order_index, lrow=order_index, add=false, first_row=order_index)
    call this%name_sscan%write(this%unit, status, frow=order_index, lrow=order_index, add=false, first_row=order_index)
    call this%tstart_sscan%write(this%unit, status, frow=order_index, lrow=order_index, add=false, first_row=order_index)
    call dealloc(order, status)

  end subroutine set_and_write_sscan_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine get_skyfield_radec_box_tSkymap.
  !> @brief
  !> return the skyfields included in a radec box
  !
  !
  !> @param this, ra_min, dec_min, ra_max, dec_max, skyfield
  !> @param[in]
  !> @param[out, optional] n_iki, n_mpe, n_mixed
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine determine_skyfield_radec_box_tSkymap(this, ra_min, ra_max, dec_min, dec_max, skyfield, mpe_skyfield, warn, status, n_iki, n_mpe, n_mixed, comment)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tSkymap) :: this
    real(r8), intent(in) :: ra_min, ra_max, dec_min, dec_max
    integer, allocatable, dimension(:), intent(out) :: skyfield
    integer, allocatable, dimension(:), intent(out) :: mpe_skyfield
    integer, intent(out) :: warn
    character(buffsize), optional :: comment
    integer, optional, intent(out) :: n_iki, n_mpe, n_mixed

    !internal param
    integer(i4), dimension(this%nelements) :: index_ra, index_dec, index_multi
    integer :: add_360 ! modified ra_max
    integer(i4) :: n_skyfields, n_ii, ii ! number of skyfields

    integer :: n_iki_opt, n_mpe_opt, n_mixed_opt

    character(*), parameter :: function = "determine_skyfield_radec_box_tSkymap"

    if(this%nelements < 1) then
      call wrstop(function, "skymap file has zero elements")
    end if

    ! check for ra_min
    if(ra_max < ra_min) then
      add_360 = true
    else
      add_360 = false
    end if

    index_ra = 0
    index_dec = 0

    do ii = 1, this%nelements
      ! check whether the dec coordinates for the skyfield are within the boarders
      if((dec_min <= this%dec_min%value(ii)) .and. (this%dec_min%value(ii) <= dec_max)) index_dec(ii) = 1
      if((dec_min <= this%dec_max%value(ii)) .and. (this%dec_max%value(ii) <= dec_max)) index_dec(ii) = 1


      if(add_360 == false) then
        if((ra_min <= this%ra_min%value(ii)) .and. (this%ra_min%value(ii) <= ra_max)) index_ra(ii) = 1
        if((ra_min <= this%ra_max%value(ii)) .and. (this%ra_min%value(ii)<= ra_max)) index_ra(ii) = 1
      else
        if((ra_min <= this%ra_min%value(ii)) .and. (this%ra_min%value(ii)<= ra_max + 360.0_r8)) index_ra(ii) = 1
        if((ra_min <= this%ra_max%value(ii)) .and. (this%ra_min%value(ii)<= ra_max + 360.0_r8)) index_ra(ii) = 1
        if((ra_min <= this%ra_min%value(ii) + 360.0_r8) .and. (this%ra_min%value(ii) + 360.0_r8 <= ra_max + 360.0_r8)) index_ra(ii) = 1
          ! this does not need to be checked since all skyfields are definied between 0 and 360 and no skyfield has a boarder crossing 360
          ! if ra_max would be included so does ra_min therefore we do not need to check it
          ! if(ra_min =< this%ra_max(ii) + 360.0_r8 =< ra_max + 360.0_r8) index_ra(ii) = 1
      end if
    end do

    index_multi = index_ra * index_dec
    n_skyfields = sum(index_multi)

    n_iki_opt = 0
    n_mpe_opt = 0
    n_mixed_opt = 0
    n_ii = 1

    call alloc(skyfield, n_skyfields, status)
    call alloc(mpe_skyfield, n_skyfields, status)

    do ii = 1, this%nelements
      if(index_ra(ii) * index_dec(ii) == 1) then
        skyfield(n_ii) = this%smapnr%value(ii)
        select case (this%owner%value(ii))
          case (0)
            n_mixed_opt = n_mixed_opt + 1
            mpe_skyfield(n_mixed_opt + n_mpe_opt) =  this%smapnr%value(ii)
          case (1)
            n_iki_opt = n_iki_opt + 1
          case (2)
            n_mpe_opt = n_mpe_opt + 1
            mpe_skyfield(n_mixed_opt + n_mpe_opt) =  this%smapnr%value(ii)
          case default
            call wrerr(function, warn, "contact the developer - irregular skyfield found")
        end select

        n_ii = n_ii + 1
      end if
    end do

    if(present(n_mixed)) n_mixed = n_mixed_opt
    if(present(n_iki)) n_iki = n_iki_opt
    if(present(n_mpe)) n_mpe = n_mpe_opt

    warn = false
    if((n_mpe_opt + n_mixed_opt) < 1) then
      warn = true
      if(present(comment)) then
        comment = "the radec_box " // trim(double2str(ra_min, status)) // " " // trim(double2str(ra_max, status)) // " "
        comment = trim(comment) // trim(double2str(dec_min, status)) // " " // trim(double2str(dec_max, status)) // " "
        comment = trim(comment) // "does not contain MPE data"
      end if
    end if


    if(n_mixed_opt + n_mpe_opt > 0) then
      mpe_skyfield = mpe_skyfield(1:n_mixed_opt + n_mpe_opt)
    else
      call alloc(mpe_skyfield, 0, status)
    end if 

    if(status > 100 .or. status == s_errRep) return

  end subroutine determine_skyfield_radec_box_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine check_skyfield_tSkymap.
  !> @brief
  !> check whether a skyfield exists and is mpe
  !> produce comments if skyfield nr does not exists or is iki
  !
  !
  !> @param this, skyfield, exist, comment
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine check_skyfield_tSkymap(this, skyfield, comment, warn, status, reject)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    integer, intent(in) :: skyfield
    integer, intent(out) :: warn
    character(buffsize), intent(out) :: comment
    integer, optional, intent(out) :: reject ! optional parameter to determine whether skyfield is "legal" to request
    class(tSkymap) :: this

    !internal param
    integer, dimension(:), allocatable :: order
    character(buffsize) :: skyfield_str
    integer :: owner

    character(*), parameter :: function = "check_skyfield_tSkymap"

    if(status > 100 .or. status == s_errRep) return

    comment = ""

    ! find the smapnr matching the skyfield nr
    skyfield_str = int2str(skyfield, status)
    order = where_i4(this%smapnr%value, [skyfield], status)

    if(size(order) /= 1) then
      owner = 3
    else
      owner = this%owner%value(order(1))
    end if

    warn = false
    select case (owner)
      case (0)
        comment = "selected skyfield belongs to IKI and MPE - " // trim(skyfield_str)
        if(present(reject)) reject = false
      case (1)
        comment = "selected skyfield belongs to IKI - " // trim(skyfield_str)
        if(present(reject)) reject = true
        warn = true
      case (2)
        comment = "selected skyfield belongs to MPE - " // trim(skyfield_str)
        if(present(reject)) reject = false
      case default
        comment = "could not find skyfield - " // trim(skyfield_str)
        if(present(reject)) reject = true
    end select


    call dealloc(order, status)

  end subroutine check_skyfield_tSkymap

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> interprete the content from an ASCII file as skyfields and/or skyfield boxes
  !> the skymap type must already be filled
  !
  !
  !> @param[in] filename, skymap_name
  !> @param[out, optional] comment
  !> @param[out] skyfield
  !> @param[inout,optional] status
  !---------------------------------------------------------------------------

  subroutine interpret_ASCII_skyfield(filename, skyfield, skymap_name, nreject, comment, status)
    implicit none

    character(*), intent(in) :: filename
    character(*), parameter :: function = "interpret_ASCII_skyfield"
    character(buffsize) :: iomsg, line, comment_subroutine
    character(buffsize), optional, dimension(:), allocatable, intent(out) :: comment

    character(buffsize), intent(in) :: skymap_name
    character(buffsize), dimension(:), allocatable :: regTok
    type(tSkymap) :: skymap
    integer, dimension(:), allocatable, intent(out) :: skyfield
    integer :: skyfield_int
    integer, dimension(:), allocatable :: skyfield_all, skyfield_mpe
    integer, intent(out) :: nreject ! number of total rejects

    integer :: unit, exists, iostat, pos
    integer :: n_iki, n_mpe, n_mixed, warn
    integer, intent(inout) :: status

    real(r8) :: ra_min, ra_max, dec_min, dec_max, switch_dec

    integer :: nlines, ncomment, success, reject

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) return

    ! check whether the file exists
    call ffits_file_exists(filename, exists, status)
    ! if the file does not exist return
    if (exists == 0) then
      status = s_err
      call wrerr(function, status, trim(filename) // " - does not exist")
      status = s_errRep
      return
    end if

    ! reject always if something went wrong
    reject = true

    ! read the skymap file
    call skymap%open(skymap_name, readonly, status)
    call skymap%read(status)
    call skymap%close(status)

    call ffits_get_iounit(unit, status)
    open (unit, file=filename, status="old",action="read", iostat=iostat, iomsg=iomsg)
    call FError(function // " - open", iostat, iomsg, status)

    ! get the number of lines
    nlines = 0
    do
      read(unit, '(A)', iostat=iostat, iomsg=iomsg) line
      ! have to escape the sequence instead to stop here
      if(iostat /= 0) exit
      nlines = nlines + 1
    end do

    call alloc(comment, nlines, status)

    rewind(unit, iostat=iostat)
    if(iostat /= 0) call wrstop(function, "rewind of file " // trim(filename) // " failed")

    ncomment = 0
    nlines = 0
    nreject = 0

    ! since nothing went wrong set reject to false
    reject = false
    do
      read(unit, '(A)', iostat=iostat, iomsg=iomsg) line

      ! have to escape the sequence instead to stop here
      !            call FError(function // " - read " // trim(filename), iostat, iomsg, status)
      if(iostat /= 0) exit
      if(status == s_errRep) return

      ! delete the comments marked by #
      line = adjustl(line)
      pos = index(line, "#")
      select case (pos)
        case (0) ! do nothing
        case (1) ! whole line is a comment
          cycle
        case default ! reduce the line to only contain everything before the #
          line = line(1:pos-1)
      end select

      ! ignore empty lines
      pos = verify(line, " ")
      if(pos == 0) then
        cycle
      end if

      ! split input line into multiple substrings
      call strtok(line, " ", regTok)
      success = false

      select case (size(regTok))
        case (1)
          skyfield_int = str2int(trim(line), status)
          call skymap%check_skyfield(skyfield_int, comment_subroutine, warn, status, reject=reject)
          if(reject == true) then
            ncomment = ncomment + 1
            comment(ncomment) = trim(comment_subroutine)
            nreject = nreject + 1
          else
            if(warn == true) then
              ncomment = ncomment + 1
              comment(ncomment) = trim(comment_subroutine)
            end if
            nlines = nlines + 1
            if(nlines == 1) then
              skyfield = [skyfield_int]
            else
              skyfield = [skyfield, skyfield_int]
            end if
          end if
        case (4)
          call str2deg(regTok(1), ra_min, status, convert=.true.)
          call str2deg(regTok(2), ra_max, status, convert=.true.)
          call str2deg(regTok(3), dec_min, status)
          call str2deg(regTok(4), dec_max, status)

          if(status /= ok) then
            ncomment = ncomment + 1
            comment(ncomment) = "cannot interpret line - " // trim(line)
            status = ok
            nreject = nreject + 1
            cycle
          end if

          ! check that dec_min < dec_max
          if(dec_max < dec_min) then
            ncomment = ncomment + 1
            comment(ncomment) = "dec_min > dec_max - switched dec values for - " // trim(line)
            switch_dec = dec_max
            dec_max = dec_min
            dec_min = switch_dec
          end if

          call skymap%determine_skyfield_radec_box(ra_min, ra_max, dec_min, dec_max, skyfield_all, skyfield_mpe, warn, status, comment=comment_subroutine)

          if(warn == false) then
            nlines = nlines + 1
            if(nlines == 1) then
              skyfield = skyfield_mpe
            else
              skyfield = [skyfield, skyfield_mpe]
            end if
          else
            ncomment = ncomment + 1
            comment(ncomment) = comment_subroutine
          end if

        case default
          ncomment = ncomment + 1
          comment(ncomment) = "cannot interpret line - " // trim(line)
          nreject = nreject + 1
      end select
    end do

    if(nreject > 0) then
      call alloc(skyfield, 0, status)
    else
      skyfield = skyfield(unique(skyfield, status))
    end if

    skyfield = skyfield(unique(skyfield, status))

    call skymap%free(status)

    call dealloc(regTok, status)

    close(unit, iostat=iostat, iomsg=iomsg)
    call ffits_free_iounit(unit, status)
    call FError(function // " - close "// trim(filename), iostat, iomsg, status)

  end subroutine interpret_ASCII_skyfield


end module catprep_skymap_mod
