! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_specfit_mod

    use param
    use erolib_general_mod
    use erolib_ffits_col_mod
    use erolib_sort_mod
    use erolib_tbintable_mod
    implicit none

    type, public :: tSpecfit

        integer :: nelements
        type(r4_tColumn) :: nh
        type(r4_tColumn) :: nh_u
        type(r4_tColumn) :: nh_l
        type(r4_tColumn) :: photon_u
        type(r4_tColumn) :: photon_l
        type(r4_tColumn) :: norm_u
        type(r4_tColumn) :: norm_l
        type(r4_tColumn) :: norm
        type(r4_tColumn) :: photon
        type(r4_tColumn) :: chisq

    contains
        procedure :: init => init_tSpecfit
        procedure :: free => free_tSpecfit
        procedure :: writeToFits => write_tSpecfit
        procedure :: writeToFits_frow_lrow => write_frow_lrow_tSpecfit
        generic ::  write => writeToFits, writeToFits_frow_lrow
        procedure :: readFromFits => read_tSpecfit
        procedure :: readFromFits_frow_lrow => read_frow_lrow_tSpecfit
        generic :: read => readFromFits, readFromFits_frow_lrow
        procedure :: apply_index => apply_index_tSpecfit
        procedure :: insert_into_tBinTable => insert_into_tBinTable_tSpecfit
    end type tSpecfit

contains

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge all entries of the columns with a second Varichck
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_specfit(Specfit_A, Specfit_B, Specfit_C, status)
        implicit none

        integer, intent(inout) :: status

        class(tSpecfit), intent(in) :: Specfit_A
        class(tSpecfit), intent(in) :: Specfit_B
        class(tSpecfit), intent(out) :: Specfit_C

        character(*), parameter :: function = "merge_specfit"

        if(status > 100 .or. status == s_errRep) return

        Specfit_C%nelements = Specfit_A%nelements + Specfit_B%nelements

        call merge_tColumn(Specfit_A%nh, Specfit_B%nh, Specfit_C%nh, status)
        call merge_tColumn(Specfit_A%nh_u, Specfit_B%nh_u, Specfit_C%nh_u, status)
        call merge_tColumn(Specfit_A%nh_l, Specfit_B%nh_l, Specfit_C%nh_l, status)
        call merge_tColumn(Specfit_A%photon_u, Specfit_B%photon_u, Specfit_C%photon_u, status)
        call merge_tColumn(Specfit_A%photon_l, Specfit_B%photon_l, Specfit_C%photon_l, status)
        call merge_tColumn(Specfit_A%norm_u, Specfit_B%norm_u, Specfit_C%norm_u, status)
        call merge_tColumn(Specfit_A%norm_l, Specfit_B%norm_l, Specfit_C%norm_l, status)

        call merge_tColumn(Specfit_A%norm, Specfit_B%norm, Specfit_C%norm, status)
        call merge_tColumn(Specfit_A%photon, Specfit_B%photon, Specfit_C%photon, status)
        call merge_tColumn(Specfit_A%chisq, Specfit_B%chisq, Specfit_C%chisq, status)


    end subroutine merge_specfit

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tSpecfit
    !> allocate the memory for this type
    !
    !
    !> @param[inout] this
    !> @param[in]  tel_id, band_id, readwrite, nelements
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tSpecfit(this, nelements, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) ::  nelements

        class(tSpecfit) :: this
        character(*), parameter :: function = "init_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = nelements
        call this%nh%init("NH", nelements, status, tunit="[cm^-2]")
        call this%nh_u%init("NH_U", nelements, status, tunit="[cm^-2]")
        call this%nh_l%init("NH_L", nelements, status, tunit="[cm^-2]")
        call this%photon_u%init("PHOTON_U", nelements, status)
        call this%photon_l%init("PHOTON_L", nelements, status)
        call this%norm_u%init("NORM_U", nelements, status, tunit="[erg/cm^-2]")
        call this%norm_l%init("NORM_L", nelements, status, tunit="[erg/cm^-2]")
        call this%photon%init("PHOTON", nelements, status)
        call this%norm%init("NORM", nelements, status, tunit="[erg/cm^-2]")
        call this%norm%init("CHISQ", nelements, status)

    end subroutine init_tSpecfit

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tSpecfit
    !> free the tSpecfit
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tSpecfit(this, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, jj

        class(tSpecfit) :: this
        character(*), parameter :: function = "free_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = 0
        call this%nh%free(status)
        call this%nh_u%free(status)
        call this%nh_l%free(status)
        call this%photon_u%free(status)
        call this%photon_l%free(status)
        call this%norm_u%free(status)
        call this%norm_l%free(status)
        call this%photon%free(status)
        call this%norm%free(status)

    end subroutine free_tSpecfit

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tSpecfit
    !> write a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in, optional] extname
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tSpecfit(this, unit, status, extname)
        implicit none

        integer, intent(inout) :: status
        class(tSpecfit) :: this
        integer :: ii, jj
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), intent(in), optional :: extname

        character(*), parameter :: function = "write_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        if(present(extname)) call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%nh%write(unit, status, frow=frow, lrow=lrow)
            call this%nh_u%write(unit, status, frow=frow, lrow=lrow)
            call this%nh_l%write(unit, status, frow=frow, lrow=lrow)
            call this%photon_u%write(unit, status, frow=frow, lrow=lrow)
            call this%photon_l%write(unit, status, frow=frow, lrow=lrow)
            call this%norm_u%write(unit, status, frow=frow, lrow=lrow)
            call this%norm_l%write(unit, status, frow=frow, lrow=lrow)
            call this%norm%write(unit, status, frow=frow, lrow=lrow)
            call this%photon%write(unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine write_tSpecfit

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tSpecfit
    !> read a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_tSpecfit(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        class(tSpecfit) :: this
        integer :: ii
        integer, intent(in) :: unit
        integer :: frow, lrow, nelements, rowsize, nrows

        character(*), parameter :: function = "read_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        call ffits_get_rowsize(unit, rowsize, status)
        nelements = this%nelements

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%nh%read(unit, status, frow=frow, lrow=lrow)
            call this%nh_u%read(unit, status, frow=frow, lrow=lrow)
            call this%nh_l%read(unit, status, frow=frow, lrow=lrow)
            call this%photon_u%read(unit, status, frow=frow, lrow=lrow)
            call this%photon_l%read(unit, status, frow=frow, lrow=lrow)
            call this%norm_u%read(unit, status, frow=frow, lrow=lrow)
            call this%norm_l%read(unit, status, frow=frow, lrow=lrow)
            call this%norm%read(unit, status, frow=frow, lrow=lrow)
            call this%photon%read(unit, status, frow=frow, lrow=lrow)

        end do

    end subroutine read_tSpecfit

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tSpecfit
    !> write a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_frow_lrow_tSpecfit(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tSpecfit) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "write_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        call wrlog(function, "write varichck")
        call this%nh%write(unit, status, frow=frow, lrow=lrow)
        call this%nh_u%write(unit, status, frow=frow, lrow=lrow)
        call this%nh_l%write(unit, status, frow=frow, lrow=lrow)
        call this%photon_u%write(unit, status, frow=frow, lrow=lrow)
        call this%photon_l%write(unit, status, frow=frow, lrow=lrow)
        call this%norm_u%write(unit, status, frow=frow, lrow=lrow)
        call this%norm_l%write(unit, status, frow=frow, lrow=lrow)
        call this%photon%write(unit, status, frow=frow, lrow=lrow)
        call this%norm%write(unit, status, frow=frow, lrow=lrow)

    end subroutine write_frow_lrow_tSpecfit

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tSpecfit
    !> read a subset of the tb data to fits
    !
    !
    !> @param[inout] this
    !> @param[in] unit
    !> @param[in] frow, lrow
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_frow_lrow_tSpecfit(this, unit, frow, lrow, status)
        implicit none

        integer, intent(inout) :: status
        class(tSpecfit) :: this
        integer, intent(in) :: unit,  frow, lrow

        character(*), parameter :: function = "read_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        call this%nh%read(unit, status, frow=frow, lrow=lrow)
        call this%nh_u%read(unit, status, frow=frow, lrow=lrow)
        call this%nh_l%read(unit, status, frow=frow, lrow=lrow)
        call this%photon_u%read(unit, status, frow=frow, lrow=lrow)
        call this%photon_l%read(unit, status, frow=frow, lrow=lrow)
        call this%norm_u%read(unit, status, frow=frow, lrow=lrow)
        call this%norm_l%read(unit, status, frow=frow, lrow=lrow)
        call this%photon%read(unit, status, frow=frow, lrow=lrow)
        call this%norm%read(unit, status, frow=frow, lrow=lrow)

    end subroutine read_frow_lrow_tSpecfit

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_tSpecfit(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tSpecfit) :: this

        integer, dimension(:), intent(in) :: order

        character(*), parameter :: function = "apply_index_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = size(order)
        call this%nh%apply_index(order, status)
        call this%nh_u%apply_index(order, status)
        call this%nh_l%apply_index(order, status)
        call this%photon_u%apply_index(order, status)
        call this%photon_l%apply_index(order, status)
        call this%norm_u%apply_index(order, status)
        call this%norm_l%apply_index(order, status)
        call this%photon%apply_index(order, status)
        call this%norm%apply_index(order, status)

    end subroutine apply_index_tSpecfit

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tSpecfit
    !> insert the contents into a tbintable description
    !
    !
    !> @param[inout] this
    !> @param[in]
    !> @param[out]
    !> @param[inout] table
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine insert_into_tBinTable_tSpecfit(this, table, status)
        implicit none

        integer, intent(inout) :: status

        type(tBinTable), intent(inout) :: table

        integer :: ii, jj

        class(tSpecfit) :: this
        character(*), parameter :: function = "insert_into_tBinTable_tSpecfit"

        if(status > 100 .or. status == s_errRep) return

        call wrlog(function, "CREATE SPECFIT")

        call table%insert_ffits_col(this%nh, status)
        call table%insert_ffits_col(this%nh_u, status)
        call table%insert_ffits_col(this%nh_l, status)
        call table%insert_ffits_col(this%photon_u, status)
        call table%insert_ffits_col(this%photon_l, status)
        call table%insert_ffits_col(this%norm_u, status)
        call table%insert_ffits_col(this%norm_l, status)
        call table%insert_ffits_col(this%photon, status)
        call table%insert_ffits_col(this%norm, status)

    end subroutine insert_into_tBinTable_tSpecfit


end module catprep_specfit_mod
