! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_srclist_mod
    use param
    use erolib_tbintable_mod
    use erolib_region_mod
    use erolib_ffits_col_mod
    use erolib_math_mod
    !    use erolib_sort_mod

    use catprep_ape_mod
    use catprep_eb_mod
    use catprep_catextend_mod

    !    use catprep_tb_data_mod
    !    use catprep_varichck_mod
    !    use catprep_specfit_mod

    use srctool_generic , only : coord_type
    use srctool_generic_regions , only : srctool_regions_test_if_coord_in_region, region_type
    
    use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr

    implicit none

logical, private, parameter :: doTest=.false.

    ! band ids used
    integer, dimension(5), private :: band_id =  [0,1,2,3,4]

    ! extnames used in the files
    character(*), parameter, private :: catsrclist_ext = "CATSRCL"

    ! basic type of the catalog used for the cumulative list
    type, public :: tCatSrcList
        integer :: detuid_size=32 ! number of characters to use for the detUID column.
        integer :: unit
        character(buffsize) :: filename
        character(buffsize) :: extname
        type(tBinTable) :: table ! description of the binary table
        type(stdhdr_type) :: stdhdr

        integer :: nelements ! number of elements (rows)

        type(i4_tColumn) :: id_src ! source id based on ermldet run
        type(i4_tColumn) :: id_cluster ! cluster id based on ermldet run

        type(str_tColumn) :: detuid     ! unique detection id

        type(i1_tColumn) :: owner ! owner column

        type(r8_tColumn) :: ra          ! [deg]
        type(r8_tColumn) :: dec         ! [deg]
        type(r4_tColumn) :: radec_err   ! [sec] 1 sigma error

        type(i1_tColumn) :: quality     ! quality number, higher is better (?)
        type(i1_tColumn) :: qf          !  binary quality flags

        type(r8_tColumn) :: lII         !   [deg]              Galactic longitude
        type(r8_tColumn) :: bII         !   [deg]              Galactic latitude

        !        type(r4_tColumn) :: dist_nn     !   [sec]              Distance to nearest neighbour
        type(r4_tColumn) :: ext         !   [sec]              Source extent
        type(r4_tColumn) :: ext_err     !   [sec]              1 sigma error
        type(r4_tColumn) :: ext_like    !                      Extent likelihood 1)

        !Columns specific to each detection algorithm:
        type(r4_tColumn) :: cutrad      !   [pixels]           Source cut-out radius
        type(r4_tColumn) :: maskfrac    !                      Fraction of cut-out area inside detection mask

        ! add nh_gal
        type(r4_tColumn) :: nh_gal

        ! Additional columns for pointed observations:
        type(i4_tColumn) :: rawx        !   I*4                      X detector pixel coordiante
        type(i4_tColumn) :: rawy        !   I*4                      Y detector pixel coordinate

        ! as in detect_io_mod.f90 for the mllist type
        type(r4_tColumn) :: hr_1           ! hardness rations
        type(r4_tColumn) :: hr_1_err       ! 1 sigma errors
        type(r4_tColumn) :: hr_2           ! hardness rations
        type(r4_tColumn) :: hr_2_err       ! 1 sigma errors
        type(r4_tColumn) :: hr_3           ! hardness rations
        type(r4_tColumn) :: hr_3_err       ! 1 sigma errors

        ! columns for each energy band
        type(tEB) :: eb0 ! energy band 0 - parameters
        type(tEB) :: eb1 ! energy band 1 - parameters
        type(tEB) :: eb2 ! energy band 2 - parameters
        type(tEB) :: eb3 ! energy band 3 - parameters
        type(tEB) :: eb4 ! energy band 4 - parameters

        ! auxilliary values which are not read/written by default
        type(i4_tColumn) :: region
        type(i4_tColumn) :: id_match

        ! quicklook information
        type(r4_tColumn) :: dist_nn     !   [sec]              Distance to nearest neighbour
        !        type(r4_tColumn) :: dist_Xnn     !   [sec]              Distance to X nearest neighbour
        type(r4_tColumn) :: srcdensity ! src density of X sources within

        type(r8_tColumn) :: tstart ! tstart of stdhdr for this run
        type(r8_tColumn) :: tstop  ! tstop of stdhdr for this run

        type(tApeCorr) :: ape ! aperture correction parameters

        logical :: catextend
        type(tCatExtend) :: extend ! catalog extension for other tasks

        ! columns for varichck and specfit
        logical :: vs_extend
        type(tVarichck) :: varichck
        type(tSpecFit) :: specfit

    contains
        procedure :: init => init_tCatSrcList
        procedure :: free => free_tCatSrcList
        procedure :: open => open_tCatSrcList
        procedure :: close => close_tCatSrcList
        procedure :: create_extension => create_extension_tCatSrcList
        procedure :: read => read_tCatSrcList
        procedure :: write => write_tCatSrcList
        procedure :: apply_index => apply_index_tCatSrcList
        procedure :: flag_region_ra_dec => flag_region_ra_dec_tCatSrcList
        procedure :: apply_filter_region => apply_filter_region_tCatSrcList
        procedure :: apply_filter_flux => apply_filter_flux_tCatSrcList
        procedure :: sort => sort_tCatSrcList
        procedure :: calc_srcdensity => calc_srcdensity_tCatSrcList
        !        procedure :: calc_dist_Xnn_mem => calc_dist_Xnn_mem_tCatSrcList
        procedure :: locate_souuid => locate_souuid_tCatSrcList
        procedure :: apply_filter_ownerflag => apply_filter_ownerflag_tCatSrcList

    end type tCatSrcList

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloacte n catsrclist structures
    !
    !
    !
    !> @param[inout] catsrclist
    !> @param[in] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine alloc_catsrclist(catsrclist, n, status)
        implicit none

        integer :: stat
        integer, intent(in) :: n
        integer, intent(inout) :: status

        class(tCatSrcList), allocatable, dimension(:) :: catsrclist

        character(buffsize) :: errmsg
        character(*), parameter :: function = "alloc_catsrclist"

        call dealloc_catsrclist(catsrclist, status)
        allocate(catsrclist(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

    end subroutine alloc_catsrclist

      !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> dealloacte catsrclist structures
    !
    !
    !
    !> @param[inout] catsrclist
    !> @param[inout] status
    !---------------------------------------------------------------------------
    subroutine dealloc_catsrclist(catsrclist, status)

        integer :: stat
        integer, intent(inout) :: status

        class(tCatSrcList), allocatable, dimension(:) :: catsrclist

        character(*), parameter :: function = "dealloc_catsrclist"

        if(status > 100 .or. status == s_errRep) return

        if(allocated(catsrclist)) then
            deallocate(catsrclist, stat=stat)
            call FError(function, stat, "dealloc catsrclist type", status)
        end if

    end subroutine dealloc_catsrclist

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> init the tCatSrcList type
    !> nband = number of different energy bands
    !
    !
    !> @param[in] this
    !> @param[out] n
    !> @param[inout] status
    !> @param[in, optional] extend, vs_extend
    !---------------------------------------------------------------------------

    subroutine init_tCatSrcList(this, n, status, extend, vs_extend)
        implicit none

        integer, intent(inout) :: status
        class(tCatSrcList) :: this
        integer, intent(in) :: n

        logical, optional, intent(in) :: extend, vs_extend

        character(*), parameter :: function = "init_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = n
        this%extname = catsrclist_ext

        call this%id_src%init("ID_SRC", n, status)
        call this%id_cluster%init("ID_CLUSTER", n, status)

!        call this%detuid%init("detUID", n, status, nvec=32)
        call this%detuid%init("detUID", n, status, nvec=this%detuid_size)

        call this%owner%init("OWNER", n, status)

        call this%ra%init("RA", n, status)
        call this%dec%init("DEC", n, status)
        call this%radec_err%init("RADEC_ERR", n, status)

        call this%quality%init("QUALITY", n, status)
        call this%qf%init("QF", n, status)
        call this%lII%init("LII", n, status)
        call this%bII%init("BII", n, status)
        call this%ext%init("EXT", n, status, tunit="sec")
        call this%ext_err%init("EXT_ERR", n, status, tunit="sec")
        call this%ext_like%init("EXT_LIKE", n, status)
        call this%cutrad%init("ML_RADIUS", n, status, tunit="min")
        call this%maskfrac%init("MASKFRAC", n, status)

        call this%nh_gal%init("NH_GAL", n, status, tunit="cm^-2")
! 20200907 IMS: nothing seems to be done with this.

        call this%rawx%init("RAWX", n, status)
        call this%rawy%init("RAWY", n, status)

        call this%hr_1%init("HR_1", n, status)
        call this%hr_1_err%init("HR_1_ERR", n, status)
        call this%hr_2%init("HR_2", n, status)
        call this%hr_2_err%init("HR_2_ERR", n, status)
        call this%hr_3%init("HR_3", n, status)
        call this%hr_3_err%init("HR_3_ERR", n, status)

        ! aux columns not written
        call this%region%init("region", n, status)
        call this%id_match%init("id_match", n, status)

        ! quick check columns
        call this%dist_nn %init("DIST_NN", n, status, tunit="sec")
        call this%srcdensity%init("SRCDENS", n, status, tunit="deg^-2")

        call this%tstart%init("TSTART", n, status, tunit="s")
        call this%tstop%init("TSTOP", n, status, tunit="s")

        ! initialise energy bands
        call this%eb0%init(n, status, 0)
        call this%eb1%init(n, status, 1)
        call this%eb2%init(n, status, 2)
        call this%eb3%init(n, status, 3)
        call this%eb4%init(n, status, 4)

        call this%ape%init(n, status)

        this%catextend = .false.
        if(present(extend)) then
            if(extend .eqv. .true.) then
                this%catextend = extend
                call this%extend%init(n, status)
            end if
        end if

        this%vs_extend = .false.
        if(present(vs_extend)) then
            if(vs_extend .eqv. .true.) then
                this%vs_extend = vs_extend
                call this%varichck%init(n, status)
                call this%specfit%init(n, status)
            end if
        end if

    end subroutine init_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the tCatSrcList type
    !
    !
    !> @param[in] this
    !> @param[out] n
    !> @param[inout] status
    !> @param[in, optional] nband
    !---------------------------------------------------------------------------

    subroutine free_tCatSrcList(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tCatSrcList) :: this

        character(*), parameter :: function = "free_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        this%nelements = 0

        call this%id_src%free(status)
        call this%id_cluster%free(status)

        call this%detuid%free(status)

        call this%ra%free(status)
        call this%dec%free(status)
        call this%radec_err%free(status)

        call this%quality%free(status)
        call this%qf%free(status)
        call this%lII%free(status)
        call this%bII%free(status)
        call this%ext%free(status)
        call this%ext_err%free(status)
        call this%ext_like%free(status)
        call this%cutrad%free(status)
        call this%maskfrac%free(status)
        call this%rawx%free(status)
        call this%rawy%free(status)

        call this%hr_1%free(status)
        call this%hr_1_err%free(status)
        call this%hr_2%free(status)
        call this%hr_2_err%free(status)
        call this%hr_3%free(status)
        call this%hr_3_err%free(status)

        ! aux
        call this%region%free(status)
        call this%id_match%free(status)

        ! quick check
        call this%dist_nn%free(status)
        call this%srcdensity%free(status)

        call this%tstart%free(status)
        call this%tstop%free(status)

        call this%eb0%free(status)
        call this%eb1%free(status)
        call this%eb2%free(status)
        call this%eb3%free(status)
        call this%eb4%free(status)

        call this%ape%free(status)

        if(this%catextend .eqv. .true.) then
            call this%extend%free(status)
        end if

        if(this%vs_extend .eqv. .true.) then
            call this%varichck%free(status)
            call this%specfit%free(status)
        end if

    end subroutine free_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> routine to open the CatSrcList
    !
    !
    !
    !> @param[inout] this
    !> @param[in] filename, rwmode
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine open_tCatSrcList(this, filename, rwmode, status, delete)
        implicit none

        integer, intent(inout) :: status
        class(tCatSrcList) :: this

        character(*), intent(in) :: filename
        integer, intent(in) :: rwmode

        integer :: exists

        integer, intent(in), optional :: delete

        character(*), parameter :: function = "open_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        this%filename = filename

        call ffits_get_iounit(this%unit, status)
        if(rwmode /= create) then
            call ffits_open_file(this%unit, filename, rwmode, status)
            if(status == ok) call wrlog(function, "CatSrcList " // trim(this%filename) // " successfully opened")
        else
            call ffits_file_exists(filename, exists, status)
            if(present(delete) .and. exists == 1) then
                if(delete == true) call delete_file_fortran(filename, status)
            end if
            call ffits_create_file(this%unit, filename, fblocksize, status)
            if(status == ok) call wrlog(function, "CatSrcList " // trim(this%filename) // " successfully created")
        end if

    end subroutine open_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> close the tCatSrcList file
    !
    !
    !
    !> @param[inout] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine close_tCatSrcList(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tCatSrcList) :: this

        character(*), parameter :: function = "close_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        call ffits_close_file(this%unit, status)
        call ffits_free_iounit(this%unit, status)

    end subroutine close_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine of class tCatSrcList
    !> create the extension of this type in a FITS file
    !
    !
    !> @param[inout] this
    !> @param[in, optional] aux
    !> @param[out]
    !> @param[inout]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tCatSrcList(this, status, aux)
        implicit none

        integer, intent(inout) :: status
        integer :: ncol ! number of columns in the file
        logical, intent(in), optional :: aux

        class(tCatSrcList) :: this
        character(*), parameter :: function = "create_extension_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        ncol = 20
        call this%table%init(ncol, status)
        call this%table%insert_ffits_col(this%id_src, status)
        call this%table%insert_ffits_col(this%id_cluster, status)

        call this%table%insert_ffits_col(this%detuid, status)
        call this%table%insert_ffits_col(this%owner, status)
    
        !        call this%table%insert_ffits_col(this%unique, status)
        call this%table%insert_ffits_col(this%ra, status)
        call this%table%insert_ffits_col(this%dec, status)
        call this%table%insert_ffits_col(this%radec_err, status)

        call this%table%insert_ffits_col(this%quality, status)
        call this%table%insert_ffits_col(this%qf, status)

        call this%table%insert_ffits_col(this%lII, status)
        call this%table%insert_ffits_col(this%bII, status)

        call this%table%insert_ffits_col(this%dist_nn, status)
        call this%table%insert_ffits_col(this%srcdensity, status)
        call this%table%insert_ffits_col(this%tstart, status)
        call this%table%insert_ffits_col(this%tstop, status)

        call this%table%insert_ffits_col(this%ext, status)
        call this%table%insert_ffits_col(this%ext_err, status)
        call this%table%insert_ffits_col(this%ext_like, status)
        call this%table%insert_ffits_col(this%cutrad, status)
        call this%table%insert_ffits_col(this%maskfrac, status)
        call this%table%insert_ffits_col(this%rawx, status)
        call this%table%insert_ffits_col(this%rawy, status)

        call this%table%insert_ffits_col(this%hr_1, status)
        call this%table%insert_ffits_col(this%hr_1_err, status)
        call this%table%insert_ffits_col(this%hr_2, status)
        call this%table%insert_ffits_col(this%hr_2_err, status)
        call this%table%insert_ffits_col(this%hr_3, status)
        call this%table%insert_ffits_col(this%hr_3_err, status)

        ! insert as blocks
        call this%eb0%insert_into_tBinTable(this%table, status)
        call this%eb1%insert_into_tBinTable(this%table, status)
        call this%eb2%insert_into_tBinTable(this%table, status)
        call this%eb3%insert_into_tBinTable(this%table, status)
        call this%eb4%insert_into_tBinTable(this%table, status)

        call this%ape%insert_into_tBinTable(this%table, status)

        if(this%catextend .eqv. .true.) then
            call this%extend%insert_into_tBinTable(this%table, status)
        end if

        if(this%vs_extend .eqv. .true.) then
            call this%varichck%insert_into_tBinTable(this%table, status)
            call this%specfit%insert_into_tBinTable(this%table, status)
        end if

        ! add aux columns if option is set
        if(present(aux)) then
            if(aux .eqv. .true.) then
                call this%table%insert_ffits_col(this%region, status)
            end if
        end if
        call this%table%insertFits(this%unit, this%extname, status)
    end subroutine create_extension_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the catsrclist type from a FITS file
    !
    !
    !
    !> @param[in] this
    !> @param[out] filename
    !> @param[inout] status
    !> @param[in, optional] extend
    !---------------------------------------------------------------------------

    subroutine read_tCatSrcList(this, status, extend, aux, filename, vs_extend)
        implicit none

        integer, intent(inout) :: status
        integer :: nelements, rowsize, frow, lrow, nrows, ii
        class(tCatSrcList) :: this
        logical, optional, intent(in) :: extend, vs_extend, aux
        character(*), optional, intent(in) :: filename

        character(fcolnamelength) :: col
        character(*), parameter :: function = "read_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(present(filename)) call this%open(filename, readonly, status)

        call ffits_movnam_hdu(this%unit, BINARY_TBL, catsrclist_ext, status)
        call ffits_get_num_rows(this%unit, nelements, status)
        call ffits_get_rowsize(this%unit, rowsize, status)

        if(present(extend) .and. present(vs_extend)) then
            call this%init(nelements, status, extend, vs_extend)
        else if(present(extend)) then
            call this%init(nelements, status, extend)
        else if (present(vs_extend)) then
            call this%init(nelements, status, vs_extend)
        else
            call this%init(nelements, status)
        end if

        call fread_stdhdr(this%unit, this%stdhdr, status)

        if(nelements == 0 .or. status == s_errRep) return

        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1
            call this%id_src%read(this%unit, status, frow=frow, lrow=lrow)
            call this%id_cluster%read(this%unit, status, frow=frow, lrow=lrow)
            call this%detuid%read(this%unit, status, frow=frow, lrow=lrow)
            call this%owner%read(this%unit, status, frow=frow, lrow=lrow)
            call this%ra%read(this%unit, status, frow=frow, lrow=lrow)
            call this%dec%read(this%unit, status, frow=frow, lrow=lrow)
            call this%radec_err%read(this%unit, status, frow=frow, lrow=lrow)
            call this%quality%read(this%unit, status, frow=frow, lrow=lrow)
            call this%qf%read(this%unit, status, frow=frow, lrow=lrow)
            call this%lII%read(this%unit, status, frow=frow, lrow=lrow)
            call this%bII%read(this%unit, status, frow=frow, lrow=lrow)
            call this%dist_nn%read(this%unit, status, frow=frow, lrow=lrow)
            call this%srcdensity%read(this%unit, status, frow=frow, lrow=lrow)
            call this%tstart%read(this%unit, status, frow=frow, lrow=lrow)
            call this%tstop%read(this%unit, status, frow=frow, lrow=lrow)
            call this%ext%read(this%unit, status, frow=frow, lrow=lrow)
            call this%ext_err%read(this%unit, status, frow=frow, lrow=lrow)
            call this%ext_like%read(this%unit, status, frow=frow, lrow=lrow)
            call this%cutrad%read(this%unit, status, frow=frow, lrow=lrow)
            call this%maskfrac%read(this%unit, status, frow=frow, lrow=lrow)
            call this%rawx%read(this%unit, status, frow=frow, lrow=lrow)
            call this%rawy%read(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_1%read(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_1_err%read(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_2%read(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_2_err%read(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_3%read(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_3_err%read(this%unit, status, frow=frow, lrow=lrow)
            call this%eb0%read(this%unit, frow, lrow, status)
            call this%eb1%read(this%unit, frow, lrow, status)
            call this%eb2%read(this%unit, frow, lrow, status)
            call this%eb3%read(this%unit, frow, lrow, status)
            call this%eb4%read(this%unit, frow, lrow, status)
            call this%ape%read(this%unit, frow, lrow, status)
            if(this%catextend .eqv. .true.) then
                ! try reading catextend
                status = ignore
                call this%extend%read(this%unit, frow, lrow, status)
                if(status /= ignore) this%catextend = .false.
                status = ok
            end if
            if(this%vs_extend .eqv. .true.) then
                ! try reading varichck and specfit extensions
                status = ignore
                call this%varichck%read(this%unit, frow, lrow, status)
                call this%specfit%read(this%unit, frow, lrow, status)
                if(status /= ignore) this%vs_extend = .false.
                status = ok
            end if

            if(present(aux)) then
                if(aux .eqv. .true.) then
                    call this%region%read(this%unit, status, frow=frow, lrow=lrow)
                end if
            end if

        end do

        if(present(filename)) call this%close(status)

    end subroutine read_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write the catsrclist type to a FITS file
    !
    !
    !
    !> @param[in] this
    !> @param[in, optional] aux
    !> @param[out] filename
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tCatSrcList(this, status, aux, filename)
        implicit none

        integer, intent(inout) :: status
        logical, intent(in), optional :: aux
        class(tCatSrcList) :: this
        integer :: rowsize, ii, frow, nelements, lrow, nrows
        character(*), optional, intent(in) :: filename

        character(*), parameter :: function = "write_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(present(filename)) then
            call this%open(filename, create, status)
            if(status /= ok) return
        end if

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(this%unit, BINARY_TBL, this%extname, status)
        if(status /= ignore) then
            !~       print *, "create extension"
            status = ok
            call this%create_extension(status)
        else
            status = ok
        !~       print *, "extension already exists"
        end if

        call ffits_get_rowsize(this%unit, rowsize, status)
        nelements = this%nelements

        call fwrite_stdhdr(this%unit, this%stdhdr, status)

        ! write 1D arrays
        do frow = 1, nelements, rowsize
            nrows = min(nelements - frow + 1, rowsize)
            lrow = frow + nrows - 1

            call this%id_src%write(this%unit, status, frow=frow, lrow=lrow)
            call this%id_cluster%write(this%unit, status, frow=frow, lrow=lrow)

            call this%detuid%write(this%unit, status, frow=frow, lrow=lrow)
            call this%owner%write(this%unit, status, frow=frow, lrow=lrow)
            call this%ra%write(this%unit, status, frow=frow, lrow=lrow)
            call this%dec%write(this%unit, status, frow=frow, lrow=lrow)
            call this%radec_err%write(this%unit, status, frow=frow, lrow=lrow)

            call this%quality%write(this%unit, status, frow=frow, lrow=lrow)
            call this%qf%write(this%unit, status, frow=frow, lrow=lrow)
            call this%lII%write(this%unit, status, frow=frow, lrow=lrow)
            call this%bII%write(this%unit, status, frow=frow, lrow=lrow)

            call this%dist_nn%write(this%unit, status, frow=frow, lrow=lrow)
            call this%srcdensity%write(this%unit, status, frow=frow, lrow=lrow)
            call this%tstart%write(this%unit, status, frow=frow, lrow=lrow)
            call this%tstop%write(this%unit, status, frow=frow, lrow=lrow)

            call this%ext%write(this%unit, status, frow=frow, lrow=lrow)
            call this%ext_err%write(this%unit, status, frow=frow, lrow=lrow)
            call this%ext_like%write(this%unit, status, frow=frow, lrow=lrow)
            call this%cutrad%write(this%unit, status, frow=frow, lrow=lrow)
            call this%maskfrac%write(this%unit, status, frow=frow, lrow=lrow)
            call this%rawx%write(this%unit, status, frow=frow, lrow=lrow)
            call this%rawy%write(this%unit, status, frow=frow, lrow=lrow)

            call this%hr_1%write(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_1_err%write(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_2%write(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_2_err%write(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_3%write(this%unit, status, frow=frow, lrow=lrow)
            call this%hr_3_err%write(this%unit, status, frow=frow, lrow=lrow)

            call this%eb0%write(this%unit, frow, lrow, status)
            call this%eb1%write(this%unit, frow, lrow, status)
            call this%eb2%write(this%unit, frow, lrow, status)
            call this%eb3%write(this%unit, frow, lrow, status)
            call this%eb4%write(this%unit, frow, lrow, status)

            call this%ape%write(this%unit, frow, lrow, status)

            if(this%catextend .eqv. .true.) then
                call this%extend%write(this%unit, frow, lrow, status)
            end if

            if(this%vs_extend .eqv. .true.) then
                call this%varichck%write(this%unit, frow, lrow, status)
                call this%specfit%write(this%unit, frow, lrow, status)
            end if

            if(present(aux)) then
                if(aux .eqv. .true.) then
                    call this%region%write(this%unit, status, frow=frow, lrow=lrow)
                end if
            end if

        end do

        if(present(filename)) call this%close(status)

    end subroutine write_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> reduce all entries in the columns to contain only a subset of the complete
    !> by providing an array list with the desired indices
    !
    !
    !> @param[in] this
    !> @param[in] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_index_tCatSrcList(this, order, status)
        implicit none

        integer, intent(inout) :: status

        class(tCatSrcList) :: this

        integer, dimension(:), intent(in) :: order

        character(*), parameter :: function = "apply_index_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(this%nelements == 0) return

        this%nelements = size(order)

        call this%id_src%apply_index(order, status)
        call this%id_cluster%apply_index(order, status)

        call this%detuid%apply_index(order, status)
        call this%owner%apply_index(order, status)
        call this%ra%apply_index(order, status)
        call this%dec%apply_index(order, status)
        call this%radec_err%apply_index(order, status)

        call this%quality%apply_index(order, status)
        call this%qf%apply_index(order, status)
        call this%lII%apply_index(order, status)
        call this%bII%apply_index(order, status)

        call this%dist_nn%apply_index(order, status)
        call this%srcdensity%apply_index(order, status)
        call this%tstart%apply_index(order, status)
        call this%tstop%apply_index(order, status)

        call this%ext%apply_index(order, status)
        call this%ext_err%apply_index(order, status)
        call this%ext_like%apply_index(order, status)
        call this%cutrad%apply_index(order, status)
        call this%maskfrac%apply_index(order, status)
        call this%rawx%apply_index(order, status)
        call this%rawy%apply_index(order, status)

        call this%hr_1%apply_index(order, status)
        call this%hr_1_err%apply_index(order, status)
        call this%hr_2%apply_index(order, status)
        call this%hr_2_err%apply_index(order, status)
        call this%hr_3%apply_index(order, status)
        call this%hr_3_err%apply_index(order, status)

        ! aux columns
        call this%region%apply_index(order, status)
        call this%id_match%apply_index(order, status)

        call this%eb0%apply_index(order, status)
        call this%eb1%apply_index(order, status)
        call this%eb2%apply_index(order, status)
        call this%eb3%apply_index(order, status)
        call this%eb4%apply_index(order, status)
        call this%ape%apply_index(order, status)

        if(this%catextend .eqv. .true.) then
            call this%extend%apply_index(order, status)
        end if

        if(this%vs_extend .eqv. .true.) then
            call this%specfit%apply_index(order, status)
            call this%varichck%apply_index(order, status)
        end if

    end subroutine apply_index_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> merge all entries of the columns with a second tCatSrcList
    !
    !
    !> @param[in] Cat_A, Cat_B
    !> @param[in] Cat_C
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine merge_CatSrcList(Cat_A, Cat_B, Cat_C, status)
        implicit none

        integer, intent(inout) :: status

        class(tCatSrcList), intent(in) :: Cat_A
        class(tCatSrcList), intent(in) :: Cat_B
        class(tCatSrcList), intent(out) :: Cat_C

        character(*), parameter :: function = "merge_CatSrcList"

        if(status > 100 .or. status == s_errRep) return

        ! copy all important informations
        Cat_C%table        = Cat_A%table
        Cat_C%stdhdr        = Cat_A%stdhdr

        if(Cat_A%nelements == 0 .and. Cat_B%nelements == 0) then
            call wrlog(function, "cats to merge have zero elements - nothing to be done here")
            return
        end if

        Cat_C%nelements = Cat_A%nelements + Cat_B%nelements

        call merge_tColumn(Cat_A%id_src, Cat_B%id_src, Cat_C%id_src, status)
        call merge_tColumn(Cat_A%id_cluster, Cat_B%id_cluster, Cat_C%id_cluster, status)

        call merge_tColumn(Cat_A%detuid, Cat_B%detuid, Cat_C%detuid, status)
        call merge_tColumn(Cat_A%owner, Cat_B%owner, Cat_C%owner, status)
        call merge_tColumn(Cat_A%ra, Cat_B%ra, Cat_C%ra, status)
        call merge_tColumn(Cat_A%dec, Cat_B%dec, Cat_C%dec, status)
        call merge_tColumn(Cat_A%radec_err, Cat_B%radec_err, Cat_C%radec_err, status)

        call merge_tColumn(Cat_A%quality, Cat_B%quality, Cat_C%quality, status)
        call merge_tColumn(Cat_A%qf, Cat_B%qf, Cat_C%qf, status)
        call merge_tColumn(Cat_A%lII, Cat_B%lII, Cat_C%lII, status)
        call merge_tColumn(Cat_A%bII, Cat_B%bII, Cat_C%bII, status)

        call merge_tColumn(Cat_A%dist_nn, Cat_B%dist_nn, Cat_C%dist_nn, status)
        call merge_tColumn(Cat_A%srcdensity, Cat_B%srcdensity, Cat_C%srcdensity, status)
        call merge_tColumn(Cat_A%tstart, Cat_B%tstart, Cat_C%tstart, status)
        call merge_tColumn(Cat_A%tstop, Cat_B%tstop, Cat_C%tstop, status)

        call merge_tColumn(Cat_A%ext, Cat_B%ext, Cat_C%ext, status)
        call merge_tColumn(Cat_A%ext_err, Cat_B%ext_err, Cat_C%ext_err, status)
        call merge_tColumn(Cat_A%ext_like, Cat_B%ext_like, Cat_C%ext_like, status)

        call merge_tColumn(Cat_A%cutrad, Cat_B%cutrad, Cat_C%cutrad, status)
        call merge_tColumn(Cat_A%maskfrac, Cat_B%maskfrac, Cat_C%maskfrac, status)
        call merge_tColumn(Cat_A%rawx, Cat_B%rawx, Cat_C%rawx, status)
        call merge_tColumn(Cat_A%rawy, Cat_B%rawy, Cat_C%rawy, status)

        call merge_tColumn(Cat_A%hr_1, Cat_B%hr_1, Cat_C%hr_1, status)
        call merge_tColumn(Cat_A%hr_1_err, Cat_B%hr_1_err, Cat_C%hr_1_err, status)
        call merge_tColumn(Cat_A%hr_2, Cat_B%hr_2, Cat_C%hr_2, status)
        call merge_tColumn(Cat_A%hr_2_err, Cat_B%hr_2_err, Cat_C%hr_2_err, status)

        call merge_tColumn(Cat_A%hr_3, Cat_B%hr_3, Cat_C%hr_3, status)
        call merge_tColumn(Cat_A%hr_3_err, Cat_B%hr_3_err, Cat_C%hr_3_err, status)

        ! aux columns
        call merge_tColumn(Cat_A%region, Cat_B%region, Cat_C%region, status)

        call merge_EB(Cat_A%eb0, Cat_B%eb0, Cat_C%eb0, status)
        call merge_EB(Cat_A%eb1, Cat_B%eb1, Cat_C%eb1, status)
        call merge_EB(Cat_A%eb2, Cat_B%eb2, Cat_C%eb2, status)
        call merge_EB(Cat_A%eb3, Cat_B%eb3, Cat_C%eb3, status)
        call merge_EB(Cat_A%eb4, Cat_B%eb4, Cat_C%eb4, status)

        call merge_Ape(Cat_A%ape, Cat_B%ape, Cat_C%ape, status)

        if((Cat_A%catextend .eqv. .true.) .and. (Cat_B%catextend .eqv. .true.)) then
            Cat_C%catextend = .true.
            call merge_CatExtend(Cat_A%extend, Cat_B%extend, Cat_C%extend, status)
        else
            Cat_C%catextend = .false.
        end if

        if((Cat_A%vs_extend .eqv. .true.) .and. (Cat_B%vs_extend .eqv. .true.)) then
            Cat_C%vs_extend = .true.
            call merge_varichck(Cat_A%varichck, Cat_B%varichck, Cat_C%varichck, status)
            call merge_specfit(Cat_A%specfit, Cat_B%specfit, Cat_C%specfit, status)
        else
            Cat_C%vs_extend = .false.
        end if

    end subroutine merge_CatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine to mark sources in regions by rectangle ra_min / ...
    !> the overlap parameter defining the boarder region (deg)
    !
    !> @param[in] this
    !> @param[in, optional] overlap
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine flag_region_ra_dec_tCatSrcList(this, ra_min_in, ra_max_in, dec_min_in, dec_max_in, status, overlap, value)
        implicit none

        integer, intent(inout) :: status
        class(tCatSrcList) :: this

        real(r8) :: dec_min, dec_max
        real(r8), intent(in) :: ra_min_in, ra_max_in, dec_min_in, dec_max_in
        real(r8), dimension(this%nelements) :: ra, dec, ra_min, ra_max, overlap_ra
        integer :: ii, old_value
        integer, intent(in), optional :: value
        real(r8), intent(in), optional :: overlap

        character(*), parameter :: function = "flag_region_ra_dec_tCatSrcList"

        ra = mod(this%ra%value, 360.0_r8)
        dec = this%dec%value
        ra_min = mod(ra_min_in, 360.0_r8)
        ra_max = mod(ra_max_in, 360.0_r8)

        ! potentially correct for negative ra
        where (ra < tiny(0.0_r8))
            ra = ra + 360.0_r8
        end where

        where (ra_min < tiny(0.0_r8))
            ra_min = ra_min + 360.0_r8
        end where

        where (ra_max < tiny(0.0_r8))
            ra_max = ra_max + 360.0_r8
        end where

        overlap_ra = overlap / cos ( dec * PI / 180.0_r8)

        if(present(overlap)) then
            ! correction center for the ra overlap
            ! adjust the overlap for declination 1 / cos (angle)
            ra_min = ra_min - overlap_ra
            ra_max = ra_max + overlap_ra
            dec_min = dec_min_in - overlap
            dec_max = dec_max_in + overlap
        else
            ! ra_min / ra_max remain unchanged besides the corrections for negative values and mod 360
            dec_min = dec_min_in
            dec_max = dec_max_in
        end if

        do ii = 1, this%nelements

            ! differ two cases to replace the old region value with a custom value or 1 if it is within a region
            if(present(value)) then
                ! since we are on a sphere and 0/360 is the same we have
                ! two cases ra_min > ra_ max and ra_min < ra_max
                if(ra_min(ii) < ra_max(ii)) then
                    old_value = this%region%value(ii)
                    this%region%value(ii) = value
                    if(ra(ii) < ra_min(ii)) this%region%value(ii) = old_value
                    if(ra(ii) > ra_max(ii)) this%region%value(ii) = old_value
                    if(dec(ii) < dec_min) this%region%value(ii) = old_value
                    if(dec(ii) > dec_max) this%region%value(ii) = old_value
                else
                    if(ra(ii) >= ra_min(ii)) this%region%value(ii) = value
                    if(ra(ii) <= ra_max(ii)) this%region%value(ii) = value
                    if(dec(ii) >= dec_min) this%region%value(ii) = value
                    if(dec(ii) <= dec_max) this%region%value(ii) = value
                end if

            else
                if(ra_min(ii) < ra_max(ii)) then
                    old_value = this%region%value(ii)
                    this%region%value(ii) = 1
                    if(ra(ii) < ra_min(ii)) this%region%value(ii) = old_value
                    if(ra(ii) > ra_max(ii)) this%region%value(ii) = old_value
                    if(dec(ii) < dec_min) this%region%value(ii) = old_value
                    if(dec(ii) > dec_max) this%region%value(ii) = old_value
                else
                    if(ra(ii) >= ra_min(ii)) this%region%value(ii) = 1
                    if(ra(ii) <= ra_max(ii)) this%region%value(ii) = 1
                    if(dec(ii) >= dec_min) this%region%value(ii) = 1
                    if(dec(ii) <= dec_max) this%region%value(ii) = 1
                end if

            end if

        end do

    end subroutine flag_region_ra_dec_tCatSrcList

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> apply the flux filter on the cat list
    !
    !
    !
    !> @param[inout] this
    !> @param[inout] flux_th
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_filter_flux_tCatSrcList(this, flux_th, status)
        implicit none

        integer, intent(inout) :: status
        integer(i4), dimension(:), allocatable :: order, mask
        integer :: ii, nelements

        class(tCatSrcList) :: this
        real(r4), intent(in) :: flux_th

        character(*), parameter :: function = "apply_filter_flux_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(this%nelements == 0) return

        nelements = this%nelements
        call alloc(order, nelements, status)
        call alloc(mask, nelements, status)
        order = [(ii,ii=1, nelements,1)]
        mask = false

        do ii = 1, nelements
          if(this%eb0%flux%value(ii) > flux_th) then
            mask(ii) = true
          end if
        end do

        order = pack(order, mask == true)

        call this%apply_index(order, status)

        call dealloc(order, status)
        call dealloc(mask, status)

    end subroutine apply_filter_flux_tCatSrcList

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> apply the region filter on the event list
    !
    !
    !
    !> @param[inout] this
    !> @param[inout] regtable
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine apply_filter_region_tCatSrcList(this, regtable, status)
        implicit none

        integer, intent(inout) :: status
        integer(i4), dimension(:), allocatable :: order, mask
        integer :: ii, nelements

        class(tCatSrcList) :: this
        type(tRegTable) :: regtable

        ! for using srctool spatial region selections
        type(region_type) :: srctool_region
        type(coord_type) :: srctool_coord

        character(*), parameter :: function = "apply_filter_region_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(this%nelements == 0) return

        nelements = this%nelements
        call alloc(order, nelements, status)
        call alloc(mask, nelements, status)
        order = [(ii,ii=1, nelements,1)]
        mask = false

        call regtable%get_srctool_region(srctool_region, status)

        ! THIS ASSUMES THAT ONLY ONE TYPE OF COORDSYS in all regions
        do ii = 1, nelements
            srctool_coord%long = this%ra%value(ii)
            srctool_coord%lat = this%dec%value(ii)
            !            srctool_coord%x = this%x(ii)
            !            srctool_coord%y = this%y(ii)
            !            srctool_coord%rawx = this%rawx(ii)
            !            srctool_coord%rawy = this%rawy(ii)
            srctool_coord%coordsys_code = COORDSYS_CODE_ICRS
            call srctool_regions_test_if_coord_in_region(srctool_coord, srctool_region, mask(ii), status)
            if(status /= ok) status = s_errRep
        end do

        order = pack(order, mask == true)

        call this%apply_index(order, status)

        call dealloc(order, status)
        call dealloc(mask, status)

    end subroutine apply_filter_region_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> sort the type entries according to one array
    !
    !
    !
    !> @param[in] this
    !> @param[in] array
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_tCatSrcList(this, array, status, ascend)
        implicit none

        integer, intent(inout) :: status
        integer, optional :: ascend

        class(tCatSrcList) :: this
        class(*), dimension(:) :: array

        integer, dimension(:), allocatable :: order

        real(r8),allocatable, dimension(:) :: tmparray_r8
        real(r4),allocatable, dimension(:) :: tmparray_r4
        integer(i4),allocatable, dimension(:) :: tmparray_i4
        integer(i2),allocatable, dimension(:) :: tmparray_i2
        integer(i1),allocatable, dimension(:) :: tmparray_i1

        character(*), parameter :: function = "sort_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(this%nelements /= size(array)) then
            call wrerr(function // " tCatSrcList(_ext) type", s_err, "length of sorting array does not conform to length of type")
            status = s_errRep
        end if

        select type(array)
            type is ( real(r4) )
            tmparray_r4 = array
            call sort_real(tmparray_r4 ,status, order)
            type is ( real(r8) )
            tmparray_r8 = array
            call sort_double(tmparray_r8 ,status, order)
            type is ( integer(i4) )
            tmparray_i4 = array
            call sort_i4(tmparray_i4 ,status, order)
            type is ( integer(i2) )
            tmparray_i2 = array
            call sort_i2(tmparray_i2 ,status, order)
            type is ( integer(i1) )
            tmparray_i1 = array
            call sort_i1(tmparray_i1 ,status, order)
            class default
            call wrstop(function, "sorting tCatSrcList with the current array type is not supported for tCatSrcList")
        end select

        if(present(ascend)) then
            if(ascend == false) call reverse_array(order)
        end if


        call this%apply_index(order, status)

        call dealloc(order, status)
        call dealloc(tmparray_i1, status)
        call dealloc(tmparray_i2, status)
        call dealloc(tmparray_i4, status)
        call dealloc(tmparray_r4, status)
        call dealloc(tmparray_r8, status)

    end subroutine sort_tCatSrcList

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> subroutine to check an array of positions to be within a rectangle
    !
    !
    !
    !> @param[in] x, y, xmin, xmax, ymin, ymax
    !> @param[out] mask
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine within_rectangle_r8(x, y, xmin, xmax, ymin, ymax, mask, status)
        implicit none

        integer, intent(inout) :: status
        real(r8), intent(in), dimension(:) :: x, y
        real(r8), intent(in) :: xmin, xmax, ymin, ymax
        integer(i4), intent(out), dimension(size(x)) :: mask

        integer :: ii

        character(*), parameter :: function = "within_rectangle_r8"

        if(status > 100 .or. status == s_errRep) return

        mask = 0

        where (x >= xmin .and. x <= xmax .and. y >= ymin .and. y <= ymax)
            mask = 1
        end where

    end subroutine within_rectangle_r8

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine calc_dist_Xnn_tCatSrcList.
    !> @brief
    !> calculate the distance to the Xth closest source
    !
    !
    !> @param this, Nth_source
    !> @param[in] p_srcdensity
    !> @param[out]
    !> @param[inout] this, status
    !---------------------------------------------------------------------------

    subroutine calc_srcdensity_tCatSrcList(this, Nth_source, status)
        implicit none

        ! input / output param
        integer, intent(inout) :: status
        class(tCatSrcList) :: this
        integer, intent(in) :: Nth_source

        ! internal param
        integer :: ii
        ! matrix to store all distances
        real(r4), dimension(this%nelements, this%nelements) :: dist_NN

        character(*), parameter :: function = "calc_srcdensity_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(Nth_source < 1) then
            call wrstop(function, "the number for the Nth closest source is < 1")
        end if

        if(Nth_source > this%nelements) then
          call wrlog(function, "no local density of source can be calculated - too few input sources")
          return
        end if

        do ii = 1, this%nelements

            dist_NN(ii, ii) = 0.0_r4

            ! exception for nelements = 1
            if(this%nelements == 1) then
                this%dist_nn%value(ii) = 0.0_r4
                exit
            end if

            if(ii<this%nelements) then
                dist_NN(ii,ii+1: ) =  distance_gc_multi(this%ra%value(ii), this%dec%value(ii), this%ra%value(ii+1:), this%dec%value(ii+1:), true)
            end if
            dist_NN(:, ii) = dist_NN(ii, :)

            ! matrix will be disturbed at this point by rearranging each row according to the distance
            ! find the closest and Xth furthest source for the distance
            ! sort the row in ascending order
            call sort(dist_NN(ii, :), status)

            ! closest source which is not itself therefore 2 not 1
            ! since ra / dec are provided in deg calculate in [sec]
            this%dist_nn%value(ii) = dist_NN(ii, 2)*3600.0_r4
            ! get the distance to Xth source
            ! including itself

            ! spherical segment 2*PI*(1-cos(THETA)) with THETA being the angle between two rays
            this%srcdensity%value(ii) = 2*PI*(1-cos(dist_NN(ii, Nth_source) * PI / 180.0_r8 ))

        end do


    end subroutine calc_srcdensity_tCatSrcList

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine locate_souuid_tCatSrcList.
    !> @brief
    !> get the index / order of a specific souuid
    !
    !
    !> @param this, souuid, order
    !> @param[in]
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine locate_souuid_tCatSrcList(this, souuid, order, status)
        implicit none

        ! input / output param
        integer, intent(inout) :: status
        class(tCatSrcList) :: this
        character(*), intent(in) :: souuid
        integer, intent(out) :: order

        !internal param

        character(*), parameter :: function = "locate_souuid_tCatSrcList"

        if(status > 100 .or. status == s_errRep) return

        if(this%nelements < 1) then
            call wrerr(function, s_err_no_exit, "catalog contains < 1 elements - cannot locate souuid")
        end if

        do order = 1, this%nelements
            if(this%extend%souuid%value(order) == souuid) then
                return
            end if
        end do

        call wrerr(function, s_err_no_exit, "cannot locate souuid in catalog - " // trim(this%filename))

    end subroutine locate_souuid_tCatSrcList

        !---------------------------------------------------------------------------
        !> @author
        !> Christoph Grossberger
        !
        ! DESCRIPTION:
        !> Brief description of routine apply_filter_ownerflag_tCatSrcList.
        !> @brief
        !> apply a filter to catsrclist using owner column and a ownerflag (IAND)
        !
        !
        !> @param this, ownerflag
        !> @param[in]
        !> @param[out]
        !> @param[inout] status
        !---------------------------------------------------------------------------

        subroutine apply_filter_ownerflag_tCatSrcList(this, ownerflag, status)
            implicit none

            ! input / output param
            integer, intent(inout) :: status
            class(tCatSrcList) :: this
            integer, intent(in) :: ownerflag

            !internal param
            integer, dimension(this%nelements) :: owner
            integer, dimension(:), allocatable :: order

            character(*), parameter :: function = "apply_filter_ownerflag_tCatSrcList"

            if(status > 100 .or. status == s_errRep) return

            if(this%nelements == 0) return

            owner = int(IOR(this%owner%value, int(ownerflag,kind=1)))
            order = where_i4(owner, [ownerflag], status)

            call this%apply_index(order, status)

        end subroutine apply_filter_ownerflag_tCatSrcList

end module catprep_srclist_mod
