! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module catprep_varichck_mod
  use param
  use erolib_general_mod
  use erolib_ffits_col_mod
  use erolib_sort_mod
  use erolib_tbintable_mod
  implicit none

  type, public :: tVarichck

    integer :: nelements
    type(r8_tColumn) :: mean_rate
    type(r8_tColumn) :: mean_error
    type(r8_tColumn) :: sigma_rms
    type(r8_tColumn) :: time4_min
    type(r8_tColumn) :: time4_max
    type(r8_tColumn) :: rate1_min
    type(r8_tColumn) :: rate1_max
    type(r8_tColumn) :: rate1_min_err
    type(r8_tColumn) :: rate1_max_err

    ! new columns varichck_extended
    type(r8_tColumn) :: ssigma_rms_err
    type(r8_tColumn) :: ampl_max
    type(r8_tColumn) :: pfactor_mav
    type(r8_tColumn) :: nlv

  contains
    procedure :: init => init_tVarichck
    procedure :: free => free_tVarichck
    procedure :: writeToFits => write_tVarichck
    procedure :: writeToFits_frow_lrow => write_frow_lrow_tVarichck
    generic ::  write => writeToFits, writeToFits_frow_lrow
    procedure :: readFromFits => read_tVarichck
    procedure :: readFromFits_frow_lrow => read_frow_lrow_tVarichck
    generic :: read => readFromFits, readFromFits_frow_lrow
    procedure :: apply_index => apply_index_tVarichck
    procedure :: insert_into_tBinTable => insert_into_tBinTable_tVarichck
  end type tVarichck

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge all entries of the columns with a second Varichck
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_varichck(Varichck_A, Varichck_B, Varichck_C, status)
    implicit none

    integer, intent(inout) :: status

    class(tVarichck), intent(in) :: Varichck_A
    class(tVarichck), intent(in) :: Varichck_B
    class(tVarichck), intent(out) :: Varichck_C

    character(*), parameter :: function = "merge_varichck"

    if(status > 100 .or. status == s_errRep) return

    Varichck_C%nelements = Varichck_A%nelements + Varichck_B%nelements

    call merge_tColumn(Varichck_A%mean_rate, Varichck_B%mean_rate, Varichck_C%mean_rate, status)
    call merge_tColumn(Varichck_A%mean_error, Varichck_B%mean_error, Varichck_C%mean_error, status)
    call merge_tColumn(Varichck_A%sigma_rms, Varichck_B%sigma_rms, Varichck_C%sigma_rms, status)
    call merge_tColumn(Varichck_A%time4_min, Varichck_B%time4_min, Varichck_C%time4_min, status)
    call merge_tColumn(Varichck_A%time4_max, Varichck_B%time4_max, Varichck_C%time4_max, status)
    call merge_tColumn(Varichck_A%rate1_min, Varichck_B%rate1_min, Varichck_C%rate1_min, status)
    call merge_tColumn(Varichck_A%rate1_max, Varichck_B%rate1_max, Varichck_C%rate1_max, status)

    call merge_tColumn(Varichck_A%rate1_min_err, Varichck_B%rate1_min_err, Varichck_C%rate1_min_err, status)
    call merge_tColumn(Varichck_A%rate1_max_err, Varichck_B%rate1_max_err, Varichck_C%rate1_max_err, status)

    call merge_tColumn(Varichck_A%ssigma_rms_err, Varichck_B%ssigma_rms_err, Varichck_C%ssigma_rms_err, status)
    call merge_tColumn(Varichck_A%ampl_max, Varichck_B%ampl_max, Varichck_C%ampl_max, status)
    call merge_tColumn(Varichck_A%pfactor_mav, Varichck_B%pfactor_mav, Varichck_C%pfactor_mav, status)
    call merge_tColumn(Varichck_A%nlv, Varichck_B%nlv, Varichck_C%nlv, status)

  end subroutine merge_varichck

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tVarichck
  !> allocate the memory for this type
  !
  !
  !> @param[inout] this
  !> @param[in]  tel_id, band_id, readwrite, nelements
  !> @param[out]
  !> @param[inout]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine init_tVarichck(this, nelements, status)
    implicit none

    integer, intent(inout) :: status
    integer, intent(in) ::  nelements

    class(tVarichck) :: this
    character(*), parameter :: function = "init_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = nelements
    call this%mean_rate%init("MEAN_RATE", nelements, status)
    call this%mean_error%init("MEAN_ERROR", nelements, status)
    call this%sigma_rms%init("SSIGMA_RMS", nelements, status)
    call this%time4_min%init("TIME4_MIN", nelements, status)
    call this%time4_max%init("TIME4_MAX", nelements, status)
    call this%rate1_min%init("RATE1_MIN", nelements, status)
    call this%rate1_max%init("RATE1_MAX", nelements, status)
    call this%rate1_min_err%init("RATE1_MIN_ERR", nelements, status)
    call this%rate1_max_err%init("RATE1_MAX_ERR", nelements, status)

    call this%ssigma_rms_err%init("SSIGMA_RMS_SS", nelements, status)
    call this%ampl_max%init("AMPL_MAX", nelements, status)
    call this%pfactor_mav%init("PFACTOR_MAV", nelements, status)
    call this%nlv%init("NLV", nelements, status)

  end subroutine init_tVarichck

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tVarichck
  !> free the tVarichck
  !
  !
  !> @param[inout] this
  !> @param[in]
  !> @param[out]
  !> @param[inout]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_tVarichck(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj

    class(tVarichck) :: this
    character(*), parameter :: function = "free_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = 0
    call this%mean_rate%free(status)
    call this%mean_error%free(status)
    call this%sigma_rms%free(status)
    call this%time4_min%free(status)
    call this%time4_max%free(status)
    call this%rate1_min%free(status)
    call this%rate1_max%free(status)
    call this%rate1_min_err%free(status)
    call this%rate1_max_err%free(status)

    call this%ssigma_rms_err%free(status)
    call this%ampl_max%free(status)
    call this%pfactor_mav%free(status)
    call this%nlv%free(status)

  end subroutine free_tVarichck

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tVarichck
  !> write a subset of the tb data to fits
  !
  !
  !> @param[inout] this
  !> @param[in] unit
  !> @param[in] frow, lrow
  !> @param[in, optional] extname
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_tVarichck(this, unit, status, extname)
    implicit none

    integer, intent(inout) :: status
    class(tVarichck) :: this
    integer :: ii, jj
    integer, intent(in) :: unit
    integer :: frow, lrow, nelements, rowsize, nrows

    character(*), intent(in), optional :: extname

    character(*), parameter :: function = "write_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    call ffits_get_rowsize(unit, rowsize, status)
    nelements = this%nelements

    if(present(extname)) call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      call this%mean_rate%write(unit, status, frow=frow, lrow=lrow)
      call this%mean_error%write(unit, status, frow=frow, lrow=lrow)
      call this%sigma_rms%write(unit, status, frow=frow, lrow=lrow)
      call this%time4_min%write(unit, status, frow=frow, lrow=lrow)
      call this%time4_max%write(unit, status, frow=frow, lrow=lrow)
      call this%rate1_min%write(unit, status, frow=frow, lrow=lrow)
      call this%rate1_max%write(unit, status, frow=frow, lrow=lrow)
      call this%rate1_min_err%write(unit, status, frow=frow, lrow=lrow)
      call this%rate1_max_err%write(unit, status, frow=frow, lrow=lrow)

      call this%ssigma_rms_err%write(unit, status, frow=frow, lrow=lrow)
      call this%ampl_max%write(unit, status, frow=frow, lrow=lrow)
      call this%pfactor_mav%write(unit, status, frow=frow, lrow=lrow)
      call this%nlv%write(unit, status, frow=frow, lrow=lrow)

    end do

  end subroutine write_tVarichck

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tVarichck
  !> read a subset of the tb data to fits
  !
  !
  !> @param[inout] this
  !> @param[in] unit
  !> @param[in] frow, lrow
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_tVarichck(this, unit, status)
    implicit none

    integer, intent(inout) :: status
    class(tVarichck) :: this
    integer :: ii
    integer, intent(in) :: unit
    integer :: frow, lrow, nelements, rowsize, nrows

    character(*), parameter :: function = "read_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    call ffits_get_rowsize(unit, rowsize, status)
    nelements = this%nelements

    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1

      call this%mean_rate%read(unit, status, frow=frow, lrow=lrow)
      call this%mean_error%read(unit, status, frow=frow, lrow=lrow)
      call this%sigma_rms%read(unit, status, frow=frow, lrow=lrow)
      call this%time4_min%read(unit, status, frow=frow, lrow=lrow)
      call this%time4_max%read(unit, status, frow=frow, lrow=lrow)
      call this%rate1_min%read(unit, status, frow=frow, lrow=lrow)
      call this%rate1_max%read(unit, status, frow=frow, lrow=lrow)
      call this%rate1_min_err%read(unit, status, frow=frow, lrow=lrow)
      call this%rate1_max_err%read(unit, status, frow=frow, lrow=lrow)

           call this%ssigma_rms_err%read(unit, status, frow=frow, lrow=lrow)
      call this%ampl_max%read(unit, status, frow=frow, lrow=lrow)
      call this%pfactor_mav%read(unit, status, frow=frow, lrow=lrow)
      call this%nlv%read(unit, status, frow=frow, lrow=lrow)

    end do

  end subroutine read_tVarichck

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tVarichck
  !> write a subset of the tb data to fits
  !
  !
  !> @param[inout] this
  !> @param[in] unit
  !> @param[in] frow, lrow
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine write_frow_lrow_tVarichck(this, unit, frow, lrow, status)
    implicit none

    integer, intent(inout) :: status
    class(tVarichck) :: this
    integer, intent(in) :: unit,  frow, lrow

    character(*), parameter :: function = "write_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    call wrlog(function, "write varichck")
    call this%mean_rate%write(unit, status, frow=frow, lrow=lrow)
    call this%mean_error%write(unit, status, frow=frow, lrow=lrow)
    call this%sigma_rms%write(unit, status, frow=frow, lrow=lrow)
    call this%time4_min%write(unit, status, frow=frow, lrow=lrow)
    call this%time4_max%write(unit, status, frow=frow, lrow=lrow)
    call this%rate1_min%write(unit, status, frow=frow, lrow=lrow)
    call this%rate1_max%write(unit, status, frow=frow, lrow=lrow)
    call this%rate1_min_err%write(unit, status, frow=frow, lrow=lrow)
    call this%rate1_max_err%write(unit, status, frow=frow, lrow=lrow)

        call this%ssigma_rms_err%write(unit, status, frow=frow, lrow=lrow)
    call this%ampl_max%write(unit, status, frow=frow, lrow=lrow)
    call this%pfactor_mav%write(unit, status, frow=frow, lrow=lrow)
    call this%nlv%write(unit, status, frow=frow, lrow=lrow)

  end subroutine write_frow_lrow_tVarichck

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tVarichck
  !> read a subset of the tb data to fits
  !
  !
  !> @param[inout] this
  !> @param[in] unit
  !> @param[in] frow, lrow
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_frow_lrow_tVarichck(this, unit, frow, lrow, status)
    implicit none

    integer, intent(inout) :: status
    class(tVarichck) :: this
    integer, intent(in) :: unit,  frow, lrow

    character(*), parameter :: function = "read_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    call this%mean_rate%read(unit, status, frow=frow, lrow=lrow)
    call this%mean_error%read(unit, status, frow=frow, lrow=lrow)
    call this%sigma_rms%read(unit, status, frow=frow, lrow=lrow)
    call this%time4_min%read(unit, status, frow=frow, lrow=lrow)
    call this%time4_max%read(unit, status, frow=frow, lrow=lrow)
    call this%rate1_min%read(unit, status, frow=frow, lrow=lrow)
    call this%rate1_max%read(unit, status, frow=frow, lrow=lrow)
    call this%rate1_min_err%read(unit, status, frow=frow, lrow=lrow)
    call this%rate1_max_err%read(unit, status, frow=frow, lrow=lrow)

        call this%ssigma_rms_err%read(unit, status, frow=frow, lrow=lrow)
    call this%ampl_max%read(unit, status, frow=frow, lrow=lrow)
    call this%pfactor_mav%read(unit, status, frow=frow, lrow=lrow)
    call this%nlv%read(unit, status, frow=frow, lrow=lrow)

  end subroutine read_frow_lrow_tVarichck

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> reduce all entries in the columns to contain only a subset of the complete
  !> by providing an array list with the desired indices
  !
  !
  !> @param[in] this
  !> @param[in] order
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_index_tVarichck(this, order, status)
    implicit none

    integer, intent(inout) :: status

    class(tVarichck) :: this

    integer, dimension(:), intent(in) :: order

    character(*), parameter :: function = "apply_index_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = size(order)
    call this%mean_rate%apply_index(order, status)
    call this%mean_error%apply_index(order, status)
    call this%sigma_rms%apply_index(order, status)
    call this%time4_min%apply_index(order, status)
    call this%time4_max%apply_index(order, status)
    call this%rate1_min%apply_index(order, status)
    call this%rate1_max%apply_index(order, status)
    call this%rate1_min_err%apply_index(order, status)
    call this%rate1_max_err%apply_index(order, status)

    call this%ssigma_rms_err%apply_index(order, status)
    call this%ampl_max%apply_index(order, status)
    call this%pfactor_mav%apply_index(order, status)
    call this%nlv%apply_index(order, status)

  end subroutine apply_index_tVarichck

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> subroutine of class tVarichck
  !> insert the contents into a tbintable description
  !
  !
  !> @param[inout] this
  !> @param[in]
  !> @param[out]
  !> @param[inout] table
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine insert_into_tBinTable_tVarichck(this, table, status)
    implicit none

    integer, intent(inout) :: status

    type(tBinTable), intent(inout) :: table

    integer :: ii, jj

    class(tVarichck) :: this
    character(*), parameter :: function = "insert_into_tBinTable_tVarichck"

    if(status > 100 .or. status == s_errRep) return

    call wrlog(function, "CREATE VARICHCK")

    call table%insert_ffits_col(this%mean_rate, status)
    call table%insert_ffits_col(this%mean_error, status)
    call table%insert_ffits_col(this%sigma_rms, status)
    call table%insert_ffits_col(this%time4_min, status)
    call table%insert_ffits_col(this%time4_max, status)
    call table%insert_ffits_col(this%rate1_min, status)
    call table%insert_ffits_col(this%rate1_max, status)
    call table%insert_ffits_col(this%rate1_min_err, status)
    call table%insert_ffits_col(this%rate1_max_err, status)

    call table%insert_ffits_col(this%ssigma_rms_err, status)
    call table%insert_ffits_col(this%ampl_max, status)
    call table%insert_ffits_col(this%pfactor_mav, status)
    call table%insert_ffits_col(this%nlv, status)

  end subroutine insert_into_tBinTable_tVarichck

end module catprep_varichck_mod
