! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

program catprep_update_main
    use param

    use catprep_mod
    use catprep_srclist_mod
    use catprep_skymap_mod
    use catprep_match_cats_mod

    use erolib_general_mod
    use erolib_str_mod

    implicit none

    character(*), parameter :: function = "catprep_update_main"

    integer :: status, ierr, skyfield, index_pos

    type(tCatprep_addnew) :: task
    type(tCatSrcList) :: CatSrcList ! product
    type(tCatSrcList) :: Catalog ! Catalog being produced
    type(tCompCat) :: CompCat ! Comparison catalog
    type(tSkymap) :: Skymap
    type(tQMatch_r4) :: Match

    status = ok

      ! read/open the .par file
    call opast(status)
    if(status /= ok) call wrstop(function, "error read par file")

    ! read the input paramters
    call wrlog(function, "reading parameters")
    call task%read_params(status)
    ! validate the input parameters
    call wrlog(function, "validate parameters")
    call task%validate(status)
    ! sanity check
    if(status /= ok) then
        call wrstop(function, "input parameters are invalid")
    else
        call wrlog(function, "parameters - valid")
    end if

    ! read the infile and the information
    call wrlog(function, "read catsrclist - input")
    call CatSrcList%open(task%srclist, readonly, status)
    call CatSrcList%read(status)
    call CatSrcList%close(status)
        ! add the additional columns
    CatSrcList%catextend = .true.
    if(CatSrcList%catextend .eqv. .true.) then
        call wrlog(function, "extend catsrclist with additional columns")
        call CatSrcList%extend%init(CatSrcList%nelements, status)
    end if
    CatSrcList%vs_extend = .true.
    if(CatSrcList%vs_extend .eqv. .true.) then
        call wrlog(function, "extend catsrclist with varichck / specfit columns")
        call CatSrcList%varichck%init(CatSrcList%nelements, status)
        call CatSrcList%specfit%init(CatSrcList%nelements, status)
    end if

    ! skyfield was read from header at this point
    ! get the comparison catalogs, assuming each filename is the same
    ! SKYMAPS.fits will be a new "Version" with skyfields, adjacent skyfields (index), CatalogName

    if(task%compcat == "") then
        ! lock Skymap directory
        index_pos = index(trim(task%skymap), "/", back=.true.)
        call lock(task%skymap(1:index_pos-1), trim(task%skymap(index_pos+1:)) // ".LOCK")
        call wrlog(function, "lock "//trim(task%skymap) // ".LOCK")

        call Skymap%open(task%skymap, readwrite, status)
        call Skymap%read(status)

        ! in the case of pipeline and survey processing
        skyfield = str2int(CatSrcList%stdhdr%skyfield, status)

        call wrlog(function, "prepare cats")
        call prepare_cats(CompCat, CatSrcList, skyfield, Skymap, task%update, status)
                ! check against old skyfield + catalog of adjacent skyfield(s)
                ! search for matches to identify SOUUID and new unregistered sources

        call wrlog(function, "matching catalogs")
        call match_cat(Match, CatSrcList, CompCat, task%cutval, status)
        ! the matrix and the used comparions catalog are freed at this point since they are not used anymore
        call Match%Cat_M%free(status)
        call dealloc(Match%value, status)

        call wrlog(function, "create new catalogs")
        call create_new_CompCat_own_skyfield(Match%Cat_N, CompCat, skyfield, Skymap, status)
        CompCat%filename = task%Catalog
        if(task%update == true) then
            call wrlog(function, "update catalog")
            call CompCat%update(Match%Cat_N, status)
        end if
    else
        call wrlog(function, "comparison catalog was provided externally")
        call wrlog(function, "no update / new sources will be added to the existing cumulative catalog")
        call CompCat%open(task%compcat, readonly, status, true)
        call CompCat%read(status)
        call CompCat%close(status)
        call wrlog(function, "match_cat")
        call match_cat(Match, CatSrcList, CompCat, task%cutval, status)
                ! the matrix and the used comparions catalog are freed at this point since they are not used anymore
        call Match%Cat_M%free(status)
        call dealloc(Match%value, status)
    end if

    call wrlog(function, "write updated input catalog")
    call wrlog(function, "open")
    call Match%Cat_N%open(task%srclist_out, create, status)
    call wrlog(function, "write")
    call Match%Cat_N%write(status)
    call wrlog(function, "close")
    call Match%Cat_N%close(status)
    call Match%Cat_N%free(status)


    if(task%compcat == "") then
        call wrlog(function, "write updated new comparison catalog")
        call CompCat%open(CompCat%filename, create, status, delete=true)
        call CompCat%write(status)
        call CompCat%close(status)
        call CompCat%free(status)

        call wrlog(function, "update skymaps")
        ! now update the SKYMAPS file
        call Skymap%set_and_write_new_catalog_name(skyfield, task%Catalog, status)
        if(task%update == false) then
            ! assume it is a single scan and update sinle scan information
            call Skymap%set_and_write_lastsscan(skyfield, task%srclist, status)
            call Skymap%set_and_write_sscan(skyfield, task%srclist, Match%Cat_N%stdhdr%tstart, status)
        end if
        call Skymap%close(status)
    end if

   ! unlock the SKYMAPS file
    call unlock(".", trim(task%skymap) // ".LOCK")
    call wrlog(function, "unlock "//trim(task%skymap) // ".LOCK")

    call cpast (status, ierr)

end program catprep_update_main
